@@ -1,6 +1,6 @@
 /*
  This file is part of GNU Taler
- (C) 2021 Taler Systems S.A.
+ (C) 2022 Taler Systems S.A.
 
  GNU Taler is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
@@ -33,7 +33,6 @@ export function QR({ text }: { text: string }): VNode {
   return (
     <div
       style={{
-        width: "100%",
         display: "flex",
         flexDirection: "column",
         alignItems: "left",
@@ -41,11 +40,9 @@ export function QR({ text }: { text: string }): VNode {
     >
       <div
         style={{
-            width: "50%",
-            minWidth: 200,
-            maxWidth: 300,
-            marginRight: "auto",
-            marginLeft: "auto",
+          width: "100%",
+          marginRight: "auto",
+          marginLeft: "auto",
         }}
         ref={divRef}
       />
