@@ -15,26 +15,26 @@
  */
 
 /**
-*
-* @author Sebastian Javier Marchano (sebasjm)
-*/
-
-import { h, VNode, FunctionalComponent } from 'preact';
-import { DepletedTip as TestedComponent } from './DepletedTip';
+ *
+ * @author Sebastian Javier Marchano (sebasjm)
+ */
 
+import { h, VNode, FunctionalComponent } from "preact";
+import { DepletedTip as TestedComponent } from "./DepletedTip";
 
 export default {
-  title: 'DepletedTip',
+  title: "DepletedTip",
   component: TestedComponent,
-  argTypes: {
-  },
+  argTypes: {},
 };
 
-function createExample<Props>(Component: FunctionalComponent<Props>, props: Partial<Props>) {
-  const r = (args: any) => <Component {...args} />
-  r.args = props
-  return r
+function createExample<Props>(
+  Component: FunctionalComponent<Props>,
+  props: Partial<Props>,
+) {
+  const r = (args: any) => <Component {...args} />;
+  r.args = props;
+  return r;
 }
 
-export const Example = createExample(TestedComponent, {
-});
+export const Example = createExample(TestedComponent, {});
