@@ -1,6 +1,6 @@
 /*
  This file is part of GNU Taler
- (C) 2021 Taler Systems S.A.
+ (C) 2021-2023 Taler Systems S.A.
 
  GNU Taler is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
@@ -19,52 +19,77 @@
  * @author Sebastian Javier Marchano (sebasjm)
  */
 
+import {
+  useTranslationContext,
+  HttpError,
+  ErrorType,
+} from "@gnu-taler/web-util/browser";
+import { format } from "date-fns";
 import { Fragment, FunctionComponent, h, VNode } from "preact";
 import { Route, route, Router } from "preact-router";
 import { useCallback, useEffect, useMemo, useState } from "preact/hooks";
-import { Loading } from "./components/exception/loading";
-import { Menu, NotificationCard } from "./components/menu";
-import { useBackendContext } from "./context/backend";
-import { InstanceContextProvider } from "./context/instance";
+import { Loading } from "./components/exception/loading.js";
+import { Menu, NotificationCard } from "./components/menu/index.js";
+import { useBackendContext } from "./context/backend.js";
+import { InstanceContextProvider } from "./context/instance.js";
 import {
   useBackendDefaultToken,
   useBackendInstanceToken,
-  useLocalStorage,
-} from "./hooks";
-import { HttpError } from "./hooks/backend";
-import { Translate, useTranslator } from "./i18n";
-import InstanceCreatePage from "./paths/admin/create";
-import InstanceListPage from "./paths/admin/list";
-import OrderCreatePage from "./paths/instance/orders/create";
-import OrderDetailsPage from "./paths/instance/orders/details";
-import OrderListPage from "./paths/instance/orders/list";
-import ProductCreatePage from "./paths/instance/products/create";
-import ProductListPage from "./paths/instance/products/list";
-import ProductUpdatePage from "./paths/instance/products/update";
-import TransferListPage from "./paths/instance/transfers/list";
-import TransferCreatePage from "./paths/instance/transfers/create";
-import ReservesCreatePage from "./paths/instance/reserves/create";
-import ReservesDetailsPage from "./paths/instance/reserves/details";
-import ReservesListPage from "./paths/instance/reserves/list";
-import ListKYCPage from "./paths/instance/kyc/list";
+  useSimpleLocalStorage,
+} from "./hooks/index.js";
+import { useInstanceKYCDetails } from "./hooks/instance.js";
+import InstanceCreatePage from "./paths/admin/create/index.js";
+import InstanceListPage from "./paths/admin/list/index.js";
+import TokenPage from "./paths/instance/token/index.js";
+import ListKYCPage from "./paths/instance/kyc/list/index.js";
+import OrderCreatePage from "./paths/instance/orders/create/index.js";
+import OrderDetailsPage from "./paths/instance/orders/details/index.js";
+import OrderListPage from "./paths/instance/orders/list/index.js";
+import ProductCreatePage from "./paths/instance/products/create/index.js";
+import ProductListPage from "./paths/instance/products/list/index.js";
+import ProductUpdatePage from "./paths/instance/products/update/index.js";
+import BankAccountCreatePage from "./paths/instance/accounts/create/index.js";
+import BankAccountListPage from "./paths/instance/accounts/list/index.js";
+import BankAccountUpdatePage from "./paths/instance/accounts/update/index.js";
+import ReservesCreatePage from "./paths/instance/reserves/create/index.js";
+import ReservesDetailsPage from "./paths/instance/reserves/details/index.js";
+import ReservesListPage from "./paths/instance/reserves/list/index.js";
+import TemplateCreatePage from "./paths/instance/templates/create/index.js";
+import TemplateUsePage from "./paths/instance/templates/use/index.js";
+import TemplateQrPage from "./paths/instance/templates/qr/index.js";
+import TemplateListPage from "./paths/instance/templates/list/index.js";
+import TemplateUpdatePage from "./paths/instance/templates/update/index.js";
+import WebhookCreatePage from "./paths/instance/webhooks/create/index.js";
+import WebhookListPage from "./paths/instance/webhooks/list/index.js";
+import WebhookUpdatePage from "./paths/instance/webhooks/update/index.js";
+import ValidatorCreatePage from "./paths/instance/otp_devices/create/index.js";
+import ValidatorListPage from "./paths/instance/otp_devices/list/index.js";
+import ValidatorUpdatePage from "./paths/instance/otp_devices/update/index.js";
+import TransferCreatePage from "./paths/instance/transfers/create/index.js";
+import TransferListPage from "./paths/instance/transfers/list/index.js";
 import InstanceUpdatePage, {
-  Props as InstanceUpdatePageProps,
   AdminUpdate as InstanceAdminUpdatePage,
-} from "./paths/instance/update";
-import LoginPage from "./paths/login";
-import NotFoundPage from "./paths/notfound";
-import { Notification } from "./utils/types";
-import { useInstanceKYCDetails } from "./hooks/instance";
-import { format } from "date-fns";
+  Props as InstanceUpdatePageProps,
+} from "./paths/instance/update/index.js";
+import { LoginPage } from "./paths/login/index.js";
+import NotFoundPage from "./paths/notfound/index.js";
+import { Notification } from "./utils/types.js";
+import { LoginToken, MerchantBackend } from "./declaration.js";
+import { Settings } from "./paths/settings/index.js";
+import { dateFormatForSettings, useSettings } from "./hooks/useSettings.js";
 
 export enum InstancePaths {
-  // details = '/',
   error = "/error",
-  update = "/update",
+  settings = "/settings",
+  token = "/token",
 
-  product_list = "/products",
-  product_update = "/product/:pid/update",
-  product_new = "/product/new",
+  bank_list = "/bank",
+  bank_update = "/bank/:bid/update",
+  bank_new = "/bank/new",
+
+  inventory_list = "/inventory",
+  inventory_update = "/inventory/:pid/update",
+  inventory_new = "/inventory/new",
 
   order_list = "/orders",
   order_new = "/order/new",
@@ -78,10 +103,26 @@ export enum InstancePaths {
 
   transfers_list = "/transfers",
   transfers_new = "/transfer/new",
+
+  templates_list = "/templates",
+  templates_update = "/templates/:tid/update",
+  templates_new = "/templates/new",
+  templates_use = "/templates/:tid/use",
+  templates_qr = "/templates/:tid/qr",
+
+  webhooks_list = "/webhooks",
+  webhooks_update = "/webhooks/:tid/update",
+  webhooks_new = "/webhooks/new",
+
+  otp_devices_list = "/otp-devices",
+  otp_devices_update = "/otp-devices/:vid/update",
+  otp_devices_new = "/otp-devices/new",
+
+  interface = "/interface",
 }
 
 // eslint-disable-next-line @typescript-eslint/no-empty-function
-const noop = () => {};
+const noop = () => { };
 
 export enum AdminPaths {
   list_instances = "/instances",
@@ -92,76 +133,86 @@ export enum AdminPaths {
 export interface Props {
   id: string;
   admin?: boolean;
+  path: string;
+  onUnauthorized: () => void;
+  onLoginPass: () => void;
   setInstanceName: (s: string) => void;
 }
 
-export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
-  const [_, updateDefaultToken] = useBackendDefaultToken();
+export function InstanceRoutes({
+  id,
+  admin,
+  path,
+  // onUnauthorized,
+  onLoginPass,
+  setInstanceName,
+}: Props): VNode {
+  const [defaultToken, updateDefaultToken] = useBackendDefaultToken();
   const [token, updateToken] = useBackendInstanceToken(id);
-  const {
-    updateLoginStatus: changeBackend,
-    addTokenCleaner,
-    clearAllTokens,
-  } = useBackendContext();
-  const cleaner = useCallback(() => {
-    updateToken(undefined);
-  }, [id]);
-  const i18n = useTranslator();
+  const { i18n } = useTranslationContext();
 
   type GlobalNotifState = (Notification & { to: string }) | undefined;
   const [globalNotification, setGlobalNotification] =
     useState<GlobalNotifState>(undefined);
 
-  useEffect(() => {
-    addTokenCleaner(cleaner);
-  }, [addTokenCleaner, cleaner]);
-
-  const changeToken = (token?: string) => {
+  const changeToken = (token?: LoginToken) => {
     if (admin) {
       updateToken(token);
     } else {
       updateDefaultToken(token);
     }
+    onLoginPass()
   };
-  const updateLoginStatus = (url: string, token?: string) => {
-    changeBackend(url);
-    if (!token) return;
-    changeToken(token);
-  };
+  // const updateLoginStatus = (url: string, token?: string) => {
+  //   changeToken(token);
+  // };
 
   const value = useMemo(
     () => ({ id, token, admin, changeToken }),
-    [id, token, admin]
+    [id, token, admin],
   );
 
   function ServerErrorRedirectTo(to: InstancePaths | AdminPaths) {
-    return function ServerErrorRedirectToImpl(error: HttpError) {
-      setGlobalNotification({
-        message: i18n`The backend reported a problem: HTTP status #${error.status}`,
-        description: i18n`Diagnostic from ${error.info?.url} is "${error.message}"`,
-        details:
-          error.clientError || error.serverError
-            ? error.error?.detail
-            : undefined,
-        type: "ERROR",
-        to,
-      });
+    return function ServerErrorRedirectToImpl(
+      error: HttpError<MerchantBackend.ErrorDetail>,
+    ) {
+      if (error.type === ErrorType.TIMEOUT) {
+        setGlobalNotification({
+          message: i18n.str`The request to the backend take too long and was cancelled`,
+          description: i18n.str`Diagnostic from ${error.info.url} is "${error.message}"`,
+          type: "ERROR",
+          to,
+        });
+      } else {
+        setGlobalNotification({
+          message: i18n.str`The backend reported a problem: HTTP status #${error.status}`,
+          description: i18n.str`Diagnostic from ${error.info.url} is '${error.message}'`,
+          details:
+            error.type === ErrorType.CLIENT || error.type === ErrorType.SERVER
+              ? error.payload.detail
+              : undefined,
+          type: "ERROR",
+          to,
+        });
+      }
       return <Redirect to={to} />;
     };
   }
 
-  const LoginPageAccessDenied = () => (
-    <Fragment>
+  // const LoginPageAccessDeniend = onUnauthorized
+  const LoginPageAccessDenied = () => {
+    return <Fragment>
       <NotificationCard
         notification={{
-          message: i18n`Access denied`,
-          description: i18n`The access token provided is invalid.`,
+          message: i18n.str`Access denied`,
+          description: i18n.str`Session expired or password changed.`,
           type: "ERROR",
         }}
       />
-      <LoginPage onConfirm={updateLoginStatus} />
+      <LoginPage onConfirm={changeToken} />
     </Fragment>
-  );
+
+  }
 
   function IfAdminCreateDefaultOr<T>(Next: FunctionComponent<any>) {
     return function IfAdminCreateDefaultOrImpl(props?: T) {
@@ -170,15 +221,15 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
           <Fragment>
             <NotificationCard
               notification={{
-                message: i18n`No 'default' instance configured yet.`,
-                description: i18n`Create a 'default' instance to begin using the merchant backoffice.`,
+                message: i18n.str`No 'default' instance configured yet.`,
+                description: i18n.str`Create a 'default' instance to begin using the merchant backoffice.`,
                 type: "INFO",
               }}
             />
             <InstanceCreatePage
               forceId="default"
               onConfirm={() => {
-                route(AdminPaths.list_instances);
+                route(InstancePaths.order_list);
               }}
             />
           </Fragment>
@@ -192,8 +243,10 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
   }
 
   const clearTokenAndGoToRoot = () => {
-    clearAllTokens();
     route("/");
+    // clear all tokens
+    updateToken(undefined)
+    updateDefaultToken(undefined)
   };
 
   return (
@@ -201,8 +254,13 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
       <Menu
         instance={id}
         admin={admin}
+        onShowSettings={() => {
+          route(InstancePaths.interface)
+        }}
+        path={path}
         onLogout={clearTokenAndGoToRoot}
         setInstanceName={setInstanceName}
+        isPasswordOk={defaultToken !== undefined}
       />
       <KycBanner />
       <NotificationCard notification={globalNotification} />
@@ -217,7 +275,6 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
         }}
       >
         <Route path="/" component={Redirect} to={InstancePaths.order_list} />
-
         {/**
          * Admin pages
          */}
@@ -236,18 +293,16 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
             onLoadError={ServerErrorRedirectTo(InstancePaths.error)}
           />
         )}
-
         {admin && (
           <Route
             path={AdminPaths.new_instance}
             component={InstanceCreatePage}
             onBack={() => route(AdminPaths.list_instances)}
             onConfirm={() => {
-              route(AdminPaths.list_instances);
+              route(InstancePaths.order_list);
             }}
           />
         )}
-
         {admin && (
           <Route
             path={AdminPaths.update_instance}
@@ -261,12 +316,11 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
             onNotFound={NotFoundPage}
           />
         )}
-
         {/**
          * Update instance page
          */}
         <Route
-          path={InstancePaths.update}
+          path={InstancePaths.settings}
           component={InstanceUpdatePage}
           onBack={() => {
             route(`/`);
@@ -279,47 +333,100 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
           onUnauthorized={LoginPageAccessDenied}
           onLoadError={ServerErrorRedirectTo(InstancePaths.error)}
         />
-
         {/**
-         * Product pages
+         * Update instance page
          */}
         <Route
-          path={InstancePaths.product_list}
+          path={InstancePaths.token}
+          component={TokenPage}
+          onChange={() => {
+            route(`/`);
+          }}
+          onCancel={() => {
+            route(InstancePaths.order_list)
+          }}
+          onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
+          onUnauthorized={LoginPageAccessDenied}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.error)}
+        />
+        {/**
+         * Inventory pages
+         */}
+        <Route
+          path={InstancePaths.inventory_list}
           component={ProductListPage}
           onUnauthorized={LoginPageAccessDenied}
-          onLoadError={ServerErrorRedirectTo(InstancePaths.update)}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.settings)}
           onCreate={() => {
-            route(InstancePaths.product_new);
+            route(InstancePaths.inventory_new);
           }}
           onSelect={(id: string) => {
-            route(InstancePaths.product_update.replace(":pid", id));
+            route(InstancePaths.inventory_update.replace(":pid", id));
           }}
           onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
         />
         <Route
-          path={InstancePaths.product_update}
+          path={InstancePaths.inventory_update}
           component={ProductUpdatePage}
           onUnauthorized={LoginPageAccessDenied}
-          onLoadError={ServerErrorRedirectTo(InstancePaths.product_list)}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.inventory_list)}
           onConfirm={() => {
-            route(InstancePaths.product_list);
+            route(InstancePaths.inventory_list);
           }}
           onBack={() => {
-            route(InstancePaths.product_list);
+            route(InstancePaths.inventory_list);
           }}
           onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
         />
         <Route
-          path={InstancePaths.product_new}
+          path={InstancePaths.inventory_new}
           component={ProductCreatePage}
           onConfirm={() => {
-            route(InstancePaths.product_list);
+            route(InstancePaths.inventory_list);
           }}
           onBack={() => {
-            route(InstancePaths.product_list);
+            route(InstancePaths.inventory_list);
+          }}
+        />
+        {/**
+         * Bank pages
+         */}
+        <Route
+          path={InstancePaths.bank_list}
+          component={BankAccountListPage}
+          onUnauthorized={LoginPageAccessDenied}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.settings)}
+          onCreate={() => {
+            route(InstancePaths.bank_new);
+          }}
+          onSelect={(id: string) => {
+            route(InstancePaths.bank_update.replace(":bid", id));
+          }}
+          onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
+        />
+        <Route
+          path={InstancePaths.bank_update}
+          component={BankAccountUpdatePage}
+          onUnauthorized={LoginPageAccessDenied}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.inventory_list)}
+          onConfirm={() => {
+            route(InstancePaths.bank_list);
+          }}
+          onBack={() => {
+            route(InstancePaths.bank_list);
+          }}
+          onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
+        />
+        <Route
+          path={InstancePaths.bank_new}
+          component={BankAccountCreatePage}
+          onConfirm={() => {
+            route(InstancePaths.bank_list);
+          }}
+          onBack={() => {
+            route(InstancePaths.bank_list);
           }}
         />
-
         {/**
          * Order pages
          */}
@@ -333,7 +440,7 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
             route(InstancePaths.order_details.replace(":oid", id));
           }}
           onUnauthorized={LoginPageAccessDenied}
-          onLoadError={ServerErrorRedirectTo(InstancePaths.update)}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.settings)}
           onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
         />
         <Route
@@ -349,14 +456,13 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
         <Route
           path={InstancePaths.order_new}
           component={OrderCreatePage}
-          onConfirm={() => {
-            route(InstancePaths.order_list);
+          onConfirm={(orderId: string) => {
+            route(InstancePaths.order_details.replace(":oid", orderId));
           }}
           onBack={() => {
             route(InstancePaths.order_list);
           }}
         />
-
         {/**
          * Transfer pages
          */}
@@ -365,12 +471,11 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
           component={TransferListPage}
           onUnauthorized={LoginPageAccessDenied}
           onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
-          onLoadError={ServerErrorRedirectTo(InstancePaths.update)}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.settings)}
           onCreate={() => {
             route(InstancePaths.transfers_new);
           }}
         />
-
         <Route
           path={InstancePaths.transfers_new}
           component={TransferCreatePage}
@@ -381,6 +486,152 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
             route(InstancePaths.transfers_list);
           }}
         />
+        {/**
+         * Webhooks pages
+         */}
+        <Route
+          path={InstancePaths.webhooks_list}
+          component={WebhookListPage}
+          onUnauthorized={LoginPageAccessDenied}
+          onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.settings)}
+          onCreate={() => {
+            route(InstancePaths.webhooks_new);
+          }}
+          onSelect={(id: string) => {
+            route(InstancePaths.webhooks_update.replace(":tid", id));
+          }}
+        />
+        <Route
+          path={InstancePaths.webhooks_update}
+          component={WebhookUpdatePage}
+          onConfirm={() => {
+            route(InstancePaths.webhooks_list);
+          }}
+          onUnauthorized={LoginPageAccessDenied}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.webhooks_list)}
+          onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
+          onBack={() => {
+            route(InstancePaths.webhooks_list);
+          }}
+        />
+        <Route
+          path={InstancePaths.webhooks_new}
+          component={WebhookCreatePage}
+          onConfirm={() => {
+            route(InstancePaths.webhooks_list);
+          }}
+          onBack={() => {
+            route(InstancePaths.webhooks_list);
+          }}
+        />
+        {/**
+         * Validator pages
+         */}
+        <Route
+          path={InstancePaths.otp_devices_list}
+          component={ValidatorListPage}
+          onUnauthorized={LoginPageAccessDenied}
+          onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.settings)}
+          onCreate={() => {
+            route(InstancePaths.otp_devices_new);
+          }}
+          onSelect={(id: string) => {
+            route(InstancePaths.otp_devices_update.replace(":vid", id));
+          }}
+        />
+        <Route
+          path={InstancePaths.otp_devices_update}
+          component={ValidatorUpdatePage}
+          onConfirm={() => {
+            route(InstancePaths.otp_devices_list);
+          }}
+          onUnauthorized={LoginPageAccessDenied}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.otp_devices_list)}
+          onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
+          onBack={() => {
+            route(InstancePaths.otp_devices_list);
+          }}
+        />
+        <Route
+          path={InstancePaths.otp_devices_new}
+          component={ValidatorCreatePage}
+          onConfirm={() => {
+            route(InstancePaths.otp_devices_list);
+          }}
+          onBack={() => {
+            route(InstancePaths.otp_devices_list);
+          }}
+        />
+        {/**
+         * Templates pages
+         */}
+        <Route
+          path={InstancePaths.templates_list}
+          component={TemplateListPage}
+          onUnauthorized={LoginPageAccessDenied}
+          onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.settings)}
+          onCreate={() => {
+            route(InstancePaths.templates_new);
+          }}
+          onNewOrder={(id: string) => {
+            route(InstancePaths.templates_use.replace(":tid", id));
+          }}
+          onQR={(id: string) => {
+            route(InstancePaths.templates_qr.replace(":tid", id));
+          }}
+          onSelect={(id: string) => {
+            route(InstancePaths.templates_update.replace(":tid", id));
+          }}
+        />
+        <Route
+          path={InstancePaths.templates_update}
+          component={TemplateUpdatePage}
+          onConfirm={() => {
+            route(InstancePaths.templates_list);
+          }}
+          onUnauthorized={LoginPageAccessDenied}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.templates_list)}
+          onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
+          onBack={() => {
+            route(InstancePaths.templates_list);
+          }}
+        />
+        <Route
+          path={InstancePaths.templates_new}
+          component={TemplateCreatePage}
+          onConfirm={() => {
+            route(InstancePaths.templates_list);
+          }}
+          onBack={() => {
+            route(InstancePaths.templates_list);
+          }}
+        />
+        <Route
+          path={InstancePaths.templates_use}
+          component={TemplateUsePage}
+          onOrderCreated={(id: string) => {
+            route(InstancePaths.order_details.replace(":oid", id));
+          }}
+          onUnauthorized={LoginPageAccessDenied}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.templates_list)}
+          onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
+          onBack={() => {
+            route(InstancePaths.templates_list);
+          }}
+        />
+        <Route
+          path={InstancePaths.templates_qr}
+          component={TemplateQrPage}
+          onUnauthorized={LoginPageAccessDenied}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.templates_list)}
+          onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
+          onBack={() => {
+            route(InstancePaths.templates_list);
+          }}
+        />
 
         {/**
          * reserves pages
@@ -390,7 +641,7 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
           component={ReservesListPage}
           onUnauthorized={LoginPageAccessDenied}
           onNotFound={IfAdminCreateDefaultOr(NotFoundPage)}
-          onLoadError={ServerErrorRedirectTo(InstancePaths.update)}
+          onLoadError={ServerErrorRedirectTo(InstancePaths.settings)}
           onSelect={(id: string) => {
             route(InstancePaths.reserves_details.replace(":rid", id));
           }}
@@ -398,7 +649,6 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
             route(InstancePaths.reserves_new);
           }}
         />
-
         <Route
           path={InstancePaths.reserves_details}
           component={ReservesDetailsPage}
@@ -409,7 +659,6 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
             route(InstancePaths.reserves_list);
           }}
         />
-
         <Route
           path={InstancePaths.reserves_new}
           component={ReservesCreatePage}
@@ -420,8 +669,8 @@ export function InstanceRoutes({ id, admin, setInstanceName }: Props): VNode {
             route(InstancePaths.reserves_list);
           }}
         />
-
         <Route path={InstancePaths.kyc} component={ListKYCPage} />
+        <Route path={InstancePaths.interface} component={Settings} />
         {/**
          * Example pages
          */}
@@ -442,38 +691,43 @@ export function Redirect({ to }: { to: string }): null {
 function AdminInstanceUpdatePage({
   id,
   ...rest
-}: { id: string } & InstanceUpdatePageProps) {
+}: { id: string } & InstanceUpdatePageProps): VNode {
   const [token, changeToken] = useBackendInstanceToken(id);
-  const { updateLoginStatus: changeBackend } = useBackendContext();
-  const updateLoginStatus = (url: string, token?: string) => {
-    changeBackend(url);
-    if (token) changeToken(token);
+  const updateLoginStatus = (token?: LoginToken): void => {
+    changeToken(token);
   };
   const value = useMemo(
     () => ({ id, token, admin: true, changeToken }),
-    [id, token]
+    [id, token],
   );
-  const i18n = useTranslator();
+  const { i18n } = useTranslationContext();
 
   return (
     <InstanceContextProvider value={value}>
       <InstanceAdminUpdatePage
         {...rest}
         instanceId={id}
-        onLoadError={(error: HttpError) => {
+        onLoadError={(error: HttpError<MerchantBackend.ErrorDetail>) => {
+          const notif =
+            error.type === ErrorType.TIMEOUT
+              ? {
+                message: i18n.str`The request to the backend take too long and was cancelled`,
+                description: i18n.str`Diagnostic from ${error.info.url} is '${error.message}'`,
+                type: "ERROR" as const,
+              }
+              : {
+                message: i18n.str`The backend reported a problem: HTTP status #${error.status}`,
+                description: i18n.str`Diagnostic from ${error.info.url} is '${error.message}'`,
+                details:
+                  error.type === ErrorType.CLIENT ||
+                    error.type === ErrorType.SERVER
+                    ? error.payload.detail
+                    : undefined,
+                type: "ERROR" as const,
+              };
           return (
             <Fragment>
-              <NotificationCard
-                notification={{
-                  message: i18n`The backend reported a problem: HTTP status #${error.status}`,
-                  description: i18n`Diagnostic from ${error.info?.url} is "${error.message}"`,
-                  details:
-                    error.clientError || error.serverError
-                      ? error.error?.detail
-                      : undefined,
-                  type: "ERROR",
-                }}
-              />
+              <NotificationCard notification={notif} />
               <LoginPage onConfirm={updateLoginStatus} />
             </Fragment>
           );
@@ -483,8 +737,8 @@ function AdminInstanceUpdatePage({
             <Fragment>
               <NotificationCard
                 notification={{
-                  message: i18n`Access denied`,
-                  description: i18n`The access token provided is invalid`,
+                  message: i18n.str`Access denied`,
+                  description: i18n.str`The access token provided is invalid`,
                   type: "ERROR",
                 }}
               />
@@ -499,8 +753,10 @@ function AdminInstanceUpdatePage({
 
 function KycBanner(): VNode {
   const kycStatus = useInstanceKYCDetails();
-  const today = format(new Date(), "yyyy-MM-dd");
-  const [lastHide, setLastHide] = useLocalStorage("kyc-last-hide");
+  const { i18n } = useTranslationContext();
+  const [settings] = useSettings();
+  const today = format(new Date(), dateFormatForSettings(settings));
+  const [lastHide, setLastHide] = useSimpleLocalStorage("kyc-last-hide");
   const hasBeenHidden = today === lastHide;
   const needsToBeShown = kycStatus.ok && kycStatus.data.type === "redirect";
   if (hasBeenHidden || !needsToBeShown) return <Fragment />;
@@ -517,7 +773,7 @@ function KycBanner(): VNode {
             </p>
             <div class="buttons is-right">
               <button class="button" onClick={() => setLastHide(today)}>
-                <Translate>Hide for today</Translate>
+                <i18n.Translate>Hide for today</i18n.Translate>
               </button>
             </div>
           </div>
