@@ -1,6 +1,6 @@
 /*
  This file is part of GNU Taler
- (C) 2021 Taler Systems S.A.
+ (C) 2021-2023 Taler Systems S.A.
 
  GNU Taler is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
@@ -15,11 +15,11 @@
  */
 
 /**
-*
-* @author Sebastian Javier Marchano (sebasjm)
-*/
+ *
+ * @author Sebastian Javier Marchano (sebasjm)
+ */
 
-import { useFormContext } from "./FormProvider";
+import { useFormContext } from "./FormProvider.js";
 
 interface Use {
   hasError?: boolean;
@@ -27,14 +27,15 @@ interface Use {
 
 export function useGroupField<T>(name: keyof T): Use {
   const f = useFormContext<T>();
-  if (!f)
-    return {};
+  if (!f) return {};
 
   return {
-    hasError: readField(f.errors, String(name))
+    hasError: readField(f.errors, String(name)),
   };
 }
 
 const readField = (object: any, name: string) => {
-  return name.split('.').reduce((prev, current) => prev && prev[current], object)
-}
+  return name
+    .split(".")
+    .reduce((prev, current) => prev && prev[current], object);
+};
