@@ -1,6 +1,6 @@
 /*
  This file is part of GNU Taler
- (C) 2021 Taler Systems S.A.
+ (C) 2021-2023 Taler Systems S.A.
 
  GNU Taler is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
@@ -19,44 +19,43 @@
  * @author Sebastian Javier Marchano (sebasjm)
  */
 
+import { useTranslationContext } from "@gnu-taler/web-util/browser";
 import { Fragment, h, VNode } from "preact";
-import { useCallback } from "preact/hooks";
-import { useBackendContext } from "../../context/backend";
-import { useConfigContext } from "../../context/config";
-import { useInstanceContext } from "../../context/instance";
-import { useInstanceKYCDetails } from "../../hooks/instance";
-import { Translate } from "../../i18n";
-import { LangSelector } from "./LangSelector";
+import { useBackendContext } from "../../context/backend.js";
+import { useConfigContext } from "../../context/config.js";
+import { useInstanceKYCDetails } from "../../hooks/instance.js";
+import { LangSelector } from "./LangSelector.js";
+
+const GIT_HASH = typeof __GIT_HASH__ !== "undefined" ? __GIT_HASH__ : undefined;
+const VERSION = typeof __VERSION__ !== "undefined" ? __VERSION__ : undefined;
 
 interface Props {
   onLogout: () => void;
+  onShowSettings: () => void;
   mobile?: boolean;
   instance: string;
   admin?: boolean;
   mimic?: boolean;
+  isPasswordOk: boolean;
 }
 
 export function Sidebar({
   mobile,
   instance,
+  onShowSettings,
   onLogout,
   admin,
   mimic,
+  isPasswordOk
 }: Props): VNode {
   const config = useConfigContext();
-  const backend = useBackendContext();
-
+  const { url: backendURL } = useBackendContext()
+  const { i18n } = useTranslationContext();
   const kycStatus = useInstanceKYCDetails();
   const needKYC = kycStatus.ok && kycStatus.data.type === "redirect";
-  // const withInstanceIdIfNeeded = useCallback(function (path: string) {
-  //   if (mimic) {
-  //     return path + '?instance=' + instance
-  //   }
-  //   return path
-  // },[instance])
 
   return (
-    <aside class="aside is-placed-left is-expanded">
+    <aside class="aside is-placed-left is-expanded" style={{ overflowY: "scroll" }}>
       {mobile && (
         <div
           class="footer"
@@ -76,63 +75,52 @@ export function Sidebar({
             class="is-size-7 has-text-right"
             style={{ lineHeight: 0, marginTop: -10 }}
           >
-            {process.env.__VERSION__} ({config.version})
+            {VERSION} ({config.version})
           </div>
         </div>
       </div>
       <div class="menu is-menu-main">
         {instance ? (
           <Fragment>
-            <p class="menu-label">
-              <Translate>Instance</Translate>
-            </p>
             <ul class="menu-list">
-              <li>
-                <a href={"/update"} class="has-icon">
-                  <span class="icon">
-                    <i class="mdi mdi-square-edit-outline" />
-                  </span>
-                  <span class="menu-item-label">
-                    <Translate>Settings</Translate>
-                  </span>
-                </a>
-              </li>
               <li>
                 <a href={"/orders"} class="has-icon">
                   <span class="icon">
                     <i class="mdi mdi-cash-register" />
                   </span>
                   <span class="menu-item-label">
-                    <Translate>Orders</Translate>
+                    <i18n.Translate>Orders</i18n.Translate>
                   </span>
                 </a>
               </li>
               <li>
-                <a href={"/products"} class="has-icon">
+                <a href={"/inventory"} class="has-icon">
                   <span class="icon">
                     <i class="mdi mdi-shopping" />
                   </span>
                   <span class="menu-item-label">
-                    <Translate>Products</Translate>
+                    <i18n.Translate>Inventory</i18n.Translate>
                   </span>
                 </a>
               </li>
               <li>
                 <a href={"/transfers"} class="has-icon">
                   <span class="icon">
-                    <i class="mdi mdi-bank" />
+                    <i class="mdi mdi-arrow-left-right" />
                   </span>
                   <span class="menu-item-label">
-                    <Translate>Transfers</Translate>
+                    <i18n.Translate>Transfers</i18n.Translate>
                   </span>
                 </a>
               </li>
               <li>
-                <a href={"/reserves"} class="has-icon">
+                <a href={"/templates"} class="has-icon">
                   <span class="icon">
-                    <i class="mdi mdi-cash" />
+                    <i class="mdi mdi-newspaper" />
+                  </span>
+                  <span class="menu-item-label">
+                    <i18n.Translate>Templates</i18n.Translate>
                   </span>
-                  <span class="menu-item-label">Reserves</span>
                 </a>
               </li>
               {needKYC && (
@@ -146,12 +134,87 @@ export function Sidebar({
                 </li>
               )}
             </ul>
+            <p class="menu-label">
+              <i18n.Translate>Configuration</i18n.Translate>
+            </p>
+            <ul class="menu-list">
+              <li>
+                <a href={"/bank"} class="has-icon">
+                  <span class="icon">
+                    <i class="mdi mdi-bank" />
+                  </span>
+                  <span class="menu-item-label">
+                    <i18n.Translate>Bank account</i18n.Translate>
+                  </span>
+                </a>
+              </li>
+              <li>
+                <a href={"/otp-devices"} class="has-icon">
+                  <span class="icon">
+                    <i class="mdi mdi-lock" />
+                  </span>
+                  <span class="menu-item-label">
+                    <i18n.Translate>OTP Devices</i18n.Translate>
+                  </span>
+                </a>
+              </li>
+              <li>
+                <a href={"/reserves"} class="has-icon">
+                  <span class="icon">
+                    <i class="mdi mdi-cash" />
+                  </span>
+                  <span class="menu-item-label">Reserves</span>
+                </a>
+              </li>
+              <li>
+                <a href={"/webhooks"} class="has-icon">
+                  <span class="icon">
+                    <i class="mdi mdi-newspaper" />
+                  </span>
+                  <span class="menu-item-label">
+                    <i18n.Translate>Webhooks</i18n.Translate>
+                  </span>
+                </a>
+              </li>
+              <li>
+                <a href={"/settings"} class="has-icon">
+                  <span class="icon">
+                    <i class="mdi mdi-square-edit-outline" />
+                  </span>
+                  <span class="menu-item-label">
+                    <i18n.Translate>Settings</i18n.Translate>
+                  </span>
+                </a>
+              </li>
+              <li>
+                <a href={"/token"} class="has-icon">
+                  <span class="icon">
+                    <i class="mdi mdi-security" />
+                  </span>
+                  <span class="menu-item-label">
+                    <i18n.Translate>Access token</i18n.Translate>
+                  </span>
+                </a>
+              </li>
+            </ul>
           </Fragment>
         ) : undefined}
         <p class="menu-label">
-          <Translate>Connection</Translate>
+          <i18n.Translate>Connection</i18n.Translate>
         </p>
         <ul class="menu-list">
+          <li>
+            <a class="has-icon is-state-info is-hoverable"
+              onClick={(): void => onShowSettings()}
+            >
+              <span class="icon">
+                <i class="mdi mdi-newspaper" />
+              </span>
+              <span class="menu-item-label">
+                <i18n.Translate>Interface</i18n.Translate>
+              </span>
+            </a>
+          </li>
           <li>
             <div>
               <span style={{ width: "3rem" }} class="icon">
@@ -166,7 +229,7 @@ export function Sidebar({
                 <i class="mdi mdi-web" />
               </span>
               <span class="menu-item-label">
-                {new URL(backend.url).hostname}
+                {new URL(backendURL).hostname}
               </span>
             </div>
           </li>
@@ -183,7 +246,7 @@ export function Sidebar({
           {admin && !mimic && (
             <Fragment>
               <p class="menu-label">
-                <Translate>Instances</Translate>
+                <i18n.Translate>Instances</i18n.Translate>
               </p>
               <li>
                 <a href={"/instance/new"} class="has-icon">
@@ -191,7 +254,7 @@ export function Sidebar({
                     <i class="mdi mdi-plus" />
                   </span>
                   <span class="menu-item-label">
-                    <Translate>New</Translate>
+                    <i18n.Translate>New</i18n.Translate>
                   </span>
                 </a>
               </li>
@@ -201,25 +264,27 @@ export function Sidebar({
                     <i class="mdi mdi-format-list-bulleted" />
                   </span>
                   <span class="menu-item-label">
-                    <Translate>List</Translate>
+                    <i18n.Translate>List</i18n.Translate>
                   </span>
                 </a>
               </li>
             </Fragment>
           )}
-          <li>
-            <a
-              class="has-icon is-state-info is-hoverable"
-              onClick={(): void => onLogout()}
-            >
-              <span class="icon">
-                <i class="mdi mdi-logout default" />
-              </span>
-              <span class="menu-item-label">
-                <Translate>Log out</Translate>
-              </span>
-            </a>
-          </li>
+          {isPasswordOk ?
+            <li>
+              <a
+                class="has-icon is-state-info is-hoverable"
+                onClick={(): void => onLogout()}
+              >
+                <span class="icon">
+                  <i class="mdi mdi-logout default" />
+                </span>
+                <span class="menu-item-label">
+                  <i18n.Translate>Log out</i18n.Translate>
+                </span>
+              </a>
+            </li> : undefined
+          }
         </ul>
       </div>
     </aside>
