@@ -1,6 +1,6 @@
 /*
  This file is part of GNU Taler
- (C) 2021 Taler Systems S.A.
+ (C) 2021-2023 Taler Systems S.A.
 
  GNU Taler is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
@@ -15,21 +15,22 @@
  */
 
 /**
-*
-* @author Sebastian Javier Marchano (sebasjm)
-*/
+ *
+ * @author Sebastian Javier Marchano (sebasjm)
+ */
 
-import { createContext } from 'preact'
-import { useContext } from 'preact/hooks'
+import { createContext } from "preact";
+import { useContext } from "preact/hooks";
+import { LoginToken } from "../declaration.js";
 
 interface Type {
   id: string;
-  token?: string;
+  token?: LoginToken;
   admin?: boolean;
-  changeToken: (t?:string) => void;
+  changeToken: (t?: LoginToken) => void;
 }
 
-const Context = createContext<Type>({} as any)
+const Context = createContext<Type>({} as any);
 
-export const InstanceContextProvider = Context.Provider
+export const InstanceContextProvider = Context.Provider;
 export const useInstanceContext = (): Type => useContext(Context);
