@@ -21,17 +21,12 @@
 /**
  * Imports.
  */
+import { codecForAmountString } from "./amounts.js";
 import {
-  codecForString,
-  buildCodecForObject,
-  codecForList,
   Codec,
+  buildCodecForObject
 } from "./codec.js";
 import { AmountString } from "./taler-types.js";
-import {
-  ReserveTransaction,
-  codecForReserveTransaction,
-} from "./ReserveTransaction.js";
 
 /**
  * Status of a reserve.
@@ -47,5 +42,5 @@ export interface ReserveStatus {
 
 export const codecForReserveStatus = (): Codec<ReserveStatus> =>
   buildCodecForObject<ReserveStatus>()
-    .property("balance", codecForString())
+    .property("balance", codecForAmountString())
     .build("ReserveStatus");
