@@ -25,7 +25,7 @@
  * Imports.
  */
 
-import { codecForAmountString } from "./amounts.js";
+import { Amounts, codecForAmountString } from "./amounts.js";
 import {
   buildCodecForObject,
   buildCodecForUnion,
@@ -41,6 +41,10 @@ import {
   codecOptional,
 } from "./codec.js";
 import { strcmp } from "./helpers.js";
+import {
+  CurrencySpecification,
+  codecForCurrencySpecificiation,
+} from "./index.js";
 import { AgeCommitmentProof, Edx25519PublicKeyEnc } from "./taler-crypto.js";
 import {
   codecForAbsoluteTime,
@@ -448,11 +452,6 @@ export interface MerchantContractTerms {
    */
   amount: string;
 
-  /**
-   * Auditors accepted by the merchant.
-   */
-  auditors: AuditorHandle[];
-
   /**
    * Deadline to pay for the contract.
    */
@@ -725,9 +724,11 @@ export class ExchangeSignKeyJson {
  */
 export class ExchangeKeysJson {
   /**
-   * List of offered denominations.
+   * Canonical, public base URL of the exchange.
    */
-  denoms: ExchangeDenomination[];
+  base_url: string;
+
+  currency: string;
 
   /**
    * The exchange's master public key.
@@ -763,6 +764,111 @@ export class ExchangeKeysJson {
   reserve_closing_delay: TalerProtocolDuration;
 
   global_fees: GlobalFees[];
+
+  accounts: ExchangeWireAccount[];
+
+  wire_fees: { [methodName: string]: WireFeesJson[] };
+
+  denominations: DenomGroup[];
+}
+
+export type DenomGroup =
+  | DenomGroupRsa
+  | DenomGroupCs
+  | DenomGroupRsaAgeRestricted
+  | DenomGroupCsAgeRestricted;
+
+export interface DenomGroupCommon {
+  // How much are coins of this denomination worth?
+  value: AmountString;
+
+  // Fee charged by the exchange for withdrawing a coin of this denomination.
+  fee_withdraw: AmountString;
+
+  // Fee charged by the exchange for depositing a coin of this denomination.
+  fee_deposit: AmountString;
+
+  // Fee charged by the exchange for refreshing a coin of this denomination.
+  fee_refresh: AmountString;
+
+  // Fee charged by the exchange for refunding a coin of this denomination.
+  fee_refund: AmountString;
+
+  // XOR of all the SHA-512 hash values of the denominations' public keys
+  // in this group.  Note that for hashing, the binary format of the
+  // public keys is used, and not their base32 encoding.
+  hash: HashCodeString;
+}
+
+export interface DenomCommon {
+  // Signature of TALER_DenominationKeyValidityPS.
+  master_sig: EddsaSignatureString;
+
+  // When does the denomination key become valid?
+  stamp_start: TalerProtocolTimestamp;
+
+  // When is it no longer possible to deposit coins
+  // of this denomination?
+  stamp_expire_withdraw: TalerProtocolTimestamp;
+
+  // Timestamp indicating by when legal disputes relating to these coins must
+  // be settled, as the exchange will afterwards destroy its evidence relating to
+  // transactions involving this coin.
+  stamp_expire_legal: TalerProtocolTimestamp;
+
+  stamp_expire_deposit: TalerProtocolTimestamp;
+
+  // Set to 'true' if the exchange somehow "lost"
+  // the private key. The denomination was not
+  // necessarily revoked, but still cannot be used
+  // to withdraw coins at this time (theoretically,
+  // the private key could be recovered in the
+  // future; coins signed with the private key
+  // remain valid).
+  lost?: boolean;
+}
+
+export type RsaPublicKeySring = string;
+export type AgeMask = number;
+
+/**
+ * 32-byte value representing a point on Curve25519.
+ */
+export type Cs25519Point = string;
+
+export interface DenomGroupRsa extends DenomGroupCommon {
+  cipher: "RSA";
+
+  denoms: ({
+    rsa_pub: RsaPublicKeySring;
+  } & DenomCommon)[];
+}
+
+export interface DenomGroupRsaAgeRestricted extends DenomGroupCommon {
+  cipher: "RSA+age_restricted";
+  age_mask: AgeMask;
+
+  denoms: ({
+    rsa_pub: RsaPublicKeySring;
+  } & DenomCommon)[];
+}
+
+export interface DenomGroupCs extends DenomGroupCommon {
+  cipher: "CS";
+  age_mask: AgeMask;
+
+  denoms: ({
+    cs_pub: Cs25519Point;
+  } & DenomCommon)[];
+}
+
+export interface DenomGroupCsAgeRestricted extends DenomGroupCommon {
+  cipher: "CS+age_restricted";
+  age_mask: AgeMask;
+
+  denoms: ({
+    cs_pub: Cs25519Point;
+  } & DenomCommon)[];
 }
 
 export interface GlobalFees {
@@ -837,16 +943,6 @@ export class WireFeesJson {
   end_date: TalerProtocolTimestamp;
 }
 
-export interface AccountInfo {
-  payto_uri: string;
-  master_sig: string;
-}
-
-export interface ExchangeWireJson {
-  accounts: AccountInfo[];
-  fees: { [methodName: string]: WireFeesJson[] };
-}
-
 /**
  * Proposal returned from the contract URL.
  */
@@ -900,11 +996,13 @@ export class WithdrawOperationStatusResponse {
 /**
  * Response from the merchant.
  */
-export class TipPickupGetResponse {
-  tip_amount: string;
+export class RewardPickupGetResponse {
+  reward_amount: string;
 
   exchange_url: string;
 
+  next_url?: string;
+
   expiration: TalerProtocolTimestamp;
 }
 
@@ -949,16 +1047,17 @@ export const codecForBlindedDenominationSignature = () =>
     .alternative(DenomKeyType.Rsa, codecForRsaBlindedDenominationSignature())
     .build("BlindedDenominationSignature");
 
-export class WithdrawResponse {
+export class ExchangeWithdrawResponse {
   ev_sig: BlindedDenominationSignature;
 }
 
-export class WithdrawBatchResponse {
-  ev_sigs: WithdrawResponse[];
+export class ExchangeWithdrawBatchResponse {
+  ev_sigs: ExchangeWithdrawResponse[];
 }
 
 export interface MerchantPayResponse {
   sig: string;
+  pos_confirmation?: string;
 }
 
 export interface ExchangeMeltRequest {
@@ -1201,13 +1300,9 @@ export const codecForDenominationPubKey = () =>
     .alternative(DenomKeyType.ClauseSchnorr, codecForCsDenominationPubKey())
     .build("DenominationPubKey");
 
-export const codecForBankWithdrawalOperationPostResponse =
-  (): Codec<BankWithdrawalOperationPostResponse> =>
-    buildCodecForObject<BankWithdrawalOperationPostResponse>()
-      .property("transfer_done", codecForBoolean())
-      .build("BankWithdrawalOperationPostResponse");
-
-export type AmountString = string;
+declare const __amount_str: unique symbol;
+export type AmountString = string & { [__amount_str]: true };
+// export type AmountString = string;
 export type Base32String = string;
 export type EddsaSignatureString = string;
 export type EddsaPublicKeyString = string;
@@ -1278,7 +1373,7 @@ export const codecForMerchantInfo = (): Codec<MerchantInfo> =>
 export const codecForTax = (): Codec<Tax> =>
   buildCodecForObject<Tax>()
     .property("name", codecForString())
-    .property("tax", codecForString())
+    .property("tax", codecForAmountString())
     .build("Tax");
 
 export const codecForInternationalizedString =
@@ -1294,10 +1389,10 @@ export const codecForProduct = (): Codec<Product> =>
     )
     .property("quantity", codecOptional(codecForNumber()))
     .property("unit", codecOptional(codecForString()))
-    .property("price", codecOptional(codecForString()))
+    .property("price", codecOptional(codecForAmountString()))
     .build("Tax");
 
-export const codecForContractTerms = (): Codec<MerchantContractTerms> =>
+export const codecForMerchantContractTerms = (): Codec<MerchantContractTerms> =>
   buildCodecForObject<MerchantContractTerms>()
     .property("order_id", codecForString())
     .property("fulfillment_url", codecOptional(codecForString()))
@@ -1314,7 +1409,6 @@ export const codecForContractTerms = (): Codec<MerchantContractTerms> =>
     .property("summary_i18n", codecOptional(codecForInternationalizedString()))
     .property("nonce", codecForString())
     .property("amount", codecForString())
-    .property("auditors", codecForList(codecForAuditorHandle()))
     .property("pay_deadline", codecForTimestamp)
     .property("refund_deadline", codecForTimestamp)
     .property("wire_transfer_deadline", codecForTimestamp)
@@ -1329,7 +1423,14 @@ export const codecForContractTerms = (): Codec<MerchantContractTerms> =>
     .property("products", codecOptional(codecForList(codecForProduct())))
     .property("extra", codecForAny())
     .property("minimum_age", codecOptional(codecForNumber()))
-    .build("ContractTerms");
+    .build("MerchantContractTerms");
+
+export const codecForPeerContractTerms = (): Codec<PeerContractTerms> =>
+  buildCodecForObject<PeerContractTerms>()
+    .property("summary", codecForString())
+    .property("amount", codecForAmountString())
+    .property("purse_expiration", codecForTimestamp)
+    .build("PeerContractTerms");
 
 export const codecForMerchantRefundPermission =
   (): Codec<MerchantAbortPayRefundDetails> =>
@@ -1390,9 +1491,13 @@ export const codecForGlobalFees = (): Codec<GlobalFees> =>
     .property("master_sig", codecForString())
     .build("GlobalFees");
 
+// FIXME: Validate properly!
+export const codecForNgDenominations: Codec<DenomGroup> = codecForAny();
+
 export const codecForExchangeKeysJson = (): Codec<ExchangeKeysJson> =>
   buildCodecForObject<ExchangeKeysJson>()
-    .property("denoms", codecForList(codecForDenomination()))
+    .property("base_url", codecForString())
+    .property("currency", codecForString())
     .property("master_public_key", codecForString())
     .property("auditors", codecForList(codecForAuditor()))
     .property("list_issue_date", codecForTimestamp)
@@ -1401,6 +1506,9 @@ export const codecForExchangeKeysJson = (): Codec<ExchangeKeysJson> =>
     .property("version", codecForString())
     .property("reserve_closing_delay", codecForDuration)
     .property("global_fees", codecForList(codecForGlobalFees()))
+    .property("accounts", codecForList(codecForExchangeWireAccount()))
+    .property("wire_fees", codecForMap(codecForList(codecForWireFeesJson())))
+    .property("denominations", codecForList(codecForNgDenominations))
     .build("ExchangeKeysJson");
 
 export const codecForWireFeesJson = (): Codec<WireFeesJson> =>
@@ -1412,18 +1520,6 @@ export const codecForWireFeesJson = (): Codec<WireFeesJson> =>
     .property("end_date", codecForTimestamp)
     .build("WireFeesJson");
 
-export const codecForAccountInfo = (): Codec<AccountInfo> =>
-  buildCodecForObject<AccountInfo>()
-    .property("payto_uri", codecForString())
-    .property("master_sig", codecForString())
-    .build("AccountInfo");
-
-export const codecForExchangeWireJson = (): Codec<ExchangeWireJson> =>
-  buildCodecForObject<ExchangeWireJson>()
-    .property("accounts", codecForList(codecForAccountInfo()))
-    .property("fees", codecForMap(codecForList(codecForWireFeesJson())))
-    .build("ExchangeWireJson");
-
 export const codecForProposal = (): Codec<Proposal> =>
   buildCodecForObject<Proposal>()
     .property("contract_terms", codecForAny())
@@ -1453,12 +1549,14 @@ export const codecForWithdrawOperationStatusResponse =
       .property("wire_types", codecForList(codecForString()))
       .build("WithdrawOperationStatusResponse");
 
-export const codecForTipPickupGetResponse = (): Codec<TipPickupGetResponse> =>
-  buildCodecForObject<TipPickupGetResponse>()
-    .property("tip_amount", codecForString())
-    .property("exchange_url", codecForString())
-    .property("expiration", codecForTimestamp)
-    .build("TipPickupGetResponse");
+export const codecForRewardPickupGetResponse =
+  (): Codec<RewardPickupGetResponse> =>
+    buildCodecForObject<RewardPickupGetResponse>()
+      .property("reward_amount", codecForString())
+      .property("exchange_url", codecForString())
+      .property("next_url", codecOptional(codecForString()))
+      .property("expiration", codecForTimestamp)
+      .build("TipPickupGetResponse");
 
 export const codecForRecoupConfirmation = (): Codec<RecoupConfirmation> =>
   buildCodecForObject<RecoupConfirmation>()
@@ -1466,19 +1564,21 @@ export const codecForRecoupConfirmation = (): Codec<RecoupConfirmation> =>
     .property("old_coin_pub", codecOptional(codecForString()))
     .build("RecoupConfirmation");
 
-export const codecForWithdrawResponse = (): Codec<WithdrawResponse> =>
-  buildCodecForObject<WithdrawResponse>()
+export const codecForWithdrawResponse = (): Codec<ExchangeWithdrawResponse> =>
+  buildCodecForObject<ExchangeWithdrawResponse>()
     .property("ev_sig", codecForBlindedDenominationSignature())
     .build("WithdrawResponse");
 
-export const codecForWithdrawBatchResponse = (): Codec<WithdrawBatchResponse> =>
-  buildCodecForObject<WithdrawBatchResponse>()
-    .property("ev_sigs", codecForList(codecForWithdrawResponse()))
-    .build("WithdrawBatchResponse");
+export const codecForExchangeWithdrawBatchResponse =
+  (): Codec<ExchangeWithdrawBatchResponse> =>
+    buildCodecForObject<ExchangeWithdrawBatchResponse>()
+      .property("ev_sigs", codecForList(codecForWithdrawResponse()))
+      .build("WithdrawBatchResponse");
 
 export const codecForMerchantPayResponse = (): Codec<MerchantPayResponse> =>
   buildCodecForObject<MerchantPayResponse>()
     .property("sig", codecForString())
+    .property("pos_confirmation", codecOptional(codecForString()))
     .build("MerchantPayResponse");
 
 export const codecForExchangeMeltResponse = (): Codec<ExchangeMeltResponse> =>
@@ -1508,7 +1608,7 @@ export const codecForMerchantCoinRefundSuccessStatus =
       .property("exchange_status", codecForConstNumber(200))
       .property("exchange_sig", codecForString())
       .property("rtransaction_id", codecForNumber())
-      .property("refund_amount", codecForString())
+      .property("refund_amount", codecForAmountString())
       .property("exchange_pub", codecForString())
       .property("execution_time", codecForTimestamp)
       .build("MerchantCoinRefundSuccessStatus");
@@ -1520,7 +1620,7 @@ export const codecForMerchantCoinRefundFailureStatus =
       .property("coin_pub", codecForString())
       .property("exchange_status", codecForNumber())
       .property("rtransaction_id", codecForNumber())
-      .property("refund_amount", codecForString())
+      .property("refund_amount", codecForAmountString())
       .property("exchange_code", codecOptional(codecForNumber()))
       .property("exchange_reply", codecOptional(codecForAny()))
       .property("execution_time", codecForTimestamp)
@@ -1537,8 +1637,8 @@ export const codecForMerchantCoinRefundStatus =
 export const codecForMerchantOrderStatusPaid =
   (): Codec<MerchantOrderStatusPaid> =>
     buildCodecForObject<MerchantOrderStatusPaid>()
-      .property("refund_amount", codecForString())
-      .property("refund_taken", codecForString())
+      .property("refund_amount", codecForAmountString())
+      .property("refund_taken", codecForAmountString())
       .property("refund_pending", codecForBoolean())
       .property("refunded", codecForBoolean())
       .build("MerchantOrderStatusPaid");
@@ -1547,7 +1647,7 @@ export const codecForMerchantOrderRefundPickupResponse =
   (): Codec<MerchantOrderRefundResponse> =>
     buildCodecForObject<MerchantOrderRefundResponse>()
       .property("merchant_pub", codecForString())
-      .property("refund_amount", codecForString())
+      .property("refund_amount", codecForAmountString())
       .property("refunds", codecForList(codecForMerchantCoinRefundStatus()))
       .build("MerchantOrderRefundPickupResponse");
 
@@ -1660,19 +1760,6 @@ export interface MerchantAbortPayRefundSuccessStatus {
   exchange_pub: string;
 }
 
-export interface TalerConfigResponse {
-  name: string;
-  version: string;
-  currency?: string;
-}
-
-export const codecForTalerConfigResponse = (): Codec<TalerConfigResponse> =>
-  buildCodecForObject<TalerConfigResponse>()
-    .property("name", codecForString())
-    .property("version", codecForString())
-    .property("currency", codecOptional(codecForString()))
-    .build("TalerConfigResponse");
-
 export interface FutureKeysResponse {
   future_denoms: any[];
 
@@ -1743,6 +1830,10 @@ export interface ExchangeWithdrawRequest {
   coin_ev: CoinEnvelope;
 }
 
+export interface ExchangeBatchWithdrawRequest {
+  planchets: ExchangeWithdrawRequest[];
+}
+
 export interface ExchangeRefreshRevealRequest {
   new_denoms_h: HashCodeString[];
   coin_evs: CoinEnvelope[];
@@ -1763,42 +1854,112 @@ export interface ExchangeRefreshRevealRequest {
   old_age_commitment?: Edx25519PublicKeyEnc[];
 }
 
-export interface DepositSuccess {
+interface DepositConfirmationSignature {
+  // The EdDSA signature of `TALER_DepositConfirmationPS` using a current
+  // `signing key of the exchange <sign-key-priv>` affirming the successful
+  // deposit and that the exchange will transfer the funds after the refund
+  // deadline, or as soon as possible if the refund deadline is zero.
+  exchange_sig: EddsaSignatureString;
+}
+
+export interface BatchDepositSuccess {
   // Optional base URL of the exchange for looking up wire transfers
   // associated with this transaction.  If not given,
   // the base URL is the same as the one used for this request.
-  // Can be used if the base URL for /transactions/ differs from that
-  // for /coins/, i.e. for load balancing.  Clients SHOULD
-  // respect the transaction_base_url if provided.  Any HTTP server
+  // Can be used if the base URL for ``/transactions/`` differs from that
+  // for ``/coins/``, i.e. for load balancing.  Clients SHOULD
+  // respect the ``transaction_base_url`` if provided.  Any HTTP server
   // belonging to an exchange MUST generate a 307 or 308 redirection
   // to the correct base URL should a client uses the wrong base
   // URL, or if the base URL has changed since the deposit.
   transaction_base_url?: string;
 
-  // timestamp when the deposit was received by the exchange.
+  // Timestamp when the deposit was received by the exchange.
   exchange_timestamp: TalerProtocolTimestamp;
 
-  // the EdDSA signature of TALER_DepositConfirmationPS using a current
-  // signing key of the exchange affirming the successful
-  // deposit and that the exchange will transfer the funds after the refund
-  // deadline, or as soon as possible if the refund deadline is zero.
-  exchange_sig: string;
-
-  // public EdDSA key of the exchange that was used to
+  // `Public EdDSA key of the exchange <sign-key-pub>` that was used to
   // generate the signature.
-  // Should match one of the exchange's signing keys from /keys.  It is given
+  // Should match one of the exchange's signing keys from ``/keys``.  It is given
   // explicitly as the client might otherwise be confused by clock skew as to
   // which signing key was used.
-  exchange_pub: string;
+  exchange_pub: EddsaPublicKeyString;
+
+  // Array of deposit confirmation signatures from the exchange
+  // Entries must be in the same order the coins were given
+  // in the batch deposit request.
+  exchange_sig: EddsaSignatureString;
 }
 
-export const codecForDepositSuccess = (): Codec<DepositSuccess> =>
-  buildCodecForObject<DepositSuccess>()
+export const codecForBatchDepositSuccess = (): Codec<BatchDepositSuccess> =>
+  buildCodecForObject<BatchDepositSuccess>()
     .property("exchange_pub", codecForString())
     .property("exchange_sig", codecForString())
     .property("exchange_timestamp", codecForTimestamp)
     .property("transaction_base_url", codecOptional(codecForString()))
-    .build("DepositSuccess");
+    .build("BatchDepositSuccess");
+
+export interface TrackTransactionWired {
+  // Raw wire transfer identifier of the deposit.
+  wtid: Base32String;
+
+  // When was the wire transfer given to the bank.
+  execution_time: TalerProtocolTimestamp;
+
+  // The contribution of this coin to the total (without fees)
+  coin_contribution: AmountString;
+
+  // Binary-only Signature_ with purpose TALER_SIGNATURE_EXCHANGE_CONFIRM_WIRE
+  // over a TALER_ConfirmWirePS
+  // whereby the exchange affirms the successful wire transfer.
+  exchange_sig: EddsaSignatureString;
+
+  // Public EdDSA key of the exchange that was used to generate the signature.
+  // Should match one of the exchange's signing keys from /keys.  Again given
+  // explicitly as the client might otherwise be confused by clock skew as to
+  // which signing key was used.
+  exchange_pub: EddsaPublicKeyString;
+}
+
+export const codecForTackTransactionWired = (): Codec<TrackTransactionWired> =>
+  buildCodecForObject<TrackTransactionWired>()
+    .property("wtid", codecForString())
+    .property("execution_time", codecForTimestamp)
+    .property("coin_contribution", codecForAmountString())
+    .property("exchange_sig", codecForString())
+    .property("exchange_pub", codecForString())
+    .build("TackTransactionWired");
+
+interface TrackTransactionAccepted {
+  // Legitimization target that the merchant should
+  // use to check for its KYC status using
+  // the /kyc-check/$REQUIREMENT_ROW/... endpoint.
+  // Optional, not present if the deposit has not
+  // yet been aggregated to the point that a KYC
+  // need has been evaluated.
+  requirement_row?: number;
+
+  // True if the KYC check for the merchant has been
+  // satisfied.  False does not mean that KYC
+  // is strictly needed, unless also a
+  // legitimization_uuid is provided.
+  kyc_ok: boolean;
+
+  // Time by which the exchange currently thinks the deposit will be executed.
+  // Actual execution may be later if the KYC check is not satisfied by then.
+  execution_time: TalerProtocolTimestamp;
+}
+
+export const codecForTackTransactionAccepted =
+  (): Codec<TrackTransactionAccepted> =>
+    buildCodecForObject<TrackTransactionAccepted>()
+      .property("requirement_row", codecOptional(codecForNumber()))
+      .property("kyc_ok", codecForBoolean())
+      .property("execution_time", codecForTimestamp)
+      .build("TackTransactionAccepted");
+
+export type TrackTransaction =
+  | ({ type: "accepted" } & TrackTransactionAccepted)
+  | ({ type: "wired" } & TrackTransactionWired);
 
 export interface PurseDeposit {
   /**
@@ -1959,6 +2120,9 @@ export interface ExchangePurseDeposits {
   deposits: PurseDeposit[];
 }
 
+/**
+ * @deprecated batch deposit should be used.
+ */
 export interface ExchangeDepositRequest {
   // Amount to be deposited, can be a fraction of the
   // coin's total value.
@@ -2020,3 +2184,233 @@ export interface ExchangeDepositRequest {
 
   h_age_commitment?: string;
 }
+
+export type WireSalt = string;
+
+export interface ExchangeBatchDepositRequest {
+  // The merchant's account details.
+  merchant_payto_uri: string;
+
+  // The salt is used to hide the ``payto_uri`` from customers
+  // when computing the ``h_wire`` of the merchant.
+  wire_salt: WireSalt;
+
+  // SHA-512 hash of the contract of the merchant with the customer.  Further
+  // details are never disclosed to the exchange.
+  h_contract_terms: HashCodeString;
+
+  // The list of coins that are going to be deposited with this Request.
+  coins: BatchDepositRequestCoin[];
+
+  // Timestamp when the contract was finalized.
+  timestamp: TalerProtocolTimestamp;
+
+  // Indicative time by which the exchange undertakes to transfer the funds to
+  // the merchant, in case of successful payment. A wire transfer deadline of 'never'
+  // is not allowed.
+  wire_transfer_deadline: TalerProtocolTimestamp;
+
+  // EdDSA `public key of the merchant <merchant-pub>`, so that the client can identify the
+  // merchant for refund requests.
+  merchant_pub: EddsaPublicKeyString;
+
+  // Date until which the merchant can issue a refund to the customer via the
+  // exchange, to be omitted if refunds are not allowed.
+  //
+  // THIS FIELD WILL BE DEPRICATED, once the refund mechanism becomes a
+  // policy via extension.
+  refund_deadline?: TalerProtocolTimestamp;
+
+  // CAVEAT: THIS IS WORK IN PROGRESS
+  // (Optional) policy for the batch-deposit.
+  // This might be a refund, auction or escrow policy.
+  policy?: any;
+}
+
+export interface BatchDepositRequestCoin {
+  // EdDSA public key of the coin being deposited.
+  coin_pub: EddsaPublicKeyString;
+
+  // Hash of denomination RSA key with which the coin is signed.
+  denom_pub_hash: HashCodeString;
+
+  // Exchange's unblinded RSA signature of the coin.
+  ub_sig: UnblindedSignature;
+
+  // Amount to be deposited, can be a fraction of the
+  // coin's total value.
+  contribution: Amounts;
+
+  // Signature over `TALER_DepositRequestPS`, made by the customer with the
+  // `coin's private key <coin-priv>`.
+  coin_sig: EddsaSignatureString;
+
+  h_age_commitment?: string;
+}
+
+export interface WalletKycUuid {
+  // UUID that the wallet should use when initiating
+  // the KYC check.
+  requirement_row: number;
+
+  // Hash of the payto:// account URI for the wallet.
+  h_payto: string;
+}
+
+export const codecForWalletKycUuid = (): Codec<WalletKycUuid> =>
+  buildCodecForObject<WalletKycUuid>()
+    .property("requirement_row", codecForNumber())
+    .property("h_payto", codecForString())
+    .build("WalletKycUuid");
+
+export interface MerchantUsingTemplateDetails {
+  summary?: string;
+  amount?: AmountString;
+}
+
+export interface ExchangeRefundRequest {
+  // Amount to be refunded, can be a fraction of the
+  // coin's total deposit value (including deposit fee);
+  // must be larger than the refund fee.
+  refund_amount: AmountString;
+
+  // SHA-512 hash of the contact of the merchant with the customer.
+  h_contract_terms: HashCodeString;
+
+  // 64-bit transaction id of the refund transaction between merchant and customer.
+  rtransaction_id: number;
+
+  // EdDSA public key of the merchant.
+  merchant_pub: EddsaPublicKeyString;
+
+  // EdDSA signature of the merchant over a
+  // TALER_RefundRequestPS with purpose
+  // TALER_SIGNATURE_MERCHANT_REFUND
+  // affirming the refund.
+  merchant_sig: EddsaPublicKeyString;
+}
+
+export interface ExchangeRefundSuccessResponse {
+  // The EdDSA :ref:signature (binary-only) with purpose
+  // TALER_SIGNATURE_EXCHANGE_CONFIRM_REFUND over
+  // a TALER_RecoupRefreshConfirmationPS
+  // using a current signing key of the
+  // exchange affirming the successful refund.
+  exchange_sig: EddsaSignatureString;
+
+  // Public EdDSA key of the exchange that was used to generate the signature.
+  // Should match one of the exchange's signing keys from /keys.  It is given
+  // explicitly as the client might otherwise be confused by clock skew as to
+  // which signing key was used.
+  exchange_pub: EddsaPublicKeyString;
+}
+
+export const codecForExchangeRefundSuccessResponse =
+  (): Codec<ExchangeRefundSuccessResponse> =>
+    buildCodecForObject<ExchangeRefundSuccessResponse>()
+      .property("exchange_pub", codecForString())
+      .property("exchange_sig", codecForString())
+      .build("ExchangeRefundSuccessResponse");
+
+export type AccountRestriction =
+  | RegexAccountRestriction
+  | DenyAllAccountRestriction;
+
+export interface DenyAllAccountRestriction {
+  type: "deny";
+}
+
+// Accounts interacting with this type of account
+// restriction must have a payto://-URI matching
+// the given regex.
+export interface RegexAccountRestriction {
+  type: "regex";
+
+  // Regular expression that the payto://-URI of the
+  // partner account must follow.  The regular expression
+  // should follow posix-egrep, but without support for character
+  // classes, GNU extensions, back-references or intervals. See
+  // https://www.gnu.org/software/findutils/manual/html_node/find_html/posix_002degrep-regular-expression-syntax.html
+  // for a description of the posix-egrep syntax. Applications
+  // may support regexes with additional features, but exchanges
+  // must not use such regexes.
+  payto_regex: string;
+
+  // Hint for a human to understand the restriction
+  // (that is hopefully easier to comprehend than the regex itself).
+  human_hint: string;
+
+  // Map from IETF BCP 47 language tags to localized
+  // human hints.
+  human_hint_i18n?: InternationalizedString;
+}
+
+export interface ExchangeWireAccount {
+  // payto:// URI identifying the account and wire method
+  payto_uri: string;
+
+  // URI to convert amounts from or to the currency used by
+  // this wire account of the exchange. Missing if no
+  // conversion is applicable.
+  conversion_url?: string;
+
+  // Restrictions that apply to bank accounts that would send
+  // funds to the exchange (crediting this exchange bank account).
+  // Optional, empty array for unrestricted.
+  credit_restrictions: AccountRestriction[];
+
+  // Restrictions that apply to bank accounts that would receive
+  // funds from the exchange (debiting this exchange bank account).
+  // Optional, empty array for unrestricted.
+  debit_restrictions: AccountRestriction[];
+
+  // Signature using the exchange's offline key over
+  // a TALER_MasterWireDetailsPS
+  // with purpose TALER_SIGNATURE_MASTER_WIRE_DETAILS.
+  master_sig: EddsaSignatureString;
+}
+
+export const codecForExchangeWireAccount = (): Codec<ExchangeWireAccount> =>
+  buildCodecForObject<ExchangeWireAccount>()
+    .property("conversion_url", codecOptional(codecForString()))
+    .property("credit_restrictions", codecForList(codecForAny()))
+    .property("debit_restrictions", codecForList(codecForAny()))
+    .property("master_sig", codecForString())
+    .property("payto_uri", codecForString())
+    .build("WireAccount");
+
+export type Integer = number;
+
+export interface BankConversionInfoConfig {
+  // libtool-style representation of the Bank protocol version, see
+  // https://www.gnu.org/software/libtool/manual/html_node/Versioning.html#Versioning
+  // The format is "current:revision:age".
+  version: string;
+
+  // Name of the API.
+  name: "taler-conversion-info";
+
+  regional_currency: string;
+
+  fiat_currency: string;
+
+  // Currency used by this bank.
+  regional_currency_specification: CurrencySpecification;
+
+  // External currency used during conversion.
+  fiat_currency_specification: CurrencySpecification;
+}
+
+export const codecForBankConversionInfoConfig =
+  (): Codec<BankConversionInfoConfig> =>
+    buildCodecForObject<BankConversionInfoConfig>()
+      .property("name", codecForConstString("taler-conversion-info"))
+      .property("version", codecForString())
+      .property("fiat_currency", codecForString())
+      .property("regional_currency", codecForString())
+      .property("fiat_currency_specification", codecForCurrencySpecificiation())
+      .property(
+        "regional_currency_specification",
+        codecForCurrencySpecificiation(),
+      )
+      .build("BankConversionInfoConfig");
