@@ -27,45 +27,53 @@
 import {
   AbsoluteTime,
   AmountString,
+  AttentionType,
   BackupRecovery,
-  buildCodecForObject,
-  bytesToString,
-  canonicalizeBaseUrl,
-  canonicalJson,
   Codec,
+  DenomKeyType,
+  EddsaKeyPair,
+  HttpStatusCode,
+  Logger,
+  PreparePayResult,
+  PreparePayResultType,
+  RecoveryLoadRequest,
+  RecoveryMergeStrategy,
+  TalerError,
+  TalerErrorCode,
+  TalerErrorDetail,
+  TalerPreciseTimestamp,
+  URL,
+  buildCodecForObject,
+  buildCodecForUnion,
+  bytesToString,
+  canonicalJson,
+  canonicalizeBaseUrl,
   codecForAmountString,
   codecForBoolean,
+  codecForConstString,
   codecForList,
   codecForNumber,
   codecForString,
   codecOptional,
-  ConfirmPayResultType,
   decodeCrock,
-  DenomKeyType,
   durationFromSpec,
   eddsaGetPublic,
-  EddsaKeyPair,
   encodeCrock,
   getRandomBytes,
   hash,
   hashDenomPub,
-  HttpStatusCode,
   j2s,
   kdf,
-  Logger,
   notEmpty,
-  PreparePayResultType,
-  RecoveryLoadRequest,
-  RecoveryMergeStrategy,
   rsaBlind,
   secretbox,
   secretbox_open,
   stringToBytes,
-  TalerErrorDetail,
-  TalerProtocolTimestamp,
-  URL,
-  WalletBackupContentV1,
 } from "@gnu-taler/taler-util";
+import {
+  readSuccessResponseJsonOrThrow,
+  readTalerErrorResponse,
+} from "@gnu-taler/taler-util/http";
 import { gunzipSync, gzipSync } from "fflate";
 import { TalerCryptoInterface } from "../../crypto/cryptoImplementation.js";
 import {
@@ -76,30 +84,25 @@ import {
   ConfigRecord,
   ConfigRecordKey,
   WalletBackupConfState,
+  timestampOptionalPreciseFromDb,
+  timestampPreciseFromDb,
+  timestampPreciseToDb,
 } from "../../db.js";
 import { InternalWalletState } from "../../internal-wallet-state.js";
-import {
-  readSuccessResponseJsonOrThrow,
-  readTalerErrorResponse,
-} from "../../util/http.js";
+import { assertUnreachable } from "../../util/assertUnreachable.js";
 import {
   checkDbInvariant,
   checkLogicInvariant,
 } from "../../util/invariants.js";
+import { addAttentionRequest, removeAttentionRequest } from "../attention.js";
 import {
-  OperationAttemptResult,
-  OperationAttemptResultType,
-  RetryTags,
-  scheduleRetryInTx,
-} from "../../util/retries.js";
-import {
-  checkPaymentByProposalId,
-  confirmPay,
-  preparePayForUri,
-} from "../pay-merchant.js";
-import { exportBackup } from "./export.js";
-import { BackupCryptoPrecomputedData, importBackup } from "./import.js";
-import { getWalletBackupState, provideBackupState } from "./state.js";
+  TaskRunResult,
+  TaskRunResultType,
+  TaskIdentifiers,
+} from "../common.js";
+import { checkPaymentByProposalId, preparePayForUri } from "../pay-merchant.js";
+import { WalletStoresV1 } from "../../db.js";
+import { GetReadOnlyAccess } from "../../util/query.js";
 
 const logger = new Logger("operations/backup.ts");
 
@@ -129,7 +132,7 @@ const magic = "TLRWBK01";
  */
 export async function encryptBackup(
   config: WalletBackupConfState,
-  blob: WalletBackupContentV1,
+  blob: any,
 ): Promise<Uint8Array> {
   const chunks: Uint8Array[] = [];
   chunks.push(stringToBytes(magic));
@@ -148,63 +151,6 @@ export async function encryptBackup(
   return concatArrays(chunks);
 }
 
-/**
- * Compute cryptographic values for a backup blob.
- *
- * FIXME: Take data that we already know from the DB.
- * FIXME: Move computations into crypto worker.
- */
-async function computeBackupCryptoData(
-  cryptoApi: TalerCryptoInterface,
-  backupContent: WalletBackupContentV1,
-): Promise<BackupCryptoPrecomputedData> {
-  const cryptoData: BackupCryptoPrecomputedData = {
-    coinPrivToCompletedCoin: {},
-    rsaDenomPubToHash: {},
-    proposalIdToContractTermsHash: {},
-    proposalNoncePrivToPub: {},
-    reservePrivToPub: {},
-  };
-  for (const backupExchangeDetails of backupContent.exchange_details) {
-    for (const backupDenom of backupExchangeDetails.denominations) {
-      if (backupDenom.denom_pub.cipher !== DenomKeyType.Rsa) {
-        throw Error("unsupported cipher");
-      }
-      for (const backupCoin of backupDenom.coins) {
-        const coinPub = encodeCrock(
-          eddsaGetPublic(decodeCrock(backupCoin.coin_priv)),
-        );
-        const blindedCoin = rsaBlind(
-          hash(decodeCrock(backupCoin.coin_priv)),
-          decodeCrock(backupCoin.blinding_key),
-          decodeCrock(backupDenom.denom_pub.rsa_public_key),
-        );
-        cryptoData.coinPrivToCompletedCoin[backupCoin.coin_priv] = {
-          coinEvHash: encodeCrock(hash(blindedCoin)),
-          coinPub,
-        };
-      }
-      cryptoData.rsaDenomPubToHash[backupDenom.denom_pub.rsa_public_key] =
-        encodeCrock(hashDenomPub(backupDenom.denom_pub));
-    }
-  }
-  for (const backupWg of backupContent.withdrawal_groups) {
-    cryptoData.reservePrivToPub[backupWg.reserve_priv] = encodeCrock(
-      eddsaGetPublic(decodeCrock(backupWg.reserve_priv)),
-    );
-  }
-  for (const purch of backupContent.purchases) {
-    const { h: contractTermsHash } = await cryptoApi.hashString({
-      str: canonicalJson(purch.contract_terms_raw),
-    });
-    const noncePub = encodeCrock(eddsaGetPublic(decodeCrock(purch.nonce_priv)));
-    cryptoData.proposalNoncePrivToPub[purch.nonce_priv] = noncePub;
-    cryptoData.proposalIdToContractTermsHash[purch.proposal_id] =
-      contractTermsHash;
-  }
-  return cryptoData;
-}
-
 function deriveAccountKeyPair(
   bc: WalletBackupConfState,
   providerUrl: string,
@@ -232,17 +178,11 @@ function deriveBlobSecret(bc: WalletBackupConfState): Uint8Array {
 
 interface BackupForProviderArgs {
   backupProviderBaseUrl: string;
-
-  /**
-   * Should we attempt one more upload after trying
-   * to pay?
-   */
-  retryAfterPayment: boolean;
 }
 
-function getNextBackupTimestamp(): TalerProtocolTimestamp {
+function getNextBackupTimestamp(): TalerPreciseTimestamp {
   // FIXME:  Randomize!
-  return AbsoluteTime.toTimestamp(
+  return AbsoluteTime.toPreciseTimestamp(
     AbsoluteTime.addDuration(
       AbsoluteTime.now(),
       durationFromSpec({ minutes: 5 }),
@@ -253,7 +193,7 @@ function getNextBackupTimestamp(): TalerProtocolTimestamp {
 async function runBackupCycleForProvider(
   ws: InternalWalletState,
   args: BackupForProviderArgs,
-): Promise<OperationAttemptResult> {
+): Promise<TaskRunResult> {
   const provider = await ws.db
     .mktx((x) => [x.backupProviders])
     .runReadOnly(async (tx) => {
@@ -262,13 +202,12 @@ async function runBackupCycleForProvider(
 
   if (!provider) {
     logger.warn("provider disappeared");
-    return {
-      type: OperationAttemptResultType.Finished,
-      result: undefined,
-    };
+    return TaskRunResult.finished();
   }
 
-  const backupJson = await exportBackup(ws);
+  //const backupJson = await exportBackup(ws);
+  // FIXME: re-implement backup
+  const backupJson = {};
   const backupConfig = await provideBackupState(ws);
   const encBackup = await encryptBackup(backupConfig, backupJson);
   const currentBackupHash = hash(encBackup);
@@ -294,6 +233,10 @@ async function runBackupCycleForProvider(
     provider.baseUrl,
   );
 
+  if (provider.shouldRetryFreshProposal) {
+    accountBackupUrl.searchParams.set("fresh", "yes");
+  }
+
   const resp = await ws.http.fetch(accountBackupUrl.href, {
     method: "POST",
     body: encBackup,
@@ -319,17 +262,22 @@ async function runBackupCycleForProvider(
         if (!prov) {
           return;
         }
-        prov.lastBackupCycleTimestamp = TalerProtocolTimestamp.now();
+        prov.lastBackupCycleTimestamp = timestampPreciseToDb(
+          TalerPreciseTimestamp.now(),
+        );
         prov.state = {
           tag: BackupProviderStateTag.Ready,
-          nextBackupTimestamp: getNextBackupTimestamp(),
+          nextBackupTimestamp: timestampPreciseToDb(getNextBackupTimestamp()),
         };
         await tx.backupProviders.put(prov);
       });
-    return {
-      type: OperationAttemptResultType.Finished,
-      result: undefined,
-    };
+
+    removeAttentionRequest(ws, {
+      entityId: provider.baseUrl,
+      type: AttentionType.BackupUnpaid,
+    });
+
+    return TaskRunResult.finished();
   }
 
   if (resp.status === HttpStatusCode.PaymentRequired) {
@@ -339,57 +287,73 @@ async function runBackupCycleForProvider(
     if (!talerUri) {
       throw Error("no taler URI available to pay provider");
     }
-    const res = await preparePayForUri(ws, talerUri);
-    let proposalId = res.proposalId;
-    let doPay = false;
-    switch (res.status) {
-      case PreparePayResultType.InsufficientBalance:
-        // FIXME: record in provider state!
-        logger.warn("insufficient balance to pay for backup provider");
-        proposalId = res.proposalId;
-        break;
-      case PreparePayResultType.PaymentPossible:
-        doPay = true;
-        break;
-      case PreparePayResultType.AlreadyConfirmed:
-        break;
+
+    //We can't delay downloading the proposal since we need the id
+    //FIXME: check download errors
+    let res: PreparePayResult | undefined = undefined;
+    try {
+      res = await preparePayForUri(ws, talerUri);
+    } catch (e) {
+      const error = TalerError.fromException(e);
+      if (!error.hasErrorCode(TalerErrorCode.WALLET_ORDER_ALREADY_CLAIMED)) {
+        throw error;
+      }
     }
 
-    // FIXME: check if the provider is overcharging us!
+    if (res === undefined) {
+      //claimed
+
+      await ws.db
+        .mktx((x) => [x.backupProviders, x.operationRetries])
+        .runReadWrite(async (tx) => {
+          const prov = await tx.backupProviders.get(provider.baseUrl);
+          if (!prov) {
+            logger.warn("backup provider not found anymore");
+            return;
+          }
+          prov.shouldRetryFreshProposal = true;
+          prov.state = {
+            tag: BackupProviderStateTag.Retrying,
+          };
+          await tx.backupProviders.put(prov);
+        });
+
+      return {
+        type: TaskRunResultType.Pending,
+      };
+    }
+    const result = res;
 
     await ws.db
       .mktx((x) => [x.backupProviders, x.operationRetries])
       .runReadWrite(async (tx) => {
-        const provRec = await tx.backupProviders.get(provider.baseUrl);
-        checkDbInvariant(!!provRec);
-        const ids = new Set(provRec.paymentProposalIds);
-        ids.add(proposalId);
-        provRec.paymentProposalIds = Array.from(ids).sort();
-        provRec.currentPaymentProposalId = proposalId;
-        // FIXME: allocate error code for this!
-        await tx.backupProviders.put(provRec);
-        const opId = RetryTags.forBackup(provRec);
-        await scheduleRetryInTx(ws, tx, opId);
+        const prov = await tx.backupProviders.get(provider.baseUrl);
+        if (!prov) {
+          logger.warn("backup provider not found anymore");
+          return;
+        }
+        const opId = TaskIdentifiers.forBackup(prov);
+        //await scheduleRetryInTx(ws, tx, opId);
+        prov.currentPaymentProposalId = result.proposalId;
+        prov.shouldRetryFreshProposal = false;
+        prov.state = {
+          tag: BackupProviderStateTag.Retrying,
+        };
+        await tx.backupProviders.put(prov);
       });
 
-    if (doPay) {
-      const confirmRes = await confirmPay(ws, proposalId);
-      switch (confirmRes.type) {
-        case ConfirmPayResultType.Pending:
-          logger.warn("payment not yet finished yet");
-          break;
-      }
-    }
+    addAttentionRequest(
+      ws,
+      {
+        type: AttentionType.BackupUnpaid,
+        provider_base_url: provider.baseUrl,
+        talerUri,
+      },
+      provider.baseUrl,
+    );
 
-    if (args.retryAfterPayment) {
-      return await runBackupCycleForProvider(ws, {
-        ...args,
-        retryAfterPayment: false,
-      });
-    }
     return {
-      type: OperationAttemptResultType.Pending,
-      result: undefined,
+      type: TaskRunResultType.Pending,
     };
   }
 
@@ -402,16 +366,23 @@ async function runBackupCycleForProvider(
           return;
         }
         prov.lastBackupHash = encodeCrock(currentBackupHash);
-        prov.lastBackupCycleTimestamp = TalerProtocolTimestamp.now();
+        prov.lastBackupCycleTimestamp = timestampPreciseToDb(
+          TalerPreciseTimestamp.now(),
+        );
         prov.state = {
           tag: BackupProviderStateTag.Ready,
-          nextBackupTimestamp: getNextBackupTimestamp(),
+          nextBackupTimestamp: timestampPreciseToDb(getNextBackupTimestamp()),
         };
         await tx.backupProviders.put(prov);
       });
+
+    removeAttentionRequest(ws, {
+      entityId: provider.baseUrl,
+      type: AttentionType.BackupUnpaid,
+    });
+
     return {
-      type: OperationAttemptResultType.Finished,
-      result: undefined,
+      type: TaskRunResultType.Finished,
     };
   }
 
@@ -419,9 +390,9 @@ async function runBackupCycleForProvider(
     logger.info("conflicting backup found");
     const backupEnc = new Uint8Array(await resp.bytes());
     const backupConfig = await provideBackupState(ws);
-    const blob = await decryptBackup(backupConfig, backupEnc);
-    const cryptoData = await computeBackupCryptoData(ws.cryptoApi, blob);
-    await importBackup(ws, blob, cryptoData);
+    // const blob = await decryptBackup(backupConfig, backupEnc);
+    // FIXME: Re-implement backup import with merging
+    // await importBackup(ws, blob, cryptoData);
     await ws.db
       .mktx((x) => [x.backupProviders, x.operationRetries])
       .runReadWrite(async (tx) => {
@@ -433,8 +404,8 @@ async function runBackupCycleForProvider(
         prov.lastBackupHash = encodeCrock(hash(backupEnc));
         // FIXME: Allocate error code for this situation?
         // FIXME: Add operation retry record!
-        const opId = RetryTags.forBackup(prov);
-        await scheduleRetryInTx(ws, tx, opId);
+        const opId = TaskIdentifiers.forBackup(prov);
+        //await scheduleRetryInTx(ws, tx, opId);
         prov.state = {
           tag: BackupProviderStateTag.Retrying,
         };
@@ -442,10 +413,7 @@ async function runBackupCycleForProvider(
       });
     logger.info("processed existing backup");
     // Now upload our own, merged backup.
-    return await runBackupCycleForProvider(ws, {
-      ...args,
-      retryAfterPayment: false,
-    });
+    return await runBackupCycleForProvider(ws, args);
   }
 
   // Some other response that we did not expect!
@@ -455,7 +423,7 @@ async function runBackupCycleForProvider(
   const err = await readTalerErrorResponse(resp);
   logger.error(`got error response from backup provider: ${j2s(err)}`);
   return {
-    type: OperationAttemptResultType.Error,
+    type: TaskRunResultType.Error,
     errorDetail: err,
   };
 }
@@ -463,7 +431,7 @@ async function runBackupCycleForProvider(
 export async function processBackupForProvider(
   ws: InternalWalletState,
   backupProviderBaseUrl: string,
-): Promise<OperationAttemptResult> {
+): Promise<TaskRunResult> {
   const provider = await ws.db
     .mktx((x) => [x.backupProviders])
     .runReadOnly(async (tx) => {
@@ -477,7 +445,6 @@ export async function processBackupForProvider(
 
   return await runBackupCycleForProvider(ws, {
     backupProviderBaseUrl: provider.baseUrl,
-    retryAfterPayment: true,
   });
 }
 
@@ -540,12 +507,11 @@ export async function runBackupCycle(
   for (const provider of providers) {
     await runBackupCycleForProvider(ws, {
       backupProviderBaseUrl: provider.baseUrl,
-      retryAfterPayment: true,
     });
   }
 }
 
-interface SyncTermsOfServiceResponse {
+export interface SyncTermsOfServiceResponse {
   // maximum backup size supported
   storage_limit_in_megabytes: number;
 
@@ -557,7 +523,7 @@ interface SyncTermsOfServiceResponse {
   version: string;
 }
 
-const codecForSyncTermsOfServiceResponse =
+export const codecForSyncTermsOfServiceResponse =
   (): Codec<SyncTermsOfServiceResponse> =>
     buildCodecForObject<SyncTermsOfServiceResponse>()
       .property("storage_limit_in_megabytes", codecForNumber())
@@ -584,10 +550,45 @@ export const codecForAddBackupProviderRequest =
       .property("activate", codecOptional(codecForBoolean()))
       .build("AddBackupProviderRequest");
 
+export type AddBackupProviderResponse =
+  | AddBackupProviderOk
+  | AddBackupProviderPaymentRequired;
+
+interface AddBackupProviderOk {
+  status: "ok";
+}
+interface AddBackupProviderPaymentRequired {
+  status: "payment-required";
+  talerUri?: string;
+}
+
+export const codecForAddBackupProviderOk = (): Codec<AddBackupProviderOk> =>
+  buildCodecForObject<AddBackupProviderOk>()
+    .property("status", codecForConstString("ok"))
+    .build("AddBackupProviderOk");
+
+export const codecForAddBackupProviderPaymenrRequired =
+  (): Codec<AddBackupProviderPaymentRequired> =>
+    buildCodecForObject<AddBackupProviderPaymentRequired>()
+      .property("status", codecForConstString("payment-required"))
+      .property("talerUri", codecOptional(codecForString()))
+      .build("AddBackupProviderPaymentRequired");
+
+export const codecForAddBackupProviderResponse =
+  (): Codec<AddBackupProviderResponse> =>
+    buildCodecForUnion<AddBackupProviderResponse>()
+      .discriminateOn("status")
+      .alternative("ok", codecForAddBackupProviderOk())
+      .alternative(
+        "payment-required",
+        codecForAddBackupProviderPaymenrRequired(),
+      )
+      .build("AddBackupProviderResponse");
+
 export async function addBackupProvider(
   ws: InternalWalletState,
   req: AddBackupProviderRequest,
-): Promise<void> {
+): Promise<AddBackupProviderResponse> {
   logger.info(`adding backup provider ${j2s(req)}`);
   await provideBackupState(ws);
   const canonUrl = canonicalizeBaseUrl(req.backupProviderBaseUrl);
@@ -600,7 +601,9 @@ export async function addBackupProvider(
         if (req.activate) {
           oldProv.state = {
             tag: BackupProviderStateTag.Ready,
-            nextBackupTimestamp: TalerProtocolTimestamp.now(),
+            nextBackupTimestamp: timestampPreciseToDb(
+              TalerPreciseTimestamp.now(),
+            ),
           };
           logger.info("setting existing backup provider to active");
           await tx.backupProviders.put(oldProv);
@@ -609,7 +612,7 @@ export async function addBackupProvider(
       }
     });
   const termsUrl = new URL("config", canonUrl);
-  const resp = await ws.http.get(termsUrl.href);
+  const resp = await ws.http.fetch(termsUrl.href);
   const terms = await readSuccessResponseJsonOrThrow(
     resp,
     codecForSyncTermsOfServiceResponse(),
@@ -618,10 +621,13 @@ export async function addBackupProvider(
     .mktx((x) => [x.backupProviders])
     .runReadWrite(async (tx) => {
       let state: BackupProviderState;
+      //FIXME: what is the difference provisional and ready?
       if (req.activate) {
         state = {
           tag: BackupProviderStateTag.Ready,
-          nextBackupTimestamp: TalerProtocolTimestamp.now(),
+          nextBackupTimestamp: timestampPreciseToDb(
+            TalerPreciseTimestamp.now(),
+          ),
         };
       } else {
         state = {
@@ -636,11 +642,46 @@ export async function addBackupProvider(
           storageLimitInMegabytes: terms.storage_limit_in_megabytes,
           supportedProtocolVersion: terms.version,
         },
+        shouldRetryFreshProposal: false,
         paymentProposalIds: [],
         baseUrl: canonUrl,
         uids: [encodeCrock(getRandomBytes(32))],
       });
     });
+
+  return await runFirstBackupCycleForProvider(ws, {
+    backupProviderBaseUrl: canonUrl,
+  });
+}
+
+async function runFirstBackupCycleForProvider(
+  ws: InternalWalletState,
+  args: BackupForProviderArgs,
+): Promise<AddBackupProviderResponse> {
+  const resp = await runBackupCycleForProvider(ws, args);
+  switch (resp.type) {
+    case TaskRunResultType.Error:
+      throw TalerError.fromDetail(
+        TalerErrorCode.WALLET_UNEXPECTED_EXCEPTION,
+        resp.errorDetail as any, //FIXME create an error for backup problems
+      );
+    case TaskRunResultType.Finished:
+      return {
+        status: "ok",
+      };
+    case TaskRunResultType.Longpoll:
+      throw Error(
+        "unexpected runFirstBackupCycleForProvider result (longpoll)",
+      );
+    case TaskRunResultType.Pending:
+      return {
+        status: "payment-required",
+        talerUri: "FIXME",
+        //talerUri: resp.result.talerUri,
+      };
+    default:
+      assertUnreachable(resp);
+  }
 }
 
 export async function restoreFromRecoverySecret(): Promise<void> {
@@ -662,8 +703,8 @@ export interface ProviderInfo {
    * Last communication issue with the provider.
    */
   lastError?: TalerErrorDetail;
-  lastSuccessfulBackupTimestamp?: TalerProtocolTimestamp;
-  lastAttemptedBackupTimestamp?: TalerProtocolTimestamp;
+  lastSuccessfulBackupTimestamp?: TalerPreciseTimestamp;
+  lastAttemptedBackupTimestamp?: TalerPreciseTimestamp;
   paymentProposalIds: string[];
   backupProblem?: BackupProblem;
   paymentStatus: ProviderPaymentStatus;
@@ -701,18 +742,6 @@ export interface BackupInfo {
   providers: ProviderInfo[];
 }
 
-export async function importBackupPlain(
-  ws: InternalWalletState,
-  blob: any,
-): Promise<void> {
-  // FIXME: parse
-  const backup: WalletBackupContentV1 = blob;
-
-  const cryptoData = await computeBackupCryptoData(ws.cryptoApi, backup);
-
-  await importBackup(ws, blob, cryptoData);
-}
-
 export enum ProviderPaymentType {
   Unpaid = "unpaid",
   Pending = "pending",
@@ -727,10 +756,12 @@ export interface ProviderPaymentUnpaid {
 
 export interface ProviderPaymentInsufficientBalance {
   type: ProviderPaymentType.InsufficientBalance;
+  amount: AmountString;
 }
 
 export interface ProviderPaymentPending {
   type: ProviderPaymentType.Pending;
+  talerUri?: string;
 }
 
 export interface ProviderPaymentPaid {
@@ -757,33 +788,43 @@ async function getProviderPaymentInfo(
   const status = await checkPaymentByProposalId(
     ws,
     provider.currentPaymentProposalId,
-  );
-  if (status.status === PreparePayResultType.InsufficientBalance) {
+  ).catch(() => undefined);
+
+  if (!status) {
     return {
-      type: ProviderPaymentType.InsufficientBalance,
+      type: ProviderPaymentType.Unpaid,
     };
   }
-  if (status.status === PreparePayResultType.PaymentPossible) {
-    return {
-      type: ProviderPaymentType.Pending,
-    };
-  }
-  if (status.status === PreparePayResultType.AlreadyConfirmed) {
-    if (status.paid) {
+
+  switch (status.status) {
+    case PreparePayResultType.InsufficientBalance:
       return {
-        type: ProviderPaymentType.Paid,
-        paidUntil: AbsoluteTime.addDuration(
-          AbsoluteTime.fromTimestamp(status.contractTerms.timestamp),
-          durationFromSpec({ years: 1 }),
-        ),
+        type: ProviderPaymentType.InsufficientBalance,
+        amount: status.amountRaw,
       };
-    } else {
+    case PreparePayResultType.PaymentPossible:
       return {
         type: ProviderPaymentType.Pending,
+        talerUri: status.talerUri,
       };
-    }
+    case PreparePayResultType.AlreadyConfirmed:
+      if (status.paid) {
+        return {
+          type: ProviderPaymentType.Paid,
+          paidUntil: AbsoluteTime.addDuration(
+            AbsoluteTime.fromProtocolTimestamp(status.contractTerms.timestamp),
+            durationFromSpec({ years: 1 }), //FIXME: take this from the contract term
+          ),
+        };
+      } else {
+        return {
+          type: ProviderPaymentType.Pending,
+          talerUri: status.talerUri,
+        };
+      }
+    default:
+      assertUnreachable(status);
   }
-  throw Error("not reached");
 }
 
 /**
@@ -797,7 +838,7 @@ export async function getBackupInfo(
     .mktx((x) => [x.backupProviders, x.operationRetries])
     .runReadOnly(async (tx) => {
       return await tx.backupProviders.iter().mapAsync(async (bp) => {
-        const opId = RetryTags.forBackup(bp);
+        const opId = TaskIdentifiers.forBackup(bp);
         const retryRecord = await tx.operationRetries.get(opId);
         return {
           provider: bp,
@@ -810,7 +851,9 @@ export async function getBackupInfo(
     providers.push({
       active: x.provider.state.tag !== BackupProviderStateTag.Provisional,
       syncProviderBaseUrl: x.provider.baseUrl,
-      lastSuccessfulBackupTimestamp: x.provider.lastBackupCycleTimestamp,
+      lastSuccessfulBackupTimestamp: timestampOptionalPreciseFromDb(
+        x.provider.lastBackupCycleTimestamp,
+      ),
       paymentProposalIds: x.provider.paymentProposalIds,
       lastError:
         x.provider.state.tag === BackupProviderStateTag.Retrying
@@ -884,9 +927,12 @@ async function backupRecoveryTheirs(
             baseUrl: prov.url,
             name: prov.name,
             paymentProposalIds: [],
+            shouldRetryFreshProposal: false,
             state: {
               tag: BackupProviderStateTag.Ready,
-              nextBackupTimestamp: TalerProtocolTimestamp.now(),
+              nextBackupTimestamp: timestampPreciseToDb(
+                TalerPreciseTimestamp.now(),
+              ),
             },
             uids: [encodeCrock(getRandomBytes(32))],
           });
@@ -935,23 +981,10 @@ export async function loadBackupRecovery(
   }
 }
 
-export async function exportBackupEncrypted(
-  ws: InternalWalletState,
-): Promise<Uint8Array> {
-  await provideBackupState(ws);
-  const blob = await exportBackup(ws);
-  const bs = await ws.db
-    .mktx((x) => [x.config])
-    .runReadOnly(async (tx) => {
-      return await getWalletBackupState(ws, tx);
-    });
-  return encryptBackup(bs, blob);
-}
-
 export async function decryptBackup(
   backupConfig: WalletBackupConfState,
   data: Uint8Array,
-): Promise<WalletBackupContentV1> {
+): Promise<any> {
   const rMagic = bytesToString(data.slice(0, 8));
   if (rMagic != magic) {
     throw Error("invalid backup file (magic tag mismatch)");
@@ -967,12 +1000,85 @@ export async function decryptBackup(
   return JSON.parse(bytesToString(gunzipSync(dataCompressed)));
 }
 
-export async function importBackupEncrypted(
+export async function provideBackupState(
   ws: InternalWalletState,
-  data: Uint8Array,
+): Promise<WalletBackupConfState> {
+  const bs: ConfigRecord | undefined = await ws.db
+    .mktx((stores) => [stores.config])
+    .runReadOnly(async (tx) => {
+      return await tx.config.get(ConfigRecordKey.WalletBackupState);
+    });
+  if (bs) {
+    checkDbInvariant(bs.key === ConfigRecordKey.WalletBackupState);
+    return bs.value;
+  }
+  // We need to generate the key outside of the transaction
+  // due to how IndexedDB works.
+  const k = await ws.cryptoApi.createEddsaKeypair({});
+  const d = getRandomBytes(5);
+  // FIXME: device ID should be configured when wallet is initialized
+  // and be based on hostname
+  const deviceId = `wallet-core-${encodeCrock(d)}`;
+  return await ws.db
+    .mktx((x) => [x.config])
+    .runReadWrite(async (tx) => {
+      let backupStateEntry: ConfigRecord | undefined = await tx.config.get(
+        ConfigRecordKey.WalletBackupState,
+      );
+      if (!backupStateEntry) {
+        backupStateEntry = {
+          key: ConfigRecordKey.WalletBackupState,
+          value: {
+            deviceId,
+            walletRootPub: k.pub,
+            walletRootPriv: k.priv,
+            lastBackupPlainHash: undefined,
+          },
+        };
+        await tx.config.put(backupStateEntry);
+      }
+      checkDbInvariant(
+        backupStateEntry.key === ConfigRecordKey.WalletBackupState,
+      );
+      return backupStateEntry.value;
+    });
+}
+
+export async function getWalletBackupState(
+  ws: InternalWalletState,
+  tx: GetReadOnlyAccess<{ config: typeof WalletStoresV1.config }>,
+): Promise<WalletBackupConfState> {
+  const bs = await tx.config.get(ConfigRecordKey.WalletBackupState);
+  checkDbInvariant(!!bs, "wallet backup state should be in DB");
+  checkDbInvariant(bs.key === ConfigRecordKey.WalletBackupState);
+  return bs.value;
+}
+
+export async function setWalletDeviceId(
+  ws: InternalWalletState,
+  deviceId: string,
 ): Promise<void> {
-  const backupConfig = await provideBackupState(ws);
-  const blob = await decryptBackup(backupConfig, data);
-  const cryptoData = await computeBackupCryptoData(ws.cryptoApi, blob);
-  await importBackup(ws, blob, cryptoData);
+  await provideBackupState(ws);
+  await ws.db
+    .mktx((x) => [x.config])
+    .runReadWrite(async (tx) => {
+      let backupStateEntry: ConfigRecord | undefined = await tx.config.get(
+        ConfigRecordKey.WalletBackupState,
+      );
+      if (
+        !backupStateEntry ||
+        backupStateEntry.key !== ConfigRecordKey.WalletBackupState
+      ) {
+        return;
+      }
+      backupStateEntry.value.deviceId = deviceId;
+      await tx.config.put(backupStateEntry);
+    });
+}
+
+export async function getWalletDeviceId(
+  ws: InternalWalletState,
+): Promise<string> {
+  const bs = await provideBackupState(ws);
+  return bs.deviceId;
 }
