@@ -23,102 +23,75 @@ import {
   CancellationToken,
   canonicalizeBaseUrl,
   codecForExchangeKeysJson,
-  codecForExchangeWireJson,
   DenominationPubKey,
+  DenomKeyType,
   Duration,
   durationFromSpec,
   encodeCrock,
   ExchangeAuditor,
-  ExchangeDenomination,
   ExchangeGlobalFees,
   ExchangeSignKeyJson,
-  ExchangeWireJson,
+  ExchangeWireAccount,
   GlobalFees,
   hashDenomPub,
   j2s,
   LibtoolVersion,
   Logger,
+  makeErrorDetail,
   NotificationType,
   parsePaytoUri,
   Recoup,
+  TalerError,
   TalerErrorCode,
+  TalerErrorDetail,
+  TalerPreciseTimestamp,
   TalerProtocolDuration,
   TalerProtocolTimestamp,
   URL,
   WireFee,
   WireFeeMap,
+  WireFeesJson,
   WireInfo,
 } from "@gnu-taler/taler-util";
-import {
-  DenominationRecord,
-  DenominationVerificationStatus,
-  ExchangeDetailsRecord,
-  ExchangeRecord,
-  WalletStoresV1,
-} from "../db.js";
-import { TalerError } from "../errors.js";
-import { InternalWalletState, TrustInfo } from "../internal-wallet-state.js";
 import {
   getExpiry,
   HttpRequestLibrary,
   readSuccessResponseJsonOrThrow,
   readSuccessResponseTextOrThrow,
-} from "../util/http.js";
+} from "@gnu-taler/taler-util/http";
+import {
+  DenominationRecord,
+  DenominationVerificationStatus,
+  ExchangeDetailsRecord,
+  ExchangeEntryRecord,
+  WalletStoresV1,
+} from "../db.js";
+import {
+  ExchangeEntryDbRecordStatus,
+  ExchangeEntryDbUpdateStatus,
+  isWithdrawableDenom,
+  timestampPreciseFromDb,
+  timestampPreciseToDb,
+  timestampProtocolToDb,
+  WalletDbReadWriteTransaction,
+} from "../index.js";
+import { InternalWalletState } from "../internal-wallet-state.js";
 import { checkDbInvariant } from "../util/invariants.js";
 import {
   DbAccess,
   GetReadOnlyAccess,
   GetReadWriteAccess,
 } from "../util/query.js";
-import {
-  OperationAttemptResult,
-  OperationAttemptResultType,
-  RetryTags,
-  unwrapOperationHandlerResultOrThrow,
-} from "../util/retries.js";
 import { WALLET_EXCHANGE_PROTOCOL_VERSION } from "../versions.js";
-import { runOperationWithErrorReporting } from "./common.js";
-import { isWithdrawableDenom } from "./withdraw.js";
+import {
+  runTaskWithErrorReporting,
+  TaskIdentifiers,
+  TaskRunResult,
+  TaskRunResultType,
+} from "./common.js";
 
 const logger = new Logger("exchanges.ts");
 
-function denominationRecordFromKeys(
-  exchangeBaseUrl: string,
-  exchangeMasterPub: string,
-  listIssueDate: TalerProtocolTimestamp,
-  denomIn: ExchangeDenomination,
-): DenominationRecord {
-  let denomPub: DenominationPubKey;
-  denomPub = denomIn.denom_pub;
-  const denomPubHash = encodeCrock(hashDenomPub(denomPub));
-  const value = Amounts.parseOrThrow(denomIn.value);
-  const d: DenominationRecord = {
-    denomPub,
-    denomPubHash,
-    exchangeBaseUrl,
-    exchangeMasterPub,
-    fees: {
-      feeDeposit: Amounts.stringify(denomIn.fee_deposit),
-      feeRefresh: Amounts.stringify(denomIn.fee_refresh),
-      feeRefund: Amounts.stringify(denomIn.fee_refund),
-      feeWithdraw: Amounts.stringify(denomIn.fee_withdraw),
-    },
-    isOffered: true,
-    isRevoked: false,
-    masterSig: denomIn.master_sig,
-    stampExpireDeposit: denomIn.stamp_expire_deposit,
-    stampExpireLegal: denomIn.stamp_expire_legal,
-    stampExpireWithdraw: denomIn.stamp_expire_withdraw,
-    stampStart: denomIn.stamp_start,
-    verificationStatus: DenominationVerificationStatus.Unverified,
-    amountFrac: value.fraction,
-    amountVal: value.value,
-    currency: value.currency,
-    listIssueDate,
-  };
-  return d;
-}
-
 export function getExchangeRequestTimeout(): Duration {
   return Duration.fromSpec({
     seconds: 5,
@@ -137,13 +110,13 @@ export async function downloadExchangeWithTermsOfService(
   timeout: Duration,
   contentType: string,
 ): Promise<ExchangeTosDownloadResult> {
-  logger.info(`downloading exchange tos (type ${contentType})`);
+  logger.trace(`downloading exchange tos (type ${contentType})`);
   const reqUrl = new URL("terms", exchangeBaseUrl);
   const headers = {
     Accept: contentType,
   };
 
-  const resp = await http.get(reqUrl.href, {
+  const resp = await http.fetch(reqUrl.href, {
     headers,
     timeout,
   });
@@ -183,35 +156,6 @@ export async function getExchangeDetails(
 getExchangeDetails.makeContext = (db: DbAccess<typeof WalletStoresV1>) =>
   db.mktx((x) => [x.exchanges, x.exchangeDetails]);
 
-/**
- * Update the database based on the download of the terms of service.
- */
-export async function updateExchangeTermsOfService(
-  ws: InternalWalletState,
-  exchangeBaseUrl: string,
-  tos: ExchangeTosDownloadResult,
-): Promise<void> {
-  await ws.db
-    .mktx((x) => [x.exchanges, x.exchangeTos, x.exchangeDetails])
-    .runReadWrite(async (tx) => {
-      const d = await getExchangeDetails(tx, exchangeBaseUrl);
-      let tosRecord = await tx.exchangeTos.get([exchangeBaseUrl, tos.tosEtag]);
-      if (!tosRecord) {
-        tosRecord = {
-          etag: tos.tosEtag,
-          exchangeBaseUrl,
-          termsOfServiceContentType: tos.tosContentType,
-          termsOfServiceText: tos.tosText,
-        };
-        await tx.exchangeTos.put(tosRecord);
-      }
-      if (d) {
-        d.tosCurrentEtag = tos.tosEtag;
-        await tx.exchangeDetails.put(d);
-      }
-    });
-}
-
 /**
  * Mark a ToS version as accepted by the user.
  *
@@ -229,7 +173,7 @@ export async function acceptExchangeTermsOfService(
       if (d) {
         d.tosAccepted = {
           etag: etag || d.tosCurrentEtag,
-          timestamp: TalerProtocolTimestamp.now(),
+          timestamp: timestampPreciseToDb(TalerPreciseTimestamp.now()),
         };
         await tx.exchangeDetails.put(d);
       }
@@ -239,13 +183,13 @@ export async function acceptExchangeTermsOfService(
 async function validateWireInfo(
   ws: InternalWalletState,
   versionCurrent: number,
-  wireInfo: ExchangeWireJson,
+  wireInfo: ExchangeKeysDownloadResult,
   masterPublicKey: string,
 ): Promise<WireInfo> {
   for (const a of wireInfo.accounts) {
     logger.trace("validating exchange acct");
     let isValid = false;
-    if (ws.insecureTrustExchange) {
+    if (ws.config.testing.insecureTrustExchange) {
       isValid = true;
     } else {
       const { valid: v } = await ws.cryptoApi.isValidWireAccount({
@@ -253,6 +197,9 @@ async function validateWireInfo(
         paytoUri: a.payto_uri,
         sig: a.master_sig,
         versionCurrent,
+        conversionUrl: a.conversion_url,
+        creditRestrictions: a.credit_restrictions,
+        debitRestrictions: a.debit_restrictions,
       });
       isValid = v;
     }
@@ -260,10 +207,11 @@ async function validateWireInfo(
       throw Error("exchange acct signature invalid");
     }
   }
+  logger.trace("account validation done");
   const feesForType: WireFeeMap = {};
-  for (const wireMethod of Object.keys(wireInfo.fees)) {
+  for (const wireMethod of Object.keys(wireInfo.wireFees)) {
     const feeList: WireFee[] = [];
-    for (const x of wireInfo.fees[wireMethod]) {
+    for (const x of wireInfo.wireFees[wireMethod]) {
       const startStamp = x.start_date;
       const endStamp = x.end_date;
       const fee: WireFee = {
@@ -274,7 +222,7 @@ async function validateWireInfo(
         wireFee: Amounts.stringify(x.wire_fee),
       };
       let isValid = false;
-      if (ws.insecureTrustExchange) {
+      if (ws.config.testing.insecureTrustExchange) {
         isValid = true;
       } else {
         const { valid: v } = await ws.cryptoApi.isValidWireFee({
@@ -307,7 +255,7 @@ async function validateGlobalFees(
   for (const gf of fees) {
     logger.trace("validating exchange global fees");
     let isValid = false;
-    if (ws.insecureTrustExchange) {
+    if (ws.config.testing.insecureTrustExchange) {
       isValid = true;
     } else {
       const { valid: v } = await ws.cryptoApi.isValidGlobalFees({
@@ -337,7 +285,6 @@ async function validateGlobalFees(
 }
 
 export interface ExchangeInfo {
-  wire: ExchangeWireJson;
   keys: ExchangeKeysDownloadResult;
 }
 
@@ -345,11 +292,6 @@ export async function downloadExchangeInfo(
   exchangeBaseUrl: string,
   http: HttpRequestLibrary,
 ): Promise<ExchangeInfo> {
-  const wireInfo = await downloadExchangeWireInfo(
-    exchangeBaseUrl,
-    http,
-    Duration.getForever(),
-  );
   const keysInfo = await downloadExchangeKeysInfo(
     exchangeBaseUrl,
     http,
@@ -357,34 +299,36 @@ export async function downloadExchangeInfo(
   );
   return {
     keys: keysInfo,
-    wire: wireInfo,
   };
 }
 
-/**
- * Fetch wire information for an exchange.
- *
- * @param exchangeBaseUrl Exchange base URL, assumed to be already normalized.
- */
-async function downloadExchangeWireInfo(
+export async function addPresetExchangeEntry(
+  tx: WalletDbReadWriteTransaction<"exchanges">,
   exchangeBaseUrl: string,
-  http: HttpRequestLibrary,
-  timeout: Duration,
-): Promise<ExchangeWireJson> {
-  const reqUrl = new URL("wire", exchangeBaseUrl);
-
-  const resp = await http.get(reqUrl.href, {
-    timeout,
-  });
-  const wireInfo = await readSuccessResponseJsonOrThrow(
-    resp,
-    codecForExchangeWireJson(),
-  );
-
-  return wireInfo;
+  currencyHint?: string,
+): Promise<void> {
+  let exchange = await tx.exchanges.get(exchangeBaseUrl);
+  if (!exchange) {
+    const r: ExchangeEntryRecord = {
+      entryStatus: ExchangeEntryDbRecordStatus.Preset,
+      updateStatus: ExchangeEntryDbUpdateStatus.Initial,
+      baseUrl: exchangeBaseUrl,
+      presetCurrencyHint: currencyHint,
+      detailsPointer: undefined,
+      lastUpdate: undefined,
+      lastKeysEtag: undefined,
+      nextRefreshCheckStamp: timestampPreciseToDb(
+        AbsoluteTime.toPreciseTimestamp(AbsoluteTime.never()),
+      ),
+      nextUpdateStamp: timestampPreciseToDb(
+        AbsoluteTime.toPreciseTimestamp(AbsoluteTime.never()),
+      ),
+    };
+    await tx.exchanges.put(r);
+  }
 }
 
-export async function provideExchangeRecordInTx(
+async function provideExchangeRecordInTx(
   ws: InternalWalletState,
   tx: GetReadWriteAccess<{
     exchanges: typeof WalletStoresV1.exchanges;
@@ -393,20 +337,24 @@ export async function provideExchangeRecordInTx(
   baseUrl: string,
   now: AbsoluteTime,
 ): Promise<{
-  exchange: ExchangeRecord;
+  exchange: ExchangeEntryRecord;
   exchangeDetails: ExchangeDetailsRecord | undefined;
 }> {
   let exchange = await tx.exchanges.get(baseUrl);
   if (!exchange) {
-    const r: ExchangeRecord = {
-      permanent: true,
+    const r: ExchangeEntryRecord = {
+      entryStatus: ExchangeEntryDbRecordStatus.Ephemeral,
+      updateStatus: ExchangeEntryDbUpdateStatus.InitialUpdate,
       baseUrl: baseUrl,
       detailsPointer: undefined,
       lastUpdate: undefined,
-      nextUpdate: AbsoluteTime.toTimestamp(now),
-      nextRefreshCheck: AbsoluteTime.toTimestamp(now),
+      nextUpdateStamp: timestampPreciseToDb(
+        AbsoluteTime.toPreciseTimestamp(AbsoluteTime.never()),
+      ),
+      nextRefreshCheckStamp: timestampPreciseToDb(
+        AbsoluteTime.toPreciseTimestamp(AbsoluteTime.never()),
+      ),
       lastKeysEtag: undefined,
-      lastWireEtag: undefined,
     };
     await tx.exchanges.put(r);
     exchange = r;
@@ -416,6 +364,7 @@ export async function provideExchangeRecordInTx(
 }
 
 interface ExchangeKeysDownloadResult {
+  baseUrl: string;
   masterPublicKey: string;
   currency: string;
   auditors: ExchangeAuditor[];
@@ -427,6 +376,8 @@ interface ExchangeKeysDownloadResult {
   recoup: Recoup[];
   listIssueDate: TalerProtocolTimestamp;
   globalFees: GlobalFees[];
+  accounts: ExchangeWireAccount[];
+  wireFees: { [methodName: string]: WireFeesJson[] };
 }
 
 /**
@@ -439,7 +390,7 @@ async function downloadExchangeKeysInfo(
 ): Promise<ExchangeKeysDownloadResult> {
   const keysUrl = new URL("keys", baseUrl);
 
-  const resp = await http.get(keysUrl.href, {
+  const resp = await http.fetch(keysUrl.href, {
     timeout,
   });
   const exchangeKeysJsonUnchecked = await readSuccessResponseJsonOrThrow(
@@ -447,7 +398,7 @@ async function downloadExchangeKeysInfo(
     codecForExchangeKeysJson(),
   );
 
-  if (exchangeKeysJsonUnchecked.denoms.length === 0) {
+  if (exchangeKeysJsonUnchecked.denominations.length === 0) {
     throw TalerError.fromDetail(
       TalerErrorCode.WALLET_EXCHANGE_DENOMINATIONS_INSUFFICIENT,
       {
@@ -474,26 +425,81 @@ async function downloadExchangeKeysInfo(
     );
   }
 
-  const currency = Amounts.parseOrThrow(
-    exchangeKeysJsonUnchecked.denoms[0].value,
-  ).currency.toUpperCase();
+  const currency = exchangeKeysJsonUnchecked.currency;
+
+  const currentDenominations: DenominationRecord[] = [];
+
+  for (const denomGroup of exchangeKeysJsonUnchecked.denominations) {
+    switch (denomGroup.cipher) {
+      case "RSA":
+      case "RSA+age_restricted": {
+        let ageMask = 0;
+        if (denomGroup.cipher === "RSA+age_restricted") {
+          ageMask = denomGroup.age_mask;
+        }
+        for (const denomIn of denomGroup.denoms) {
+          const denomPub: DenominationPubKey = {
+            age_mask: ageMask,
+            cipher: DenomKeyType.Rsa,
+            rsa_public_key: denomIn.rsa_pub,
+          };
+          const denomPubHash = encodeCrock(hashDenomPub(denomPub));
+          const value = Amounts.parseOrThrow(denomGroup.value);
+          const rec: DenominationRecord = {
+            denomPub,
+            denomPubHash,
+            exchangeBaseUrl: baseUrl,
+            exchangeMasterPub: exchangeKeysJsonUnchecked.master_public_key,
+            isOffered: true,
+            isRevoked: false,
+            value: Amounts.stringify(value),
+            currency: value.currency,
+            stampExpireDeposit: timestampProtocolToDb(
+              denomIn.stamp_expire_deposit,
+            ),
+            stampExpireLegal: timestampProtocolToDb(denomIn.stamp_expire_legal),
+            stampExpireWithdraw: timestampProtocolToDb(
+              denomIn.stamp_expire_withdraw,
+            ),
+            stampStart: timestampProtocolToDb(denomIn.stamp_start),
+            verificationStatus: DenominationVerificationStatus.Unverified,
+            masterSig: denomIn.master_sig,
+            listIssueDate: timestampProtocolToDb(
+              exchangeKeysJsonUnchecked.list_issue_date,
+            ),
+            fees: {
+              feeDeposit: Amounts.stringify(denomGroup.fee_deposit),
+              feeRefresh: Amounts.stringify(denomGroup.fee_refresh),
+              feeRefund: Amounts.stringify(denomGroup.fee_refund),
+              feeWithdraw: Amounts.stringify(denomGroup.fee_withdraw),
+            },
+          };
+          currentDenominations.push(rec);
+        }
+        break;
+      }
+      case "CS+age_restricted":
+      case "CS":
+        logger.warn("Clause-Schnorr denominations not supported");
+        continue;
+      default:
+        logger.warn(
+          `denomination type ${(denomGroup as any).cipher} not supported`,
+        );
+        continue;
+    }
+  }
 
   return {
     masterPublicKey: exchangeKeysJsonUnchecked.master_public_key,
     currency,
+    baseUrl: exchangeKeysJsonUnchecked.base_url,
     auditors: exchangeKeysJsonUnchecked.auditors,
-    currentDenominations: exchangeKeysJsonUnchecked.denoms.map((d) =>
-      denominationRecordFromKeys(
-        baseUrl,
-        exchangeKeysJsonUnchecked.master_public_key,
-        exchangeKeysJsonUnchecked.list_issue_date,
-        d,
-      ),
-    ),
+    currentDenominations,
     protocolVersion: exchangeKeysJsonUnchecked.version,
     signingKeys: exchangeKeysJsonUnchecked.signkeys,
     reserveClosingDelay: exchangeKeysJsonUnchecked.reserve_closing_delay,
-    expiry: AbsoluteTime.toTimestamp(
+    expiry: AbsoluteTime.toProtocolTimestamp(
       getExpiry(resp, {
         minDuration: durationFromSpec({ hours: 1 }),
       }),
@@ -501,6 +507,8 @@ async function downloadExchangeKeysInfo(
     recoup: exchangeKeysJsonUnchecked.recoup ?? [],
     listIssueDate: exchangeKeysJsonUnchecked.list_issue_date,
     globalFees: exchangeKeysJsonUnchecked.global_fees,
+    accounts: exchangeKeysJsonUnchecked.accounts,
+    wireFees: exchangeKeysJsonUnchecked.wire_fees,
   };
 }
 
@@ -537,25 +545,51 @@ export async function downloadTosFromAcceptedFormat(
   );
 }
 
+/**
+ * FIXME: Split this into two parts: (a) triggering the exchange
+ * to be updated and (b) waiting for the update to finish.
+ */
 export async function updateExchangeFromUrl(
   ws: InternalWalletState,
   baseUrl: string,
   options: {
+    checkMasterPub?: string;
     forceNow?: boolean;
     cancellationToken?: CancellationToken;
   } = {},
 ): Promise<{
-  exchange: ExchangeRecord;
+  exchange: ExchangeEntryRecord;
   exchangeDetails: ExchangeDetailsRecord;
 }> {
   const canonUrl = canonicalizeBaseUrl(baseUrl);
-  return unwrapOperationHandlerResultOrThrow(
-    await runOperationWithErrorReporting(
-      ws,
-      RetryTags.forExchangeUpdateFromUrl(canonUrl),
-      () => updateExchangeFromUrlHandler(ws, canonUrl, options),
-    ),
+  const res = await runTaskWithErrorReporting(
+    ws,
+    TaskIdentifiers.forExchangeUpdateFromUrl(canonUrl),
+    () => updateExchangeFromUrlHandler(ws, canonUrl, options),
   );
+  switch (res.type) {
+    case TaskRunResultType.Finished: {
+      const now = AbsoluteTime.now();
+      const { exchange, exchangeDetails } = await ws.db
+        .mktx((x) => [x.exchanges, x.exchangeDetails])
+        .runReadWrite(async (tx) => {
+          let exchange = await tx.exchanges.get(canonUrl);
+          const exchangeDetails = await getExchangeDetails(tx, canonUrl);
+          return { exchange, exchangeDetails };
+        });
+      if (!exchange) {
+        throw Error("exchange not found");
+      }
+      if (!exchangeDetails) {
+        throw Error("exchange details not found");
+      }
+      return { exchange, exchangeDetails };
+    }
+    case TaskRunResultType.Error:
+      throw TalerError.fromUncheckedDetail(res.errorDetail);
+    default:
+      throw Error(`unexpected operation result (${res.type})`);
+  }
 }
 
 /**
@@ -567,17 +601,13 @@ export async function updateExchangeFromUrlHandler(
   ws: InternalWalletState,
   exchangeBaseUrl: string,
   options: {
+    checkMasterPub?: string;
     forceNow?: boolean;
     cancellationToken?: CancellationToken;
   } = {},
-): Promise<
-  OperationAttemptResult<{
-    exchange: ExchangeRecord;
-    exchangeDetails: ExchangeDetailsRecord;
-  }>
-> {
+): Promise<TaskRunResult> {
   const forceNow = options.forceNow ?? false;
-  logger.info(
+  logger.trace(
     `updating exchange info for ${exchangeBaseUrl}, forced: ${forceNow}`,
   );
 
@@ -597,16 +627,24 @@ export async function updateExchangeFromUrlHandler(
   if (
     !forceNow &&
     exchangeDetails !== undefined &&
-    !AbsoluteTime.isExpired(AbsoluteTime.fromTimestamp(exchange.nextUpdate))
+    !AbsoluteTime.isExpired(
+      AbsoluteTime.fromPreciseTimestamp(
+        timestampPreciseFromDb(exchange.nextUpdateStamp),
+      ),
+    )
   ) {
-    logger.info("using existing exchange info");
-    return {
-      type: OperationAttemptResultType.Finished,
-      result: { exchange, exchangeDetails },
-    };
+    logger.trace("using existing exchange info");
+
+    if (options.checkMasterPub) {
+      if (exchangeDetails.masterPublicKey !== options.checkMasterPub) {
+        throw Error(`master public key mismatch`);
+      }
+    }
+
+    return TaskRunResult.finished();
   }
 
-  logger.info("updating exchange /keys info");
+  logger.trace("updating exchange /keys info");
 
   const timeout = getExchangeRequestTimeout();
 
@@ -616,14 +654,13 @@ export async function updateExchangeFromUrlHandler(
     timeout,
   );
 
-  logger.info("updating exchange /wire info");
-  const wireInfoDownload = await downloadExchangeWireInfo(
-    exchangeBaseUrl,
-    ws.http,
-    timeout,
-  );
+  if (options.checkMasterPub) {
+    if (keysInfo.masterPublicKey !== options.checkMasterPub) {
+      throw Error(`master public key mismatch`);
+    }
+  }
 
-  logger.info("validating exchange /wire info");
+  logger.trace("validating exchange wire info");
 
   const version = LibtoolVersion.parseVersion(keysInfo.protocolVersion);
   if (!version) {
@@ -634,7 +671,7 @@ export async function updateExchangeFromUrlHandler(
   const wireInfo = await validateWireInfo(
     ws,
     version.current,
-    wireInfoDownload,
+    keysInfo,
     keysInfo.masterPublicKey,
   );
 
@@ -643,8 +680,22 @@ export async function updateExchangeFromUrlHandler(
     keysInfo.globalFees,
     keysInfo.masterPublicKey,
   );
+  if (keysInfo.baseUrl != exchangeBaseUrl) {
+    logger.warn("exchange base URL mismatch");
+    const errorDetail: TalerErrorDetail = makeErrorDetail(
+      TalerErrorCode.WALLET_EXCHANGE_BASE_URL_MISMATCH,
+      {
+        urlWallet: exchangeBaseUrl,
+        urlExchange: keysInfo.baseUrl,
+      },
+    );
+    return {
+      type: TaskRunResultType.Error,
+      errorDetail,
+    };
+  }
 
-  logger.info("finished validating exchange /wire info");
+  logger.trace("finished validating exchange /wire info");
 
   // We download the text/plain version here,
   // because that one needs to exist, and we
@@ -664,7 +715,10 @@ export async function updateExchangeFromUrlHandler(
 
   let ageMask = 0;
   for (const x of keysInfo.currentDenominations) {
-    if (isWithdrawableDenom(x) && x.denomPub.age_mask != 0) {
+    if (
+      isWithdrawableDenom(x, ws.config.testing.denomselAllowLate) &&
+      x.denomPub.age_mask != 0
+    ) {
       ageMask = x.denomPub.age_mask;
       break;
     }
@@ -673,9 +727,8 @@ export async function updateExchangeFromUrlHandler(
   const updated = await ws.db
     .mktx((x) => [
       x.exchanges,
-      x.exchangeTos,
       x.exchangeDetails,
-      x.exchangeSignkeys,
+      x.exchangeSignKeys,
       x.denominations,
       x.coins,
       x.refreshGroups,
@@ -717,49 +770,40 @@ export async function updateExchangeFromUrlHandler(
       if (existingDetails?.rowId) {
         newDetails.rowId = existingDetails.rowId;
       }
-      r.lastUpdate = TalerProtocolTimestamp.now();
-      r.nextUpdate = keysInfo.expiry;
+      r.lastUpdate = timestampPreciseToDb(TalerPreciseTimestamp.now());
+      r.nextUpdateStamp = timestampPreciseToDb(
+        AbsoluteTime.toPreciseTimestamp(
+          AbsoluteTime.fromProtocolTimestamp(keysInfo.expiry),
+        ),
+      );
       // New denominations might be available.
-      r.nextRefreshCheck = TalerProtocolTimestamp.now();
+      r.nextRefreshCheckStamp = timestampPreciseToDb(
+        TalerPreciseTimestamp.now(),
+      );
       if (detailsPointerChanged) {
         r.detailsPointer = {
           currency: newDetails.currency,
           masterPublicKey: newDetails.masterPublicKey,
-          updateClock: TalerProtocolTimestamp.now(),
+          updateClock: timestampPreciseToDb(TalerPreciseTimestamp.now()),
         };
       }
       await tx.exchanges.put(r);
       const drRowId = await tx.exchangeDetails.put(newDetails);
       checkDbInvariant(typeof drRowId.key === "number");
 
-      let tosRecord = await tx.exchangeTos.get([
-        exchangeBaseUrl,
-        tosDownload.tosEtag,
-      ]);
-
-      if (!tosRecord || tosRecord.etag !== existingTosAccepted?.etag) {
-        tosRecord = {
-          etag: tosDownload.tosEtag,
-          exchangeBaseUrl,
-          termsOfServiceContentType: tosDownload.tosContentType,
-          termsOfServiceText: tosDownload.tosText,
-        };
-        await tx.exchangeTos.put(tosRecord);
-      }
-
       for (const sk of keysInfo.signingKeys) {
         // FIXME: validate signing keys before inserting them
         await tx.exchangeSignKeys.put({
           exchangeDetailsRowId: drRowId.key,
           masterSig: sk.master_sig,
           signkeyPub: sk.key,
-          stampEnd: sk.stamp_end,
-          stampExpire: sk.stamp_expire,
-          stampStart: sk.stamp_start,
+          stampEnd: timestampProtocolToDb(sk.stamp_end),
+          stampExpire: timestampProtocolToDb(sk.stamp_expire),
+          stampStart: timestampProtocolToDb(sk.stamp_start),
         });
       }
 
-      logger.info("updating denominations in database");
+      logger.trace("updating denominations in database");
       const currentDenomSet = new Set<string>(
         keysInfo.currentDenominations.map((x) => x.denomPubHash),
       );
@@ -790,7 +834,7 @@ export async function updateExchangeFromUrlHandler(
               );
             }
           } else {
-            x.listIssueDate = keysInfo.listIssueDate;
+            x.listIssueDate = timestampProtocolToDb(keysInfo.listIssueDate);
             if (!x.isOffered) {
               x.isOffered = true;
               logger.info(
@@ -851,9 +895,7 @@ export async function updateExchangeFromUrlHandler(
   if (recoupGroupId) {
     // Asynchronously start recoup.  This doesn't need to finish
     // for the exchange update to be considered finished.
-    ws.recoupOps.processRecoupGroup(ws, recoupGroupId).catch((e) => {
-      logger.error("error while recouping coins:", e);
-    });
+    ws.workAvailable.trigger();
   }
 
   if (!updated) {
@@ -868,13 +910,7 @@ export async function updateExchangeFromUrlHandler(
     });
   }
 
-  return {
-    type: OperationAttemptResultType.Finished,
-    result: {
-      exchange: updated.exchange,
-      exchangeDetails: updated.exchangeDetails,
-    },
-  };
+  return TaskRunResult.finished();
 }
 
 /**
@@ -911,54 +947,3 @@ export async function getExchangePaytoUri(
     )}`,
   );
 }
-
-/**
- * Check if and how an exchange is trusted and/or audited.
- */
-export async function getExchangeTrust(
-  ws: InternalWalletState,
-  exchangeInfo: ExchangeRecord,
-): Promise<TrustInfo> {
-  let isTrusted = false;
-  let isAudited = false;
-
-  return await ws.db
-    .mktx((x) => [
-      x.exchanges,
-      x.exchangeDetails,
-      x.exchangeTrust,
-      x.auditorTrust,
-    ])
-    .runReadOnly(async (tx) => {
-      const exchangeDetails = await getExchangeDetails(
-        tx,
-        exchangeInfo.baseUrl,
-      );
-
-      if (!exchangeDetails) {
-        throw Error(`exchange ${exchangeInfo.baseUrl} details not available`);
-      }
-      const exchangeTrustRecord =
-        await tx.exchangeTrust.indexes.byExchangeMasterPub.get(
-          exchangeDetails.masterPublicKey,
-        );
-      if (
-        exchangeTrustRecord &&
-        exchangeTrustRecord.uids.length > 0 &&
-        exchangeTrustRecord.currency === exchangeDetails.currency
-      ) {
-        isTrusted = true;
-      }
-
-      for (const auditor of exchangeDetails.auditors) {
-        const auditorTrustRecord =
-          await tx.auditorTrust.indexes.byAuditorPub.get(auditor.auditor_pub);
-        if (auditorTrustRecord && auditorTrustRecord.uids.length > 0) {
-          isAudited = true;
-          break;
-        }
-      }
-
-      return { isTrusted, isAudited };
-    });
-}
