@@ -36,23 +36,36 @@ import {
   ExchangeProtocolVersion,
   ExchangeRefreshRevealRequest,
   fnutil,
+  getErrorDetailFromException,
   getRandomBytes,
   HashCodeString,
   HttpStatusCode,
   j2s,
   Logger,
+  makeErrorDetail,
   NotificationType,
   RefreshGroupId,
   RefreshReason,
-  TalerProtocolTimestamp,
+  TalerError,
+  TalerErrorCode,
+  TalerErrorDetail,
+  TalerPreciseTimestamp,
+  TransactionAction,
+  TransactionMajorState,
+  TransactionState,
+  TransactionType,
   URL,
 } from "@gnu-taler/taler-util";
+import {
+  readSuccessResponseJsonOrThrow,
+  readUnexpectedResponseDetails,
+} from "@gnu-taler/taler-util/http";
 import { TalerCryptoInterface } from "../crypto/cryptoImplementation.js";
 import {
   DerivedRefreshSession,
   RefreshNewDenomInfo,
 } from "../crypto/cryptoTypes.js";
-import { CryptoApiStoppedError } from "../crypto/workers/cryptoDispatcher.js";
+import { CryptoApiStoppedError } from "../crypto/workers/crypto-dispatcher.js";
 import {
   CoinRecord,
   CoinSourceType,
@@ -60,30 +73,36 @@ import {
   RefreshCoinStatus,
   RefreshGroupRecord,
   RefreshOperationStatus,
+  RefreshReasonDetails,
   WalletStoresV1,
 } from "../db.js";
-import { TalerError } from "../errors.js";
+import {
+  isWithdrawableDenom,
+  PendingTaskType,
+  RefreshSessionRecord,
+  timestampPreciseToDb,
+  timestampProtocolFromDb,
+} from "../index.js";
 import {
   EXCHANGE_COINS_LOCK,
   InternalWalletState,
 } from "../internal-wallet-state.js";
 import { assertUnreachable } from "../util/assertUnreachable.js";
-import {
-  readSuccessResponseJsonOrThrow,
-  readUnexpectedResponseDetails,
-} from "../util/http.js";
+import { selectWithdrawalDenominations } from "../util/coinSelection.js";
 import { checkDbInvariant } from "../util/invariants.js";
-import { GetReadWriteAccess } from "../util/query.js";
+import { GetReadOnlyAccess, GetReadWriteAccess } from "../util/query.js";
 import {
-  OperationAttemptResult,
-  OperationAttemptResultType,
-} from "../util/retries.js";
-import { makeCoinAvailable } from "./common.js";
+  constructTaskIdentifier,
+  makeCoinAvailable,
+  makeCoinsVisible,
+  TaskRunResult,
+  TaskRunResultType,
+} from "./common.js";
 import { updateExchangeFromUrl } from "./exchanges.js";
 import {
-  isWithdrawableDenom,
-  selectWithdrawalDenominations,
-} from "./withdraw.js";
+  constructTransactionIdentifier,
+  notifyTransition,
+} from "./transactions.js";
 
 const logger = new Logger("refresh.ts");
 
@@ -102,21 +121,22 @@ export function getTotalRefreshCost(
   denoms: DenominationRecord[],
   refreshedDenom: DenominationInfo,
   amountLeft: AmountJson,
+  denomselAllowLate: boolean,
 ): AmountJson {
   const withdrawAmount = Amounts.sub(
     amountLeft,
     refreshedDenom.feeRefresh,
   ).amount;
   const denomMap = Object.fromEntries(denoms.map((x) => [x.denomPubHash, x]));
-  const withdrawDenoms = selectWithdrawalDenominations(withdrawAmount, denoms);
+  const withdrawDenoms = selectWithdrawalDenominations(
+    withdrawAmount,
+    denoms,
+    denomselAllowLate,
+  );
   const resultingAmount = Amounts.add(
     Amounts.zeroOfCurrency(withdrawAmount.currency),
     ...withdrawDenoms.selectedDenoms.map(
-      (d) =>
-        Amounts.mult(
-          DenominationRecord.getValue(denomMap[d.denomPubHash]),
-          d.count,
-        ).amount,
+      (d) => Amounts.mult(denomMap[d.denomPubHash].value, d.count).amount,
     ),
   ).amount;
   const totalCost = Amounts.sub(amountLeft, resultingAmount).amount;
@@ -128,40 +148,47 @@ export function getTotalRefreshCost(
   return totalCost;
 }
 
-function updateGroupStatus(rg: RefreshGroupRecord): void {
-  const allDone = fnutil.all(
+function updateGroupStatus(rg: RefreshGroupRecord): { final: boolean } {
+  const allFinal = fnutil.all(
     rg.statusPerCoin,
-    (x) => x === RefreshCoinStatus.Finished || x === RefreshCoinStatus.Frozen,
+    (x) => x === RefreshCoinStatus.Finished || x === RefreshCoinStatus.Failed,
   );
-  const anyFrozen = fnutil.any(
+  const anyFailed = fnutil.any(
     rg.statusPerCoin,
-    (x) => x === RefreshCoinStatus.Frozen,
+    (x) => x === RefreshCoinStatus.Failed,
   );
-  if (allDone) {
-    if (anyFrozen) {
-      rg.timestampFinished = AbsoluteTime.toTimestamp(AbsoluteTime.now());
-      rg.operationStatus = RefreshOperationStatus.FinishedWithError;
+  if (allFinal) {
+    if (anyFailed) {
+      rg.timestampFinished = timestampPreciseToDb(TalerPreciseTimestamp.now());
+      rg.operationStatus = RefreshOperationStatus.Failed;
     } else {
-      rg.timestampFinished = AbsoluteTime.toTimestamp(AbsoluteTime.now());
+      rg.timestampFinished = timestampPreciseToDb(TalerPreciseTimestamp.now());
       rg.operationStatus = RefreshOperationStatus.Finished;
     }
+    return { final: true };
   }
+  return { final: false };
 }
 
 /**
  * Create a refresh session for one particular coin inside a refresh group.
+ *
+ * If the session already exists, return the existing one.
+ *
+ * If the session doesn't need to be created (refresh group gone or session already
+ * finished), return undefined.
  */
-async function refreshCreateSession(
+async function provideRefreshSession(
   ws: InternalWalletState,
   refreshGroupId: string,
   coinIndex: number,
-): Promise<void> {
+): Promise<RefreshSessionRecord | undefined> {
   logger.trace(
     `creating refresh session for coin ${coinIndex} in refresh group ${refreshGroupId}`,
   );
 
   const d = await ws.db
-    .mktx((x) => [x.refreshGroups, x.coins])
+    .mktx((x) => [x.refreshGroups, x.coins, x.refreshSessions])
     .runReadWrite(async (tx) => {
       const refreshGroup = await tx.refreshGroups.get(refreshGroupId);
       if (!refreshGroup) {
@@ -172,21 +199,24 @@ async function refreshCreateSession(
       ) {
         return;
       }
-      const existingRefreshSession =
-        refreshGroup.refreshSessionPerCoin[coinIndex];
-      if (existingRefreshSession) {
-        return;
-      }
+      const existingRefreshSession = await tx.refreshSessions.get([
+        refreshGroupId,
+        coinIndex,
+      ]);
       const oldCoinPub = refreshGroup.oldCoinPubs[coinIndex];
       const coin = await tx.coins.get(oldCoinPub);
       if (!coin) {
         throw Error("Can't refresh, coin not found");
       }
-      return { refreshGroup, coin };
+      return { refreshGroup, coin, existingRefreshSession };
     });
 
   if (!d) {
-    return;
+    return undefined;
+  }
+
+  if (d.existingRefreshSession) {
+    return d.existingRefreshSession;
   }
 
   const { refreshGroup, coin } = d;
@@ -229,44 +259,62 @@ async function refreshCreateSession(
   const newCoinDenoms = selectWithdrawalDenominations(
     availableAmount,
     availableDenoms,
+    ws.config.testing.denomselAllowLate,
   );
 
+  const transactionId = constructTransactionIdentifier({
+    tag: TransactionType.Refresh,
+    refreshGroupId,
+  });
+
   if (newCoinDenoms.selectedDenoms.length === 0) {
     logger.trace(
       `not refreshing, available amount ${amountToPretty(
         availableAmount,
       )} too small`,
     );
-    await ws.db
-      .mktx((x) => [x.coins, x.refreshGroups])
+    const transitionInfo = await ws.db
+      .mktx((x) => [x.coins, x.coinAvailability, x.refreshGroups])
       .runReadWrite(async (tx) => {
         const rg = await tx.refreshGroups.get(refreshGroupId);
         if (!rg) {
           return;
         }
+        const oldTxState = computeRefreshTransactionState(rg);
         rg.statusPerCoin[coinIndex] = RefreshCoinStatus.Finished;
-        updateGroupStatus(rg);
-
+        const updateRes = updateGroupStatus(rg);
+        if (updateRes.final) {
+          await makeCoinsVisible(ws, tx, transactionId);
+        }
         await tx.refreshGroups.put(rg);
+        const newTxState = computeRefreshTransactionState(rg);
+        return { oldTxState, newTxState };
       });
-    ws.notify({ type: NotificationType.RefreshUnwarranted });
+    ws.notify({ type: NotificationType.BalanceChange });
+    notifyTransition(ws, transactionId, transitionInfo);
     return;
   }
 
   const sessionSecretSeed = encodeCrock(getRandomBytes(64));
 
   // Store refresh session for this coin in the database.
-  await ws.db
-    .mktx((x) => [x.refreshGroups, x.coins])
+  const mySession = await ws.db
+    .mktx((x) => [x.refreshGroups, x.coins, x.refreshSessions])
     .runReadWrite(async (tx) => {
       const rg = await tx.refreshGroups.get(refreshGroupId);
       if (!rg) {
         return;
       }
-      if (rg.refreshSessionPerCoin[coinIndex]) {
-        return;
+      const existingSession = await tx.refreshSessions.get([
+        refreshGroupId,
+        coinIndex,
+      ]);
+      if (existingSession) {
+        return existingSession;
       }
-      rg.refreshSessionPerCoin[coinIndex] = {
+      const newSession: RefreshSessionRecord = {
+        coinIndex,
+        refreshGroupId,
         norevealIndex: undefined,
         sessionSecretSeed: sessionSecretSeed,
         newDenoms: newCoinDenoms.selectedDenoms.map((x) => ({
@@ -275,12 +323,13 @@ async function refreshCreateSession(
         })),
         amountRefreshOutput: Amounts.stringify(newCoinDenoms.totalCoinValue),
       };
-      await tx.refreshGroups.put(rg);
+      await tx.refreshSessions.put(newSession);
+      return newSession;
     });
-  logger.info(
-    `created refresh session for coin #${coinIndex} in ${refreshGroupId}`,
+  logger.trace(
+    `found/created refresh session for coin #${coinIndex} in ${refreshGroupId}`,
   );
-  ws.notify({ type: NotificationType.RefreshStarted });
+  return mySession;
 }
 
 function getRefreshRequestTimeout(rg: RefreshGroupRecord): Duration {
@@ -295,13 +344,16 @@ async function refreshMelt(
   coinIndex: number,
 ): Promise<void> {
   const d = await ws.db
-    .mktx((x) => [x.refreshGroups, x.coins, x.denominations])
+    .mktx((x) => [x.refreshGroups, x.refreshSessions, x.coins, x.denominations])
     .runReadWrite(async (tx) => {
       const refreshGroup = await tx.refreshGroups.get(refreshGroupId);
       if (!refreshGroup) {
         return;
       }
-      const refreshSession = refreshGroup.refreshSessionPerCoin[coinIndex];
+      const refreshSession = await tx.refreshSessions.get([
+        refreshGroupId,
+        coinIndex,
+      ]);
       if (!refreshSession) {
         return;
       }
@@ -403,10 +455,20 @@ async function refreshMelt(
     });
   });
 
+  const transactionId = constructTransactionIdentifier({
+    tag: TransactionType.Refresh,
+    refreshGroupId,
+  });
+
   if (resp.status === HttpStatusCode.NotFound) {
     const errDetails = await readUnexpectedResponseDetails(resp);
-    await ws.db
-      .mktx((x) => [x.refreshGroups])
+    const transitionInfo = await ws.db
+      .mktx((x) => [
+        x.refreshGroups,
+        x.refreshSessions,
+        x.coins,
+        x.coinAvailability,
+      ])
       .runReadWrite(async (tx) => {
         const rg = await tx.refreshGroups.get(refreshGroupId);
         if (!rg) {
@@ -418,11 +480,32 @@ async function refreshMelt(
         if (rg.statusPerCoin[coinIndex] !== RefreshCoinStatus.Pending) {
           return;
         }
-        rg.statusPerCoin[coinIndex] = RefreshCoinStatus.Frozen;
-        rg.lastErrorPerCoin[coinIndex] = errDetails;
-        updateGroupStatus(rg);
+        const oldTxState = computeRefreshTransactionState(rg);
+        rg.statusPerCoin[coinIndex] = RefreshCoinStatus.Failed;
+        const refreshSession = await tx.refreshSessions.get([
+          refreshGroupId,
+          coinIndex,
+        ]);
+        if (!refreshSession) {
+          throw Error(
+            "db invariant failed: missing refresh session in database",
+          );
+        }
+        refreshSession.lastError = errDetails;
+        const updateRes = updateGroupStatus(rg);
+        if (updateRes.final) {
+          await makeCoinsVisible(ws, tx, transactionId);
+        }
         await tx.refreshGroups.put(rg);
+        await tx.refreshSessions.put(refreshSession);
+        const newTxState = computeRefreshTransactionState(rg);
+        return {
+          oldTxState,
+          newTxState,
+        };
       });
+    ws.notify({ type: NotificationType.BalanceChange });
+    notifyTransition(ws, transactionId, transitionInfo);
     return;
   }
 
@@ -446,7 +529,7 @@ async function refreshMelt(
   refreshSession.norevealIndex = norevealIndex;
 
   await ws.db
-    .mktx((x) => [x.refreshGroups])
+    .mktx((x) => [x.refreshGroups, x.refreshSessions])
     .runReadWrite(async (tx) => {
       const rg = await tx.refreshGroups.get(refreshGroupId);
       if (!rg) {
@@ -455,7 +538,7 @@ async function refreshMelt(
       if (rg.timestampFinished) {
         return;
       }
-      const rs = rg.refreshSessionPerCoin[coinIndex];
+      const rs = await tx.refreshSessions.get([refreshGroupId, coinIndex]);
       if (!rs) {
         return;
       }
@@ -463,12 +546,8 @@ async function refreshMelt(
         return;
       }
       rs.norevealIndex = norevealIndex;
-      await tx.refreshGroups.put(rg);
+      await tx.refreshSessions.put(rs);
     });
-
-  ws.notify({
-    type: NotificationType.RefreshMelted,
-  });
 }
 
 export async function assembleRefreshRevealRequest(args: {
@@ -534,17 +613,20 @@ async function refreshReveal(
   refreshGroupId: string,
   coinIndex: number,
 ): Promise<void> {
-  logger.info(
+  logger.trace(
     `doing refresh reveal for ${refreshGroupId} (old coin ${coinIndex})`,
   );
   const d = await ws.db
-    .mktx((x) => [x.refreshGroups, x.coins, x.denominations])
+    .mktx((x) => [x.refreshGroups, x.refreshSessions, x.coins, x.denominations])
     .runReadOnly(async (tx) => {
       const refreshGroup = await tx.refreshGroups.get(refreshGroupId);
       if (!refreshGroup) {
         return;
       }
-      const refreshSession = refreshGroup.refreshSessionPerCoin[coinIndex];
+      const refreshSession = await tx.refreshSessions.get([
+        refreshGroupId,
+        coinIndex,
+      ]);
       if (!refreshSession) {
         return;
       }
@@ -661,6 +743,11 @@ async function refreshReveal(
 
   const coins: CoinRecord[] = [];
 
+  const transactionId = constructTransactionIdentifier({
+    tag: TransactionType.Refresh,
+    refreshGroupId,
+  });
+
   for (let i = 0; i < refreshSession.newDenoms.length; i++) {
     const ncd = newCoinDenoms[i];
     for (let j = 0; j < refreshSession.newDenoms[i].count; j++) {
@@ -690,6 +777,7 @@ async function refreshReveal(
           refreshGroupId,
           oldCoinPub: refreshGroup.oldCoinPubs[coinIndex],
         },
+        sourceTransactionId: transactionId,
         coinEvHash: pc.coinEvHash,
         maxAge: pc.maxAge,
         ageCommitmentProof: pc.ageCommitmentProof,
@@ -700,12 +788,13 @@ async function refreshReveal(
     }
   }
 
-  await ws.db
+  const transitionInfo = await ws.db
     .mktx((x) => [
       x.coins,
       x.denominations,
       x.coinAvailability,
       x.refreshGroups,
+      x.refreshSessions,
     ])
     .runReadWrite(async (tx) => {
       const rg = await tx.refreshGroups.get(refreshGroupId);
@@ -713,56 +802,55 @@ async function refreshReveal(
         logger.warn("no refresh session found");
         return;
       }
-      const rs = rg.refreshSessionPerCoin[coinIndex];
+      const rs = await tx.refreshSessions.get([refreshGroupId, coinIndex]);
       if (!rs) {
         return;
       }
+      const oldTxState = computeRefreshTransactionState(rg);
       rg.statusPerCoin[coinIndex] = RefreshCoinStatus.Finished;
       updateGroupStatus(rg);
       for (const coin of coins) {
         await makeCoinAvailable(ws, tx, coin);
       }
+      await makeCoinsVisible(ws, tx, transactionId);
       await tx.refreshGroups.put(rg);
+      const newTxState = computeRefreshTransactionState(rg);
+      return { oldTxState, newTxState };
     });
+  notifyTransition(ws, transactionId, transitionInfo);
   logger.trace("refresh finished (end of reveal)");
-  ws.notify({
-    type: NotificationType.RefreshRevealed,
-  });
 }
 
 export async function processRefreshGroup(
   ws: InternalWalletState,
   refreshGroupId: string,
-  options: {
-    forceNow?: boolean;
-  } = {},
-): Promise<OperationAttemptResult> {
-  logger.info(`processing refresh group ${refreshGroupId}`);
+  options: Record<string, never> = {},
+): Promise<TaskRunResult> {
+  logger.trace(`processing refresh group ${refreshGroupId}`);
 
   const refreshGroup = await ws.db
     .mktx((x) => [x.refreshGroups])
     .runReadOnly(async (tx) => tx.refreshGroups.get(refreshGroupId));
   if (!refreshGroup) {
-    return {
-      type: OperationAttemptResultType.Finished,
-      result: undefined,
-    };
+    return TaskRunResult.finished();
   }
   if (refreshGroup.timestampFinished) {
-    return {
-      type: OperationAttemptResultType.Finished,
-      result: undefined,
-    };
+    return TaskRunResult.finished();
   }
   // Process refresh sessions of the group in parallel.
   logger.trace("processing refresh sessions for old coins");
+  let errors: TalerErrorDetail[] = [];
+  let inShutdown = false;
   const ps = refreshGroup.oldCoinPubs.map((x, i) =>
     processRefreshSession(ws, refreshGroupId, i).catch((x) => {
       if (x instanceof CryptoApiStoppedError) {
+        inShutdown = true;
         logger.info(
           "crypto API stopped while processing refresh group, probably the wallet is currently shutting down.",
         );
-      } else if (x instanceof TalerError) {
+        return;
+      }
+      if (x instanceof TalerError) {
         logger.warn("process refresh session got exception (TalerError)");
         logger.warn(`exc ${x}`);
         logger.warn(`exc stack ${x.stack}`);
@@ -772,20 +860,34 @@ export async function processRefreshGroup(
         logger.warn(`exc ${x}`);
         logger.warn(`exc stack ${x.stack}`);
       }
+      errors.push(getErrorDetailFromException(x));
     }),
   );
   try {
-    logger.trace("waiting for refreshes");
+    logger.info("waiting for refreshes");
     await Promise.all(ps);
-    logger.trace("refresh finished");
+    logger.info("refresh group finished");
   } catch (e) {
     logger.warn("process refresh sessions got exception");
     logger.warn(`exception: ${e}`);
   }
-  return {
-    type: OperationAttemptResultType.Finished,
-    result: undefined,
-  };
+  if (inShutdown) {
+    return TaskRunResult.pending();
+  }
+  if (errors.length > 0) {
+    return {
+      type: TaskRunResultType.Error,
+      errorDetail: makeErrorDetail(
+        TalerErrorCode.WALLET_REFRESH_GROUP_INCOMPLETE,
+        {
+          numErrors: errors.length,
+          errors: errors.slice(0, 5),
+        },
+      ),
+    };
+  }
+
+  return TaskRunResult.pending();
 }
 
 async function processRefreshSession(
@@ -793,13 +895,18 @@ async function processRefreshSession(
   refreshGroupId: string,
   coinIndex: number,
 ): Promise<void> {
-  logger.info(
+  logger.trace(
     `processing refresh session for coin ${coinIndex} of group ${refreshGroupId}`,
   );
-  let refreshGroup = await ws.db
-    .mktx((x) => [x.refreshGroups])
+  let { refreshGroup, refreshSession } = await ws.db
+    .mktx((x) => [x.refreshGroups, x.refreshSessions])
     .runReadOnly(async (tx) => {
-      return tx.refreshGroups.get(refreshGroupId);
+      const rg = await tx.refreshGroups.get(refreshGroupId);
+      const rs = await tx.refreshSessions.get([refreshGroupId, coinIndex]);
+      return {
+        refreshGroup: rg,
+        refreshSession: rs,
+      };
     });
   if (!refreshGroup) {
     return;
@@ -807,24 +914,13 @@ async function processRefreshSession(
   if (refreshGroup.statusPerCoin[coinIndex] === RefreshCoinStatus.Finished) {
     return;
   }
-  if (!refreshGroup.refreshSessionPerCoin[coinIndex]) {
-    await refreshCreateSession(ws, refreshGroupId, coinIndex);
-    refreshGroup = await ws.db
-      .mktx((x) => [x.refreshGroups])
-      .runReadOnly(async (tx) => {
-        return tx.refreshGroups.get(refreshGroupId);
-      });
-    if (!refreshGroup) {
-      return;
-    }
-  }
-  const refreshSession = refreshGroup.refreshSessionPerCoin[coinIndex];
   if (!refreshSession) {
-    if (refreshGroup.statusPerCoin[coinIndex] !== RefreshCoinStatus.Finished) {
-      throw Error(
-        "BUG: refresh session was not created and coin not marked as finished",
-      );
-    }
+    refreshSession = await provideRefreshSession(ws, refreshGroupId, coinIndex);
+  }
+  if (!refreshSession) {
+    // We tried to create the refresh session, but didn't get a result back.
+    // This means that either the session is finished, or that creating
+    // one isn't necessary.
     return;
   }
   if (refreshSession.norevealIndex === undefined) {
@@ -833,29 +929,21 @@ async function processRefreshSession(
   await refreshReveal(ws, refreshGroupId, coinIndex);
 }
 
-/**
- * Create a refresh group for a list of coins.
- *
- * Refreshes the remaining amount on the coin, effectively capturing the remaining
- * value in the refresh group.
- *
- * The caller must also ensure that the coins that should be refreshed exist
- * in the current database transaction.
- */
-export async function createRefreshGroup(
+export interface RefreshOutputInfo {
+  outputPerCoin: AmountJson[];
+}
+
+export async function calculateRefreshOutput(
   ws: InternalWalletState,
-  tx: GetReadWriteAccess<{
+  tx: GetReadOnlyAccess<{
     denominations: typeof WalletStoresV1.denominations;
     coins: typeof WalletStoresV1.coins;
     refreshGroups: typeof WalletStoresV1.refreshGroups;
     coinAvailability: typeof WalletStoresV1.coinAvailability;
   }>,
+  currency: string,
   oldCoinPubs: CoinRefreshRequest[],
-  reason: RefreshReason,
-): Promise<RefreshGroupId> {
-  const refreshGroupId = encodeCrock(getRandomBytes(32));
-
-  const inputPerCoin: AmountJson[] = [];
+): Promise<RefreshOutputInfo> {
   const estimatedOutputPerCoin: AmountJson[] = [];
 
   const denomsPerExchange: Record<string, DenominationRecord[]> = {};
@@ -869,12 +957,53 @@ export async function createRefreshGroup(
     const allDenoms = await tx.denominations.indexes.byExchangeBaseUrl
       .iter(exchangeBaseUrl)
       .filter((x) => {
-        return isWithdrawableDenom(x);
+        return isWithdrawableDenom(x, ws.config.testing.denomselAllowLate);
       });
     denomsPerExchange[exchangeBaseUrl] = allDenoms;
     return allDenoms;
   };
 
+  for (const ocp of oldCoinPubs) {
+    const coin = await tx.coins.get(ocp.coinPub);
+    checkDbInvariant(!!coin, "coin must be in database");
+    const denom = await ws.getDenomInfo(
+      ws,
+      tx,
+      coin.exchangeBaseUrl,
+      coin.denomPubHash,
+    );
+    checkDbInvariant(
+      !!denom,
+      "denomination for existing coin must be in database",
+    );
+    const refreshAmount = ocp.amount;
+    const denoms = await getDenoms(coin.exchangeBaseUrl);
+    const cost = getTotalRefreshCost(
+      denoms,
+      denom,
+      Amounts.parseOrThrow(refreshAmount),
+      ws.config.testing.denomselAllowLate,
+    );
+    const output = Amounts.sub(refreshAmount, cost).amount;
+    estimatedOutputPerCoin.push(output);
+  }
+
+  return {
+    outputPerCoin: estimatedOutputPerCoin,
+  };
+}
+
+async function applyRefresh(
+  ws: InternalWalletState,
+  tx: GetReadWriteAccess<{
+    denominations: typeof WalletStoresV1.denominations;
+    coins: typeof WalletStoresV1.coins;
+    refreshGroups: typeof WalletStoresV1.refreshGroups;
+    coinAvailability: typeof WalletStoresV1.coinAvailability;
+  }>,
+  oldCoinPubs: CoinRefreshRequest[],
+  refreshGroupId: string,
+): Promise<void> {
   for (const ocp of oldCoinPubs) {
     const coin = await tx.coins.get(ocp.coinPub);
     checkDbInvariant(!!coin, "coin must be in database");
@@ -916,54 +1045,74 @@ export async function createRefreshGroup(
     if (!coin.spendAllocation) {
       coin.spendAllocation = {
         amount: Amounts.stringify(ocp.amount),
-        id: `txn:refresh:${refreshGroupId}`,
+        // id: `txn:refresh:${refreshGroupId}`,
+        id: constructTransactionIdentifier({
+          tag: TransactionType.Refresh,
+          refreshGroupId,
+        }),
       };
     }
-    const refreshAmount = ocp.amount;
-    inputPerCoin.push(Amounts.parseOrThrow(refreshAmount));
     await tx.coins.put(coin);
-    const denoms = await getDenoms(coin.exchangeBaseUrl);
-    const cost = getTotalRefreshCost(
-      denoms,
-      denom,
-      Amounts.parseOrThrow(refreshAmount),
-    );
-    const output = Amounts.sub(refreshAmount, cost).amount;
-    estimatedOutputPerCoin.push(output);
   }
+}
+
+/**
+ * Create a refresh group for a list of coins.
+ *
+ * Refreshes the remaining amount on the coin, effectively capturing the remaining
+ * value in the refresh group.
+ *
+ * The caller must also ensure that the coins that should be refreshed exist
+ * in the current database transaction.
+ */
+export async function createRefreshGroup(
+  ws: InternalWalletState,
+  tx: GetReadWriteAccess<{
+    denominations: typeof WalletStoresV1.denominations;
+    coins: typeof WalletStoresV1.coins;
+    refreshGroups: typeof WalletStoresV1.refreshGroups;
+    coinAvailability: typeof WalletStoresV1.coinAvailability;
+  }>,
+  currency: string,
+  oldCoinPubs: CoinRefreshRequest[],
+  reason: RefreshReason,
+  reasonDetails?: RefreshReasonDetails,
+): Promise<RefreshGroupId> {
+  const refreshGroupId = encodeCrock(getRandomBytes(32));
+
+  const outInfo = await calculateRefreshOutput(ws, tx, currency, oldCoinPubs);
+
+  const estimatedOutputPerCoin = outInfo.outputPerCoin;
+
+  await applyRefresh(ws, tx, oldCoinPubs, refreshGroupId);
 
   const refreshGroup: RefreshGroupRecord = {
     operationStatus: RefreshOperationStatus.Pending,
+    currency,
     timestampFinished: undefined,
     statusPerCoin: oldCoinPubs.map(() => RefreshCoinStatus.Pending),
     oldCoinPubs: oldCoinPubs.map((x) => x.coinPub),
-    lastErrorPerCoin: {},
+    reasonDetails,
     reason,
     refreshGroupId,
-    refreshSessionPerCoin: oldCoinPubs.map(() => undefined),
-    inputPerCoin: inputPerCoin.map((x) => Amounts.stringify(x)),
-    estimatedOutputPerCoin: estimatedOutputPerCoin.map((x) =>
+    inputPerCoin: oldCoinPubs.map((x) => x.amount),
+    expectedOutputPerCoin: estimatedOutputPerCoin.map((x) =>
       Amounts.stringify(x),
     ),
-    timestampCreated: TalerProtocolTimestamp.now(),
+    timestampCreated: timestampPreciseToDb(TalerPreciseTimestamp.now()),
   };
 
   if (oldCoinPubs.length == 0) {
     logger.warn("created refresh group with zero coins");
-    refreshGroup.timestampFinished = TalerProtocolTimestamp.now();
+    refreshGroup.timestampFinished = timestampPreciseToDb(
+      TalerPreciseTimestamp.now(),
+    );
     refreshGroup.operationStatus = RefreshOperationStatus.Finished;
   }
 
   await tx.refreshGroups.put(refreshGroup);
 
-  logger.info(`created refresh group ${refreshGroupId}`);
-
-  processRefreshGroup(ws, refreshGroupId).catch((e) => {
-    if (e instanceof CryptoApiStoppedError) {
-      return;
-    }
-    logger.warn(`processing refresh group ${refreshGroupId} failed: ${e}`);
-  });
+  logger.trace(`created refresh group ${refreshGroupId}`);
 
   return {
     refreshGroupId,
@@ -974,8 +1123,12 @@ export async function createRefreshGroup(
  * Timestamp after which the wallet would do the next check for an auto-refresh.
  */
 function getAutoRefreshCheckThreshold(d: DenominationRecord): AbsoluteTime {
-  const expireWithdraw = AbsoluteTime.fromTimestamp(d.stampExpireWithdraw);
-  const expireDeposit = AbsoluteTime.fromTimestamp(d.stampExpireDeposit);
+  const expireWithdraw = AbsoluteTime.fromProtocolTimestamp(
+    timestampProtocolFromDb(d.stampExpireWithdraw),
+  );
+  const expireDeposit = AbsoluteTime.fromProtocolTimestamp(
+    timestampProtocolFromDb(d.stampExpireDeposit),
+  );
   const delta = AbsoluteTime.difference(expireWithdraw, expireDeposit);
   const deltaDiv = durationMul(delta, 0.75);
   return AbsoluteTime.addDuration(expireWithdraw, deltaDiv);
@@ -985,8 +1138,12 @@ function getAutoRefreshCheckThreshold(d: DenominationRecord): AbsoluteTime {
  * Timestamp after which the wallet would do an auto-refresh.
  */
 function getAutoRefreshExecuteThreshold(d: DenominationRecord): AbsoluteTime {
-  const expireWithdraw = AbsoluteTime.fromTimestamp(d.stampExpireWithdraw);
-  const expireDeposit = AbsoluteTime.fromTimestamp(d.stampExpireDeposit);
+  const expireWithdraw = AbsoluteTime.fromProtocolTimestamp(
+    timestampProtocolFromDb(d.stampExpireWithdraw),
+  );
+  const expireDeposit = AbsoluteTime.fromProtocolTimestamp(
+    timestampProtocolFromDb(d.stampExpireDeposit),
+  );
   const delta = AbsoluteTime.difference(expireWithdraw, expireDeposit);
   const deltaDiv = durationMul(delta, 0.5);
   return AbsoluteTime.addDuration(expireWithdraw, deltaDiv);
@@ -995,8 +1152,8 @@ function getAutoRefreshExecuteThreshold(d: DenominationRecord): AbsoluteTime {
 export async function autoRefresh(
   ws: InternalWalletState,
   exchangeBaseUrl: string,
-): Promise<OperationAttemptResult> {
-  logger.info(`doing auto-refresh check for '${exchangeBaseUrl}'`);
+): Promise<TaskRunResult> {
+  logger.trace(`doing auto-refresh check for '${exchangeBaseUrl}'`);
 
   // We must make sure that the exchange is up-to-date so that
   // can refresh into new denominations.
@@ -1018,7 +1175,7 @@ export async function autoRefresh(
     ])
     .runReadWrite(async (tx) => {
       const exchange = await tx.exchanges.get(exchangeBaseUrl);
-      if (!exchange) {
+      if (!exchange || !exchange.detailsPointer) {
         return;
       }
       const coins = await tx.coins.indexes.byBaseUrl
@@ -1041,11 +1198,7 @@ export async function autoRefresh(
         if (AbsoluteTime.isExpired(executeThreshold)) {
           refreshCoins.push({
             coinPub: coin.coinPub,
-            amount: Amounts.stringify({
-              value: denom.amountVal,
-              fraction: denom.amountFrac,
-              currency: denom.currency,
-            }),
+            amount: denom.value,
           });
         } else {
           const checkThreshold = getAutoRefreshCheckThreshold(denom);
@@ -1059,21 +1212,198 @@ export async function autoRefresh(
         const res = await createRefreshGroup(
           ws,
           tx,
+          exchange.detailsPointer?.currency,
           refreshCoins,
           RefreshReason.Scheduled,
         );
-        logger.info(
+        logger.trace(
           `created refresh group for auto-refresh (${res.refreshGroupId})`,
         );
       }
-      logger.info(
-        `current wallet time: ${AbsoluteTime.toIsoString(AbsoluteTime.now())}`,
-      );
-      logger.info(
+      //      logger.trace(
+      //        `current wallet time: ${AbsoluteTime.toIsoString(AbsoluteTime.now())}`,
+      //      );
+      logger.trace(
         `next refresh check at ${AbsoluteTime.toIsoString(minCheckThreshold)}`,
       );
-      exchange.nextRefreshCheck = AbsoluteTime.toTimestamp(minCheckThreshold);
+      exchange.nextRefreshCheckStamp = timestampPreciseToDb(
+        AbsoluteTime.toPreciseTimestamp(minCheckThreshold),
+      );
       await tx.exchanges.put(exchange);
     });
-  return OperationAttemptResult.finishedEmpty();
+  return TaskRunResult.finished();
+}
+
+export function computeRefreshTransactionState(
+  rg: RefreshGroupRecord,
+): TransactionState {
+  switch (rg.operationStatus) {
+    case RefreshOperationStatus.Finished:
+      return {
+        major: TransactionMajorState.Done,
+      };
+    case RefreshOperationStatus.Failed:
+      return {
+        major: TransactionMajorState.Failed,
+      };
+    case RefreshOperationStatus.Pending:
+      return {
+        major: TransactionMajorState.Pending,
+      };
+    case RefreshOperationStatus.Suspended:
+      return {
+        major: TransactionMajorState.Suspended,
+      };
+  }
+}
+
+export function computeRefreshTransactionActions(
+  rg: RefreshGroupRecord,
+): TransactionAction[] {
+  switch (rg.operationStatus) {
+    case RefreshOperationStatus.Finished:
+      return [TransactionAction.Delete];
+    case RefreshOperationStatus.Failed:
+      return [TransactionAction.Delete];
+    case RefreshOperationStatus.Pending:
+      return [TransactionAction.Suspend, TransactionAction.Fail];
+    case RefreshOperationStatus.Suspended:
+      return [TransactionAction.Resume, TransactionAction.Fail];
+  }
+}
+
+export async function suspendRefreshGroup(
+  ws: InternalWalletState,
+  refreshGroupId: string,
+): Promise<void> {
+  const transactionId = constructTransactionIdentifier({
+    tag: TransactionType.Refresh,
+    refreshGroupId,
+  });
+  let res = await ws.db
+    .mktx((x) => [x.refreshGroups])
+    .runReadWrite(async (tx) => {
+      const dg = await tx.refreshGroups.get(refreshGroupId);
+      if (!dg) {
+        logger.warn(
+          `can't suspend refresh group, refreshGroupId=${refreshGroupId} not found`,
+        );
+        return undefined;
+      }
+      const oldState = computeRefreshTransactionState(dg);
+      switch (dg.operationStatus) {
+        case RefreshOperationStatus.Finished:
+          return undefined;
+        case RefreshOperationStatus.Pending: {
+          dg.operationStatus = RefreshOperationStatus.Suspended;
+          await tx.refreshGroups.put(dg);
+          return {
+            oldTxState: oldState,
+            newTxState: computeRefreshTransactionState(dg),
+          };
+        }
+        case RefreshOperationStatus.Suspended:
+          return undefined;
+      }
+      return undefined;
+    });
+  if (res) {
+    ws.notify({
+      type: NotificationType.TransactionStateTransition,
+      transactionId,
+      oldTxState: res.oldTxState,
+      newTxState: res.newTxState,
+    });
+  }
+}
+
+export async function resumeRefreshGroup(
+  ws: InternalWalletState,
+  refreshGroupId: string,
+): Promise<void> {
+  const transactionId = constructTransactionIdentifier({
+    tag: TransactionType.Refresh,
+    refreshGroupId,
+  });
+  const transitionInfo = await ws.db
+    .mktx((x) => [x.refreshGroups])
+    .runReadWrite(async (tx) => {
+      const dg = await tx.refreshGroups.get(refreshGroupId);
+      if (!dg) {
+        logger.warn(
+          `can't resume refresh group, refreshGroupId=${refreshGroupId} not found`,
+        );
+        return;
+      }
+      const oldState = computeRefreshTransactionState(dg);
+      switch (dg.operationStatus) {
+        case RefreshOperationStatus.Finished:
+          return;
+        case RefreshOperationStatus.Pending: {
+          return;
+        }
+        case RefreshOperationStatus.Suspended:
+          dg.operationStatus = RefreshOperationStatus.Pending;
+          await tx.refreshGroups.put(dg);
+          return {
+            oldTxState: oldState,
+            newTxState: computeRefreshTransactionState(dg),
+          };
+      }
+      return undefined;
+    });
+  ws.workAvailable.trigger();
+  notifyTransition(ws, transactionId, transitionInfo);
+}
+
+export async function failRefreshGroup(
+  ws: InternalWalletState,
+  refreshGroupId: string,
+): Promise<void> {
+  throw Error("action cancel-aborting not allowed on refreshes");
+}
+
+export async function abortRefreshGroup(
+  ws: InternalWalletState,
+  refreshGroupId: string,
+): Promise<void> {
+  const transactionId = constructTransactionIdentifier({
+    tag: TransactionType.Refresh,
+    refreshGroupId,
+  });
+  const transitionInfo = await ws.db
+    .mktx((x) => [x.refreshGroups])
+    .runReadWrite(async (tx) => {
+      const dg = await tx.refreshGroups.get(refreshGroupId);
+      if (!dg) {
+        logger.warn(
+          `can't resume refresh group, refreshGroupId=${refreshGroupId} not found`,
+        );
+        return;
+      }
+      const oldState = computeRefreshTransactionState(dg);
+      let newStatus: RefreshOperationStatus | undefined;
+      switch (dg.operationStatus) {
+        case RefreshOperationStatus.Finished:
+          break;
+        case RefreshOperationStatus.Pending:
+        case RefreshOperationStatus.Suspended:
+          newStatus = RefreshOperationStatus.Failed;
+          break;
+        case RefreshOperationStatus.Failed:
+          break;
+        default:
+          assertUnreachable(dg.operationStatus);
+      }
+      if (newStatus) {
+        dg.operationStatus = newStatus;
+        await tx.refreshGroups.put(dg);
+      }
+      return {
+        oldTxState: oldState,
+        newTxState: computeRefreshTransactionState(dg),
+      };
+    });
+  ws.workAvailable.trigger();
+  notifyTransition(ws, transactionId, transitionInfo);
 }
