@@ -15,5 +15,5 @@
  */
 
 export function assertUnreachable(x: never): never {
-  throw new Error("Didn't expect to get here");
+  throw new Error(`Didn't expect to get here ${x}`);
 }
