@@ -35,7 +35,7 @@ import {
   IDBKeyPath,
   IDBKeyRange,
 } from "@gnu-taler/idb-bridge";
-import { Logger } from "@gnu-taler/taler-util";
+import { Codec, Logger, j2s } from "@gnu-taler/taler-util";
 
 const logger = new Logger("query.ts");
 
@@ -239,7 +239,7 @@ class ResultStream<T> {
 export function openDatabase(
   idbFactory: IDBFactory,
   databaseName: string,
-  databaseVersion: number,
+  databaseVersion: number | undefined,
   onVersionChange: () => void,
   onUpgradeNeeded: (
     db: IDBDatabase,
@@ -257,7 +257,7 @@ export function openDatabase(
     req.onsuccess = (e) => {
       req.result.onversionchange = (evt: IDBVersionChangeEvent) => {
         logger.info(
-          `handling live db version change from ${evt.oldVersion} to ${evt.newVersion}`,
+          `handling versionchange on ${databaseName} from ${evt.oldVersion} to ${evt.newVersion}`,
         );
         req.result.close();
         onVersionChange();
@@ -274,6 +274,9 @@ export function openDatabase(
       if (!transaction) {
         throw Error("no transaction handle available in upgrade handler");
       }
+      logger.info(
+        `handling upgradeneeded event on ${databaseName} from ${e.oldVersion} to ${e.newVersion}`,
+      );
       onUpgradeNeeded(db, e.oldVersion, newVersion, transaction);
     };
   });
@@ -303,7 +306,7 @@ export interface StoreOptions {
   autoIncrement?: boolean;
 
   /**
-   * Database version that this store was added in, or
+   * First minor database version that this store was added in, or
    * undefined if added in the first version.
    */
   versionAdded?: number;
@@ -338,7 +341,7 @@ interface IndexReadOnlyAccessor<RecordType> {
   iter(query?: IDBKeyRange | IDBValidKey): ResultStream<RecordType>;
   get(query: IDBValidKey): Promise<RecordType | undefined>;
   getAll(
-    query: IDBKeyRange | IDBValidKey,
+    query?: IDBKeyRange | IDBValidKey,
     count?: number,
   ): Promise<RecordType[]>;
 }
@@ -351,7 +354,7 @@ interface IndexReadWriteAccessor<RecordType> {
   iter(query: IDBKeyRange | IDBValidKey): ResultStream<RecordType>;
   get(query: IDBValidKey): Promise<RecordType | undefined>;
   getAll(
-    query: IDBKeyRange | IDBValidKey,
+    query?: IDBKeyRange | IDBValidKey,
     count?: number,
   ): Promise<RecordType[]>;
 }
@@ -376,19 +379,19 @@ export interface InsertResponse {
 export interface StoreReadWriteAccessor<RecordType, IndexMap> {
   get(key: IDBValidKey): Promise<RecordType | undefined>;
   iter(query?: IDBValidKey): ResultStream<RecordType>;
-  put(r: RecordType): Promise<InsertResponse>;
-  add(r: RecordType): Promise<InsertResponse>;
+  put(r: RecordType, key?: IDBValidKey): Promise<InsertResponse>;
+  add(r: RecordType, key?: IDBValidKey): Promise<InsertResponse>;
   delete(key: IDBValidKey): Promise<void>;
   indexes: GetIndexReadWriteAccess<RecordType, IndexMap>;
 }
 
 export interface StoreWithIndexes<
   StoreName extends string,
-  SD extends StoreDescriptor<unknown>,
+  RecordType,
   IndexMap,
 > {
   storeName: StoreName;
-  store: SD;
+  store: StoreDescriptor<RecordType>;
   indexMap: IndexMap;
 
   /**
@@ -398,19 +401,13 @@ export interface StoreWithIndexes<
   mark: Symbol;
 }
 
-export type GetRecordType<T> = T extends StoreDescriptor<infer X> ? X : unknown;
-
 const storeWithIndexesSymbol = Symbol("StoreWithIndexesMark");
 
-export function describeStore<
-  StoreName extends string,
-  SD extends StoreDescriptor<unknown>,
-  IndexMap,
->(
+export function describeStore<StoreName extends string, RecordType, IndexMap>(
   name: StoreName,
-  s: SD,
+  s: StoreDescriptor<RecordType>,
   m: IndexMap,
-): StoreWithIndexes<StoreName, SD, IndexMap> {
+): StoreWithIndexes<StoreName, RecordType, IndexMap> {
   return {
     storeName: name,
     store: s,
@@ -419,23 +416,122 @@ export function describeStore<
   };
 }
 
+export function describeStoreV2<
+  StoreName extends string,
+  RecordType,
+  IndexMap extends { [x: string]: IndexDescriptor } = {},
+>(args: {
+  storeName: StoreName;
+  recordCodec: Codec<RecordType>;
+  keyPath?: IDBKeyPath | IDBKeyPath[];
+  autoIncrement?: boolean;
+  /**
+   * Database version that this store was added in, or
+   * undefined if added in the first version.
+   */
+  versionAdded?: number;
+  indexes?: IndexMap;
+}): StoreWithIndexes<StoreName, RecordType, IndexMap> {
+  return {
+    storeName: args.storeName,
+    store: {
+      _dummy: undefined as any,
+      autoIncrement: args.autoIncrement,
+      keyPath: args.keyPath,
+      versionAdded: args.versionAdded,
+    },
+    indexMap: args.indexes ?? ({} as IndexMap),
+    mark: storeWithIndexesSymbol,
+  };
+}
+
+type KeyPathComponents = string | number;
+
+/**
+ * Follow a key path (dot-separated) in an object.
+ */
+type DerefKeyPath<T, P> = P extends `${infer PX extends keyof T &
+  KeyPathComponents}`
+  ? T[PX]
+  : P extends `${infer P0 extends keyof T & KeyPathComponents}.${infer Rest}`
+  ? DerefKeyPath<T[P0], Rest>
+  : unknown;
+
+/**
+ * Return a path if it is a valid dot-separate path to an object.
+ * Otherwise, return "never".
+ */
+type ValidateKeyPath<T, P> = P extends `${infer PX extends keyof T &
+  KeyPathComponents}`
+  ? PX
+  : P extends `${infer P0 extends keyof T & KeyPathComponents}.${infer Rest}`
+  ? `${P0}.${ValidateKeyPath<T[P0], Rest>}`
+  : never;
+
+// function foo<T, P>(
+//   x: T,
+//   p: P extends ValidateKeyPath<T, P> ? P : never,
+// ): void {}
+
+// foo({x: [0,1,2]}, "x.0");
+
 export type GetReadOnlyAccess<BoundStores> = {
   [P in keyof BoundStores]: BoundStores[P] extends StoreWithIndexes<
-    infer SN,
-    infer SD,
-    infer IM
+    infer StoreName,
+    infer RecordType,
+    infer IndexMap
   >
-    ? StoreReadOnlyAccessor<GetRecordType<SD>, IM>
+    ? StoreReadOnlyAccessor<RecordType, IndexMap>
     : unknown;
 };
 
+export type StoreNames<StoreMap> = StoreMap extends {
+  [P in keyof StoreMap]: StoreWithIndexes<infer SN1, infer SD1, infer IM1>;
+}
+  ? keyof StoreMap
+  : unknown;
+
+export type DbReadOnlyTransaction<
+  StoreMap,
+  Stores extends StoreNames<StoreMap> & string,
+> = StoreMap extends {
+  [P in Stores]: StoreWithIndexes<infer SN1, infer SD1, infer IM1>;
+}
+  ? {
+      [P in Stores]: StoreMap[P] extends StoreWithIndexes<
+        infer StoreName,
+        infer RecordType,
+        infer IndexMap
+      >
+        ? StoreReadOnlyAccessor<RecordType, IndexMap>
+        : unknown;
+    }
+  : unknown;
+
+export type DbReadWriteTransaction<
+  StoreMap,
+  Stores extends StoreNames<StoreMap> & string,
+> = StoreMap extends {
+  [P in Stores]: StoreWithIndexes<infer SN1, infer SD1, infer IM1>;
+}
+  ? {
+      [P in Stores]: StoreMap[P] extends StoreWithIndexes<
+        infer StoreName,
+        infer RecordType,
+        infer IndexMap
+      >
+        ? StoreReadWriteAccessor<RecordType, IndexMap>
+        : unknown;
+    }
+  : unknown;
+
 export type GetReadWriteAccess<BoundStores> = {
   [P in keyof BoundStores]: BoundStores[P] extends StoreWithIndexes<
-    infer SN,
-    infer SD,
-    infer IM
+    infer StoreName,
+    infer RecordType,
+    infer IndexMap
   >
-    ? StoreReadWriteAccessor<GetRecordType<SD>, IM>
+    ? StoreReadWriteAccessor<RecordType, IndexMap>
     : unknown;
 };
 
@@ -612,15 +708,15 @@ function makeWriteContext(
         const req = tx.objectStore(storeName).openCursor(query);
         return new ResultStream<any>(req);
       },
-      async add(r) {
-        const req = tx.objectStore(storeName).add(r);
+      async add(r, k) {
+        const req = tx.objectStore(storeName).add(r, k);
         const key = await requestToPromise(req);
         return {
           key: key,
         };
       },
-      async put(r) {
-        const req = tx.objectStore(storeName).put(r);
+      async put(r, k) {
+        const req = tx.objectStore(storeName).put(r, k);
         const key = await requestToPromise(req);
         return {
           key: key,
@@ -660,7 +756,6 @@ export class DbAccess<StoreMap> {
     const storeNames: string[] = [];
     const accessibleStores: { [x: string]: StoreWithIndexes<any, any, any> } =
       {};
-
     for (let i = 0; i < this.db.objectStoreNames.length; i++) {
       const sn = this.db.objectStoreNames[i];
       const swi = (this.stores as any)[sn] as StoreWithIndexes<any, any, any>;
@@ -671,6 +766,21 @@ export class DbAccess<StoreMap> {
       accessibleStores[sn] = swi;
     }
 
+    const storeMapKeys = Object.keys(this.stores as any);
+    for (const storeMapKey of storeMapKeys) {
+      const swi = (this.stores as any)[storeMapKey] as StoreWithIndexes<
+        any,
+        any,
+        any
+      >;
+      if (!accessibleStores[swi.storeName]) {
+        const version = this.db.version;
+        throw Error(
+          `store '${swi.storeName}' required by schema but not in database (minver=${version})`,
+        );
+      }
+    }
+
     const runReadOnly = <T>(
       txf: ReadOnlyTransactionFunction<StoreMap, T>,
     ): Promise<T> => {
