@@ -1,6 +1,6 @@
 /*
  This file is part of GNU Taler
- (C) 2019 Taler Systems S.A.
+ (C) 2019-2023 Taler Systems S.A.
 
  GNU Taler is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
@@ -19,18 +19,38 @@
  *
  * Uses libtool's current:revision:age versioning.
  */
-export const WALLET_EXCHANGE_PROTOCOL_VERSION = "12:0:0";
+export const WALLET_EXCHANGE_PROTOCOL_VERSION = "17:0:0";
 
 /**
  * Protocol version spoken with the merchant.
  *
  * Uses libtool's current:revision:age versioning.
  */
-export const WALLET_MERCHANT_PROTOCOL_VERSION = "2:0:1";
+export const WALLET_MERCHANT_PROTOCOL_VERSION = "5:0:1";
 
 /**
- * Protocol version spoken with the merchant.
+ * Protocol version spoken with the bank.
  *
  * Uses libtool's current:revision:age versioning.
  */
 export const WALLET_BANK_INTEGRATION_PROTOCOL_VERSION = "0:0:0";
+
+/**
+ * Semver of the wallet-core implementation.
+ * Will be replaced with the value from package.json in a
+ * post-compilation step (inside lib/).
+ */
+export const WALLET_CORE_IMPLEMENTATION_VERSION = "1:0:0";
+
+/**
+ * Libtool rules:
+ * 
+ * If the library source code has changed at all since the last update, 
+ *     then increment revision (‘c:r:a’ becomes ‘c:r+1:a’).
+ * If any interfaces have been added, removed, or changed since the last
+ *     update, increment current, and set revision to 0.
+ * If any interfaces have been added since the last public release, then 
+ *     increment age.
+ * If any interfaces have been removed or changed since the last public 
+ *     release, then set age to 0.
+ */
\ No newline at end of file
