@@ -24,16 +24,21 @@
 /**
  * Imports.
  */
-import { h, VNode } from "preact";
-import { JustInDevMode } from "./components/JustInDevMode.js";
+import { WalletApiOperation } from "@gnu-taler/taler-wallet-core";
+import { Fragment, h, VNode } from "preact";
+import { EnabledBySettings } from "./components/EnabledBySettings.js";
 import {
   NavigationHeader,
   NavigationHeaderHolder,
   SvgIcon,
 } from "./components/styled/index.js";
-import { useTranslationContext } from "./context/translation.js";
-import settingsIcon from "./svg/settings_black_24dp.svg";
-import qrIcon from "./svg/qr_code_24px.svg";
+import { useBackendContext } from "./context/backend.js";
+import { useAsyncAsHook } from "./hooks/useAsyncAsHook.js";
+import qrIcon from "./svg/qr_code_24px.inline.svg";
+import settingsIcon from "./svg/settings_black_24dp.inline.svg";
+import warningIcon from "./svg/warning_24px.inline.svg";
+import { parseTalerUri, TalerUriAction } from "@gnu-taler/taler-util";
+import { useTranslationContext } from "@gnu-taler/web-util/browser";
 
 /**
  * List of pages used by the wallet
@@ -41,6 +46,7 @@ import qrIcon from "./svg/qr_code_24px.svg";
  * @author sebasjm
  */
 
+// eslint-disable-next-line @typescript-eslint/ban-types
 type PageLocation<DynamicPart extends object> = {
   pattern: string;
   (params: DynamicPart): string;
@@ -53,11 +59,15 @@ function replaceAll(
 ): string {
   let result = pattern;
   for (const v in vars) {
-    result = result.replace(vars[v], !values[v] ? "" : values[v]);
+    result = result.replace(
+      vars[v],
+      !values[v] ? "" : encodeURIComponent(values[v]),
+    );
   }
   return result;
 }
 
+// eslint-disable-next-line @typescript-eslint/ban-types
 function pageDefinition<T extends object>(pattern: string): PageLocation<T> {
   const patternParams = pattern.match(/(:[\w?]*)/g);
   if (!patternParams)
@@ -74,7 +84,7 @@ function pageDefinition<T extends object>(pattern: string): PageLocation<T> {
     return { ...prev, [name]: cur };
   }, {} as Record<string, string>);
 
-  const f = (values: T): string => replaceAll(pattern, vars, values);
+  const f = (values: T): string => replaceAll(pattern, vars, values ?? {});
   f.pattern = pattern;
   return f;
 }
@@ -103,13 +113,16 @@ export const Pages = {
   backupProviderAdd: "/backup/provider/add",
 
   qr: "/qr",
+  notifications: "/notifications",
   settings: "/settings",
   settingsExchangeAdd: pageDefinition<{ currency?: string }>(
     "/settings/exchange/add/:currency?",
   ),
 
+  defaultCta: pageDefinition<{ uri: string }>("/taler-uri/:uri"),
   cta: pageDefinition<{ action: string }>("/cta/:action"),
   ctaPay: "/cta/pay",
+  ctaPayTemplate: "/cta/pay/template",
   ctaRecovery: "/cta/recovery",
   ctaRefund: "/cta/refund",
   ctaTips: "/cta/tip",
@@ -128,25 +141,73 @@ export const Pages = {
   ),
 };
 
-export function PopupNavBar({ path = "" }: { path?: string }): VNode {
+const talerUriActionToPageName: {
+  [t in TalerUriAction]: keyof typeof Pages | undefined;
+} = {
+  [TalerUriAction.Withdraw]: "ctaWithdraw",
+  [TalerUriAction.Pay]: "ctaPay",
+  [TalerUriAction.Reward]: "ctaTips",
+  [TalerUriAction.Refund]: "ctaRefund",
+  [TalerUriAction.PayPull]: "ctaInvoicePay",
+  [TalerUriAction.PayPush]: "ctaTransferPickup",
+  [TalerUriAction.Restore]: "ctaRecovery",
+  [TalerUriAction.PayTemplate]: "ctaPayTemplate",
+  [TalerUriAction.WithdrawExchange]: "ctaWithdrawManual",
+  [TalerUriAction.DevExperiment]: undefined,
+  [TalerUriAction.Exchange]: undefined,
+  [TalerUriAction.Auditor]: undefined,
+};
+
+export function getPathnameForTalerURI(talerUri: string): string | undefined {
+  const uri = parseTalerUri(talerUri);
+  if (!uri) {
+    return undefined;
+  }
+  const pageName = talerUriActionToPageName[uri.type];
+  if (!pageName) {
+    return undefined;
+  }
+  const pageString: string =
+    typeof Pages[pageName] === "function"
+      ? (Pages[pageName] as any)()
+      : Pages[pageName];
+  return `${pageString}?talerUri=${encodeURIComponent(talerUri)}`;
+}
+
+export type PopupNavBarOptions = "balance" | "backup" | "dev";
+export function PopupNavBar({ path }: { path?: PopupNavBarOptions }): VNode {
+  const api = useBackendContext();
+  const hook = useAsyncAsHook(async () => {
+    return await api.wallet.call(
+      WalletApiOperation.GetUserAttentionUnreadCount,
+      {},
+    );
+  });
+  const attentionCount = !hook || hook.hasError ? 0 : hook.response.total;
+
   const { i18n } = useTranslationContext();
   return (
     <NavigationHeader>
-      <a
-        href={Pages.balance}
-        class={path.startsWith("/balance") ? "active" : ""}
-      >
+      <a href={Pages.balance} class={path === "balance" ? "active" : ""}>
         <i18n.Translate>Balance</i18n.Translate>
       </a>
-      <JustInDevMode>
-        <a
-          href={Pages.backup}
-          class={path.startsWith("/backup") ? "active" : ""}
-        >
+      <EnabledBySettings name="backup">
+        <a href={Pages.backup} class={path === "backup" ? "active" : ""}>
           <i18n.Translate>Backup</i18n.Translate>
         </a>
-      </JustInDevMode>
+      </EnabledBySettings>
       <div style={{ display: "flex", paddingTop: 4, justifyContent: "right" }}>
+        {attentionCount > 0 ? (
+          <a href={Pages.notifications}>
+            <SvgIcon
+              title={i18n.str`Notifications`}
+              dangerouslySetInnerHTML={{ __html: warningIcon }}
+              color="yellow"
+            />
+          </a>
+        ) : (
+          <Fragment />
+        )}
         <a href={Pages.qr}>
           <SvgIcon
             title={i18n.str`QR Reader and Taler URI`}
@@ -165,30 +226,45 @@ export function PopupNavBar({ path = "" }: { path?: string }): VNode {
     </NavigationHeader>
   );
 }
-
-export function WalletNavBar({ path = "" }: { path?: string }): VNode {
+export type WalletNavBarOptions = "balance" | "backup" | "dev";
+export function WalletNavBar({ path }: { path?: WalletNavBarOptions }): VNode {
   const { i18n } = useTranslationContext();
+
+  const api = useBackendContext();
+  const hook = useAsyncAsHook(async () => {
+    return await api.wallet.call(
+      WalletApiOperation.GetUserAttentionUnreadCount,
+      {},
+    );
+  });
+  const attentionCount =
+    (!hook || hook.hasError ? 0 : hook.response?.total) ?? 0;
+
   return (
     <NavigationHeaderHolder>
       <NavigationHeader>
-        <a
-          href={Pages.balance}
-          class={path.startsWith("/balance") ? "active" : ""}
-        >
+        <a href={Pages.balance} class={path === "balance" ? "active" : ""}>
           <i18n.Translate>Balance</i18n.Translate>
         </a>
-        <JustInDevMode>
-          <a
-            href={Pages.backup}
-            class={path.startsWith("/backup") ? "active" : ""}
-          >
+        <EnabledBySettings name="backup">
+          <a href={Pages.backup} class={path === "backup" ? "active" : ""}>
             <i18n.Translate>Backup</i18n.Translate>
           </a>
+        </EnabledBySettings>
 
-          <a href={Pages.dev} class={path.startsWith("/dev") ? "active" : ""}>
-            <i18n.Translate>Dev</i18n.Translate>
+        {attentionCount > 0 ? (
+          <a href={Pages.notifications}>
+            <i18n.Translate>Notifications</i18n.Translate>
           </a>
-        </JustInDevMode>
+        ) : (
+          <Fragment />
+        )}
+
+        <EnabledBySettings name="advanceMode">
+          <a href={Pages.dev} class={path === "dev" ? "active" : ""}>
+            <i18n.Translate>Dev tools</i18n.Translate>
+          </a>
+        </EnabledBySettings>
 
         <div
           style={{ display: "flex", paddingTop: 4, justifyContent: "right" }}
