@@ -19,11 +19,11 @@
  * @author Sebastian Javier Marchano (sebasjm)
  */
 
-import { createExample } from "../../test-utils.js";
+import * as tests from "@gnu-taler/web-util/testing";
 // import { ReadyView } from "./views.js";
 
 export default {
   title: "TermsOfService",
 };
 
-// export const Ready = createExample(ReadyView, {});
+// export const Ready = tests.createExample(ReadyView, {});
