@@ -15,31 +15,34 @@
  */
 
 import {
+  AbsoluteTime,
   Amounts,
-  Balance,
   NotificationType,
   Transaction,
+  WalletBalance,
 } from "@gnu-taler/taler-util";
 import { WalletApiOperation } from "@gnu-taler/taler-wallet-core";
-import { Fragment, h, VNode } from "preact";
+import { useTranslationContext } from "@gnu-taler/web-util/browser";
+import { Fragment, VNode, h } from "preact";
 import { useEffect, useState } from "preact/hooks";
+import { ErrorAlertView } from "../components/CurrentAlerts.js";
+import { HistoryItem } from "../components/HistoryItem.js";
 import { Loading } from "../components/Loading.js";
-import { LoadingError } from "../components/LoadingError.js";
+import { Time } from "../components/Time.js";
 import {
   CenteredBoldText,
   CenteredText,
   DateSeparator,
   NiceSelect,
 } from "../components/styled/index.js";
-import { Time } from "../components/Time.js";
-import { TransactionItem } from "../components/TransactionItem.js";
-import { useTranslationContext } from "../context/translation.js";
+import { alertFromError, useAlertContext } from "../context/alert.js";
+import { useBackendContext } from "../context/backend.js";
 import { useAsyncAsHook } from "../hooks/useAsyncAsHook.js";
 import { Button } from "../mui/Button.js";
 import { NoBalanceHelp } from "../popup/NoBalanceHelp.js";
-import DownloadIcon from "../svg/download_24px.svg";
-import UploadIcon from "../svg/upload_24px.svg";
-import { wxApi } from "../wxApi.js";
+import DownloadIcon from "../svg/download_24px.inline.svg";
+import UploadIcon from "../svg/upload_24px.inline.svg";
+import { getDate, startOfDay } from "date-fns";
 
 interface Props {
   currency?: string;
@@ -52,14 +55,15 @@ export function HistoryPage({
   goToWalletDeposit,
 }: Props): VNode {
   const { i18n } = useTranslationContext();
+  const api = useBackendContext();
   const state = useAsyncAsHook(async () => ({
-    b: await wxApi.wallet.call(WalletApiOperation.GetBalances, {}),
-    tx: await wxApi.wallet.call(WalletApiOperation.GetTransactions, {}),
+    b: await api.wallet.call(WalletApiOperation.GetBalances, {}),
+    tx: await api.wallet.call(WalletApiOperation.GetTransactions, {}),
   }));
 
   useEffect(() => {
-    return wxApi.listener.onUpdateNotification(
-      [NotificationType.WithdrawGroupFinished],
+    return api.listener.onUpdateNotification(
+      [NotificationType.TransactionStateTransition],
       state?.retry,
     );
   });
@@ -70,13 +74,11 @@ export function HistoryPage({
 
   if (state.hasError) {
     return (
-      <LoadingError
-        title={
-          <i18n.Translate>
-            Could not load the list of transactions
-          </i18n.Translate>
-        }
-        error={state}
+      <ErrorAlertView
+        error={alertFromError(
+          i18n.str`Could not load the list of transactions`,
+          state,
+        )}
       />
     );
   }
@@ -108,10 +110,30 @@ export function HistoryView({
   goToWalletManualWithdraw: (currency?: string) => Promise<void>;
   defaultCurrency?: string;
   transactions: Transaction[];
-  balances: Balance[];
+  balances: WalletBalance[];
 }): VNode {
   const { i18n } = useTranslationContext();
-  const currencies = balances.map((b) => b.available.split(":")[0]);
+  const { pushAlertOnError } = useAlertContext();
+
+  const transactionByCurrency = transactions.reduce((prev, cur) => {
+    const c = Amounts.parseOrThrow(cur.amountEffective).currency;
+    if (!prev[c]) {
+      prev[c] = [];
+    }
+    prev[c].push(cur);
+    return prev;
+  }, {} as Record<string, Transaction[]>);
+
+  const currencies = balances
+    .filter((b) => {
+      const av = Amounts.parseOrThrow(b.available);
+      return (
+        Amounts.isNonZero(av) ||
+        (transactionByCurrency[av.currency] &&
+          transactionByCurrency[av.currency].length > 0)
+      );
+    })
+    .map((b) => b.available.split(":")[0]);
 
   const defaultCurrencyIndex = currencies.findIndex(
     (c) => c === defaultCurrency,
@@ -126,26 +148,31 @@ export function HistoryView({
     ? Amounts.jsonifyAmount(balances[currencyIndex].available)
     : undefined;
 
-  const byDate = transactions
-    .filter((t) => t.amountRaw.split(":")[0] === selectedCurrency)
-    .reduce((rv, x) => {
-      const theDate =
-        x.timestamp.t_s === "never"
-          ? 0
-          : normalizeToDay(x.timestamp.t_s * 1000);
-      if (theDate) {
-        (rv[theDate] = rv[theDate] || []).push(x);
-      }
+  const ts =
+    selectedCurrency === undefined
+      ? []
+      : transactionByCurrency[selectedCurrency] ?? [];
 
-      return rv;
-    }, {} as { [x: string]: Transaction[] });
-  const datesWithTransaction = Object.keys(byDate);
+  const datesWithTransaction: string[] = [];
+  const byDate = ts.reduce((rv, x) => {
+    const startDay =
+      x.timestamp.t_s === "never" ? 0 : startOfDay(x.timestamp.t_s * 1000).getTime();
+    if (startDay) {
+      if (!rv[startDay]) {
+        rv[startDay] = [];
+        datesWithTransaction.push(String(startDay));
+      }
+      rv[startDay].push(x);
+    }
+
+    return rv;
+  }, {} as { [x: string]: Transaction[] });
 
   if (balances.length === 0 || !selectedCurrency) {
     return (
       <NoBalanceHelp
         goToWalletManualWithdraw={{
-          onClick: goToWalletManualWithdraw,
+          onClick: pushAlertOnError(goToWalletManualWithdraw),
         }}
       />
     );
@@ -240,12 +267,14 @@ export function HistoryView({
               <Fragment key={i}>
                 <DateSeparator>
                   <Time
-                    timestamp={{ t_ms: Number.parseInt(d, 10) }}
+                    timestamp={AbsoluteTime.fromMilliseconds(
+                      Number.parseInt(d, 10),
+                    )}
                     format="dd MMMM yyyy"
                   />
                 </DateSeparator>
                 {byDate[d].map((tx, i) => (
-                  <TransactionItem key={i} tx={tx} />
+                  <HistoryItem key={i} tx={tx} />
                 ))}
               </Fragment>
             );
