@@ -17,13 +17,16 @@
 import {
   ExchangeListItem,
   ExchangeTosStatus,
+  LibtoolVersion,
+  TranslatedString,
   WalletCoreVersion,
 } from "@gnu-taler/taler-util";
 import { WalletApiOperation } from "@gnu-taler/taler-wallet-core";
-import { Fragment, h, VNode } from "preact";
+import { useTranslationContext } from "@gnu-taler/web-util/browser";
+import { Fragment, VNode, h } from "preact";
+import { Pages } from "../NavigationBar.js";
 import { Checkbox } from "../components/Checkbox.js";
-import { ErrorTalerOperation } from "../components/ErrorTalerOperation.js";
-import { JustInDevMode } from "../components/JustInDevMode.js";
+import { EnabledBySettings } from "../components/EnabledBySettings.js";
 import { Part } from "../components/Part.js";
 import { SelectList } from "../components/SelectList.js";
 import {
@@ -32,31 +35,31 @@ import {
   LinkPrimary,
   SubTitle,
   SuccessText,
+  WarningBox,
   WarningText,
 } from "../components/styled/index.js";
-import { useDevContext } from "../context/devContext.js";
-import { useTranslationContext } from "../context/translation.js";
+import { useAlertContext } from "../context/alert.js";
+import { useBackendContext } from "../context/backend.js";
 import { useAsyncAsHook } from "../hooks/useAsyncAsHook.js";
-import { useAutoOpenPermissions } from "../hooks/useAutoOpenPermissions.js";
 import { useBackupDeviceName } from "../hooks/useBackupDeviceName.js";
-import { useClipboardPermissions } from "../hooks/useClipboardPermissions.js";
+import { useSettings } from "../hooks/useSettings.js";
 import { ToggleHandler } from "../mui/handlers.js";
-import { Pages } from "../NavigationBar.js";
-import { platform } from "../platform/api.js";
-import { wxApi } from "../wxApi.js";
+import { Settings } from "../platform/api.js";
+import { platform } from "../platform/foreground.js";
+import { WALLET_CORE_SUPPORTED_VERSION } from "../wxApi.js";
 
 const GIT_HASH = typeof __GIT_HASH__ !== "undefined" ? __GIT_HASH__ : undefined;
 
 export function SettingsPage(): VNode {
-  const autoOpenToggle = useAutoOpenPermissions();
-  const clipboardToggle = useClipboardPermissions();
-  const { devModeToggle } = useDevContext();
+  const [settings, updateSettings] = useSettings();
+  const { safely } = useAlertContext();
   const { name, update } = useBackupDeviceName();
   const webex = platform.getWalletWebExVersion();
+  const api = useBackendContext();
 
   const exchangesHook = useAsyncAsHook(async () => {
-    const list = await wxApi.wallet.call(WalletApiOperation.ListExchanges, {});
-    const version = await wxApi.wallet.call(WalletApiOperation.GetVersion, {});
+    const list = await api.wallet.call(WalletApiOperation.ListExchanges, {});
+    const version = await api.wallet.call(WalletApiOperation.GetVersion, {});
     return { exchanges: list.exchanges, version };
   });
   const { exchanges, version } =
@@ -69,9 +72,38 @@ export function SettingsPage(): VNode {
       knownExchanges={exchanges}
       deviceName={name}
       setDeviceName={update}
-      autoOpenToggle={autoOpenToggle}
-      clipboardToggle={clipboardToggle}
-      devModeToggle={devModeToggle}
+      autoOpenToggle={{
+        value: settings.autoOpen,
+        button: {
+          onClick: safely("update support injection", async () => {
+            updateSettings("autoOpen", !settings.autoOpen);
+          }),
+        },
+      }}
+      injectTalerToggle={{
+        value: settings.injectTalerSupport,
+        button: {
+          onClick: safely("update support injection", async () => {
+            updateSettings("injectTalerSupport", !settings.injectTalerSupport);
+          }),
+        },
+      }}
+      advanceToggle={{
+        value: settings.advanceMode,
+        button: {
+          onClick: safely("update advance mode", async () => {
+            updateSettings("advanceMode", !settings.advanceMode);
+          }),
+        },
+      }}
+      langToggle={{
+        value: settings.langSelector,
+        button: {
+          onClick: safely("update lang selector", async () => {
+            updateSettings("langSelector", !settings.langSelector);
+          }),
+        },
+      }}
       webexVersion={{
         version: webex.version,
         hash: GIT_HASH,
@@ -85,8 +117,9 @@ export interface ViewProps {
   deviceName: string;
   setDeviceName: (s: string) => Promise<void>;
   autoOpenToggle: ToggleHandler;
-  clipboardToggle: ToggleHandler;
-  devModeToggle: ToggleHandler;
+  injectTalerToggle: ToggleHandler;
+  advanceToggle: ToggleHandler;
+  langToggle: ToggleHandler;
   knownExchanges: Array<ExchangeListItem>;
   coreVersion: WalletCoreVersion | undefined;
   webexVersion: {
@@ -98,8 +131,9 @@ export interface ViewProps {
 export function SettingsView({
   knownExchanges,
   autoOpenToggle,
-  clipboardToggle,
-  devModeToggle,
+  injectTalerToggle,
+  advanceToggle,
+  langToggle,
   coreVersion,
   webexVersion,
 }: ViewProps): VNode {
@@ -108,54 +142,6 @@ export function SettingsView({
   return (
     <Fragment>
       <section>
-        {autoOpenToggle.button.error && (
-          <ErrorTalerOperation
-            title={<i18n.Translate>Could not toggle auto-open</i18n.Translate>}
-            error={autoOpenToggle.button.error.errorDetail}
-          />
-        )}
-        {clipboardToggle.button.error && (
-          <ErrorTalerOperation
-            title={<i18n.Translate>Could not toggle clipboard</i18n.Translate>}
-            error={clipboardToggle.button.error.errorDetail}
-          />
-        )}
-        <SubTitle>
-          <i18n.Translate>Navigator</i18n.Translate>
-        </SubTitle>
-        <Checkbox
-          label={
-            <i18n.Translate>
-              Automatically open wallet based on page content
-            </i18n.Translate>
-          }
-          name="autoOpen"
-          description={
-            <i18n.Translate>
-              Enabling this option below will make using the wallet faster, but
-              requires more permissions from your browser.
-            </i18n.Translate>
-          }
-          enabled={autoOpenToggle.value!}
-          onToggle={autoOpenToggle.button.onClick!}
-        />
-        <Checkbox
-          label={
-            <i18n.Translate>
-              Automatically check clipboard for Taler URI
-            </i18n.Translate>
-          }
-          name="clipboard"
-          description={
-            <i18n.Translate>
-              Enabling this option below will make using the wallet faster, but
-              requires more permissions from your browser.
-            </i18n.Translate>
-          }
-          enabled={clipboardToggle.value!}
-          onToggle={clipboardToggle.button.onClick!}
-        />
-
         <SubTitle>
           <i18n.Translate>Trust</i18n.Translate>
         </SubTitle>
@@ -189,20 +175,17 @@ export function SettingsView({
                             <i18n.Translate>ok</i18n.Translate>
                           </SuccessText>
                         );
-                      case ExchangeTosStatus.Changed:
+                      case ExchangeTosStatus.Pending:
                         return (
                           <WarningText>
-                            <i18n.Translate>changed</i18n.Translate>
+                            <i18n.Translate>pending</i18n.Translate>
                           </WarningText>
                         );
-                      case ExchangeTosStatus.New:
-                      case ExchangeTosStatus.NotFound:
+                      case ExchangeTosStatus.Proposed:
                         return (
-                          <DestructiveText>
-                            <i18n.Translate>not accepted</i18n.Translate>
-                          </DestructiveText>
+                          <i18n.Translate>proposed</i18n.Translate>
                         );
-                      case ExchangeTosStatus.Unknown:
+                      default:
                         return (
                           <DestructiveText>
                             <i18n.Translate>
@@ -235,22 +218,47 @@ export function SettingsView({
           </LinkPrimary>
         </div>
 
+        {coreVersion && (<Fragment>
+          {LibtoolVersion.compare(coreVersion.version, WALLET_CORE_SUPPORTED_VERSION)?.compatible ? undefined :
+            <WarningBox>
+              <i18n.Translate>
+                The version of wallet core is not supported. (supported version: {WALLET_CORE_SUPPORTED_VERSION})
+              </i18n.Translate>
+            </WarningBox>}
+          <EnabledBySettings name="advanceMode">
+            <Part
+              title={i18n.str`Exchange compatibility`}
+              text={<span>{coreVersion.exchange}</span>}
+            />
+            <Part
+              title={i18n.str`Merchant compatibility`}
+              text={<span>{coreVersion.merchant}</span>}
+            />
+            <Part
+              title={i18n.str`Bank compatibility`}
+              text={<span>{coreVersion.bank}</span>}
+            />
+            <Part
+              title={i18n.str`Wallet Core compatibility`}
+              text={<span>{coreVersion.version}</span>}
+            />
+          </EnabledBySettings>
+        </Fragment>
+        )}
         <SubTitle>
-          <i18n.Translate>Troubleshooting</i18n.Translate>
+          <i18n.Translate>Advance mode</i18n.Translate>
         </SubTitle>
         <Checkbox
-          label={<i18n.Translate>Developer mode</i18n.Translate>}
+          label={i18n.str`Enable advance mode`}
           name="devMode"
-          description={
-            <i18n.Translate>
-              More options and information useful for debugging
-            </i18n.Translate>
-          }
-          enabled={devModeToggle.value!}
-          onToggle={devModeToggle.button.onClick!}
+          description={i18n.str`Show more information and options in the UI`}
+          enabled={advanceToggle.value!}
+          onToggle={advanceToggle.button.onClick!}
         />
-
-        <JustInDevMode>
+        <EnabledBySettings name="advanceMode">
+          <AdvanceSettings />
+        </EnabledBySettings>
+        <EnabledBySettings name="langSelector">
           <SubTitle>
             <i18n.Translate>Display</i18n.Translate>
           </SubTitle>
@@ -263,46 +271,103 @@ export function SettingsView({
               onChange={(v) => changeLanguage(v)}
             />
           </Input>
-        </JustInDevMode>
+        </EnabledBySettings>
+        <SubTitle>
+          <i18n.Translate>Navigator</i18n.Translate>
+        </SubTitle>
+        <Checkbox
+          label={i18n.str`Inject Taler support in all pages`}
+          name="inject"
+          description={
+            <i18n.Translate>
+              Disabling this option will make some web application not able to
+              trigger the wallet when clicking links but you will be able to
+              open the wallet using the keyboard shortcut
+            </i18n.Translate>
+          }
+          enabled={injectTalerToggle.value!}
+          onToggle={injectTalerToggle.button.onClick!}
+        />
+        <Checkbox
+          label={i18n.str`Automatically open wallet`}
+          name="autoOpen"
+          description={
+            <i18n.Translate>
+              Open the wallet when a payment action is found.
+            </i18n.Translate>
+          }
+          enabled={autoOpenToggle.value!}
+          onToggle={autoOpenToggle.button.onClick!}
+        />
         <SubTitle>
           <i18n.Translate>Version</i18n.Translate>
         </SubTitle>
-        {coreVersion && (
-          <Part
-            title={<i18n.Translate>Wallet Core</i18n.Translate>}
-            text={
-              <span>
-                {coreVersion.version}{" "}
-                <JustInDevMode>{coreVersion.hash}</JustInDevMode>
-              </span>
-            }
-          />
-        )}
         <Part
-          title={<i18n.Translate>Web Extension</i18n.Translate>}
+          title={i18n.str`Web Extension`}
           text={
             <span>
               {webexVersion.version}{" "}
-              <JustInDevMode>{webexVersion.hash}</JustInDevMode>
+              <EnabledBySettings name="advanceMode">
+                {webexVersion.hash}
+              </EnabledBySettings>
             </span>
           }
         />
-        {coreVersion && (
-          <JustInDevMode>
-            <Part
-              title={<i18n.Translate>Exchange compatibility</i18n.Translate>}
-              text={<span>{coreVersion.exchange}</span>}
+      </section>
+    </Fragment>
+  );
+}
+
+type Info = { label: TranslatedString; description: TranslatedString };
+type Options = {
+  [k in keyof Settings]?: Info;
+};
+function AdvanceSettings(): VNode {
+  const [settings, updateSettings] = useSettings();
+  const { i18n } = useTranslationContext();
+  const o: Options = {
+    backup: {
+      label: i18n.str`Show backup feature`,
+      description: i18n.str`Backup integration still in beta.`,
+    },
+    suspendIndividualTransaction: {
+      label: i18n.str`Show suspend/resume transaction`,
+      description: i18n.str`Prevent transaction from doing network request.`,
+    },
+    extendedAccountTypes: {
+      label: i18n.str`Show more account types on deposit`,
+      description: i18n.str`Extends the UI to more payment target types.`,
+    },
+    showJsonOnError: {
+      label: i18n.str`Show JSON on error`,
+      description: i18n.str`Print more information about the error. Useful for debugging.`,
+    },
+    walletAllowHttp: {
+      label: i18n.str`Allow HTTP connections`,
+      description: i18n.str`Using HTTP connection may be faster but unsafe (wallet restart required)`,
+    },
+    langSelector: {
+      label: i18n.str`Lang selector`,
+      description: i18n.str`Allows to manually change the language of the UI. Otherwise it will be automatically selected by your browser configuration.`,
+    },
+  };
+  return (
+    <Fragment>
+      <section>
+        {Object.entries(o).map(([name, { label, description }]) => {
+          const settingsName = name as keyof Settings;
+          return (
+            <Checkbox
+              label={label}
+              name={name}
+              description={description}
+              enabled={settings[settingsName]}
+              onToggle={async () => {
+                updateSettings(settingsName, !settings[settingsName]);
+              }}
             />
-            <Part
-              title={<i18n.Translate>Merchant compatibility</i18n.Translate>}
-              text={<span>{coreVersion.merchant}</span>}
-            />
-            <Part
-              title={<i18n.Translate>Bank compatibility</i18n.Translate>}
-              text={<span>{coreVersion.bank}</span>}
-            />
-          </JustInDevMode>
-        )}
+          );
+        })}
       </section>
     </Fragment>
   );
